"use strict";

/* Copyright 2002-2018 MarkLogic Corporation. All Rights Reserved. */

const util = require('./lib/util');

class EndpointError {
  constructor(statusCode, response) {
    this.name = "EndpointError";
    this.statusCode = statusCode;
    this.response = response;
    switch (statusCode) {
    case 400:
      this.message = 'Bad Request';
      break;
    case 401:
      this.message = 'Unauthorized';
      break;
    case 403:
      this.message = 'Forbidden';
      break;
    case 404:
      this.message = 'Not Found';
      break;
    default:
      this.statusCode = 500;
      this.message = 'Internal Server Error';
      break;
    }
  }
  toString() {
    return this.response;
  }
}
exports.EndpointError = EndpointError;

class Endpoint {
  static handleErrors(responseFunc) {
    try {
      return responseFunc() || {};
    } catch (err) {
      if (err.statusCode && err.statusCode < 500) {
        xdmp.setResponseCode(err.statusCode, err.message);
        return err;
      } else {
        throw err;
      }
    }
  }

  getRequestBody() {
    try {
      return xdmp.getRequestBody().toObject();
    }
    catch (err) {
      throw new EndpointError(400, err.stack);
    }
  }
  // http://stackoverflow.com/questions/8486099/
  //        how-do-i-parse-a-url-query-parameters-in-javascript
  // Debugged by nwalsh
  getJsonFromUrl(query) {
    var result = {};
    query.split("&").forEach(function (part) {
      if (part) {
        // replace every + with space, regexp-free version
        part = part.split("+").join(" ");
        var eq = part.indexOf("=");
        var key = eq > -1 ? part.substr(0, eq) : part;
        var val = eq > -1 ? decodeURIComponent(part.substr(eq + 1)) : "";
        var from = key.indexOf("[");
        if (from == -1) {
          key = decodeURIComponent(key);
          if (result[key] === undefined) {
            result[key] = val;
          } else if (Array.isArray(result[key])) {
            result[key].push(val);
          } else {
            let arr = [ result[key], val ];
            result[key] = arr;
          }
        } else {
          var to = key.indexOf("]");
          var index = decodeURIComponent(key.substring(from + 1, to));
          key = decodeURIComponent(key.substring(0, from));
          if (!result[key])
            result[key] = [];
          if (!index)
            result[key].push(val);
          else
            result[key][index] = val;
        }
      }
    });
    return result;
  }

  getRequestParams() {
    try {
      let origUrl = xdmp.getOriginalUrl();
      let result = {};
      if (origUrl.includes("?")) {
        result = this.getJsonFromUrl(origUrl.replace(/^[^\?]*\?/, ''));
      }
      return result;
    }
    catch (err) {
      throw new EndpointError(400, err.stack);
    }
  }

  static LOG(msg) {
    return xdmp.log('OpsDirector error: ' + msg + " (in " + xdmp.getRequestPath() + ")");
  }

  static securityAssert() {
    xdmp.securityAssert("http://marklogic.com/xdmp/privileges/opsdir-user",
                        "execute");
  }

  GET() {
    Endpoint.LOG("GET() not implemented");
    xdmp.setResponseCode(404, 'Not Found');
  }

  PUT() {
    Endpoint.LOG("PUT() not implemented");
    xdmp.setResponseCode(404, 'Not Found');
  }

  POST() {
    Endpoint.LOG("POST() not implemented");
    xdmp.setResponseCode(404, 'Not Found');
  }

  DELETE() {
    Endpoint.LOG("DELETE() not implemented");
    xdmp.setResponseCode(404, 'Not Found');
  }

  MISSING(reqMethod) {
    Endpoint.LOG(reqMethod + "() not implemented");
    xdmp.setResponseCode(404, 'Not Found');
  }

  static respond(t) {
    // We have to let the session endpoint run w/o auth because that's how we login
    if (xdmp.getRequestPath() == "/app/endpoints/session.sjs"
        || xdmp.getRequestPath() == "/app/endpoints/motd.sjs"
        || xdmp.getRequestPath() == "/app/endpoints/email.sjs") {
      // Nevermind, we don't want to do auth on these...
    } else {
      // This doesn't get caught by handleErrors so we have to do it specially
      try {
        Endpoint.securityAssert();
      } catch (error) {
        if (error.name !== "SEC-PRIV") {
          throw error;
        }
        xdmp.setResponseCode(401, "Unauthorized");
        return {
          "name": "EndpointError",
          "statusCode": 401,
          "message": "Unauthorized"
        };
      }
    }

    var reqMethod = xdmp.getRequestMethod();

    // Register endpoints with feature metrics
    var feature = xdmp.getRequestPath();
    if (feature.endsWith(".sjs")) {
      feature = feature.substring(0, feature.length - 4);
    }
    if (feature.startsWith("/app/endpoints/")) {
      feature = feature.substring(15);
    }
    feature = feature.replace("/", ".");
    feature = feature.replace("-", "_");
    feature = "opsdir.v1." + reqMethod.toLowerCase() + "." + feature;
    let fid = xdmp.featureMetricRegister(feature);
    xdmp.featureMetricIncrement(fid);

    var i = new t();
    switch (reqMethod) {
    case 'GET':
      return Endpoint.handleErrors(() => i.GET());
    case 'PUT':
      return Endpoint.handleErrors(() => i.PUT());
    case 'POST':
      return Endpoint.handleErrors(() => i.POST());
    case 'DELETE':
      return Endpoint.handleErrors(() => i.DELETE());
    default:
      return Endpoint.handleErrors(() => i.MISSING(reqMethod));
    }
  }

  // fields is an array of objects mapping property names to column names.
  //        [ { "property": "prop1", "header": "Header 1" },
  //          { "property": "prop2": "header": "Header 2" } ... ]
  // items is an array of objects; all of the properties identified
  //       in the fields object will be output as CSV.
  //
  download(filename, fields, items) {
    let acc = [];
    for (var index = 0; index < fields.length; index++) {
      let field = fields[index];
      if (field.label) {
        acc.push(field.label);
      } else if(field.header) {
        acc.push(field.header);
      } else {
        acc.push(field.name);
      }
    }
    let header = acc.map(field => `"${field.replace('"', '""')}"`).join(',') + '\r\n';

    let rows = items.map(item => fields.reduce((acc, field) => {
      let fieldInItem = item[field.name];
      if (fieldInItem !== undefined && fieldInItem !== null) {
        let value = fieldInItem;

        if (Array.isArray(fieldInItem)) {
          value = fieldInItem.join(", ");
        } else if (typeof fieldInItem == "object") {
          value="";
          for (var prop in fieldInItem) {
            if (fieldInItem.hasOwnProperty(prop)) {
              value += prop + "[" + fieldInItem[prop]+ "] ";
            }
          }
        }
        let fieldStr = value.replace ? `"${value.replace('"', '""')}"` : value;
        if (field.type === "projection") {
          fieldStr = '"'+fieldInItem+'"';
        }
        acc.push(fieldStr);
      } else {
        acc.push('""');
      }
      return acc;
    }, [])).join('\r\n') + '\r\n';

    xdmp.addResponseHeader('Content-Disposition',
                           'attachment; filename=' + filename
                           + '-' + fn.currentDateTime() + '.csv;');

      return header + rows;
  }
}
exports.Endpoint = Endpoint;
