"use strict";

/* Copyright 2002-2018 MarkLogic Corporation. All Rights Reserved. */

/* Markdown

   # /v1/endpoints/licenses.sjs

   This endpoint returns license information. It returns the JSON
   payloads for licenses that match the query.

   The following parameters can be used to tailor the search:

   * `clusterId`: to limit the results by cluster ID (one or a list)
   * `hostId`: to limit the results by host ID (one or a list)
   * `clusterName`: to limit the results by cluster name (one or a list)
   * `hostName`: to limit the results by host name (one or a list)
   * `aggregate`: specifies aggregation; must be `edition` or `licenseKey`

   As with the other endpoints, these parameters can be passed by POSTing
   a payload to the endpoint.

*/

const search = require('../lib/search');
const parseParams = require('../lib/parseParams');
const endpoint = require("../endpoint");
const jslib = require('/common/jslib.xqy');

let baseCriteria = [
  cts.collectionQuery(['http://marklogic.com/opsdirector/v1/stats']),
  cts.jsonPropertyValueQuery('opsdirResourceType', 'host-status')
];

let primaryKey = 'opsdirHostId';

// These are the fields we want returned
//
// Note this is all a bit odd because of the underlying architecture
// of the search library. We have to say that 'statusProperties' is
// returned, or we don't get it back. But in fact, we filter it
// further so this doesn't accurately reflect what's returned.
//
// Also note that filtering by 'edition' is handled separately below
// rather than adding another index because this isn't likely to be
// an endpoint where high performance is a requirement.
let fields = [
  { name: 'opsdirId', label: 'Id' },
  { name: 'ts', type: 'date', label: 'Timestamp' },
  { name: 'opsdirDtstart', type: 'date', label: 'Date Start' },
  { name: 'opsdirDtend', type: 'date', label: 'Date End' },
  { name: 'opsdirHostId', label: 'Host ID' },
  { name: 'opsdirClusterId', label: 'Cluster ID' },
  { name: 'opsdirGroupId', label: 'Group ID' },
  { name: 'opsdirHostName', label: 'Host Name' },
  { name: 'opsdirClusterName', label: 'Cluster Name' },
  { name: 'opsdirGroupName', label: 'Group Name' },
  { name: 'statusProperties', label: 'Status properties' }
];

// These are the fields to output in CVS, after transforming search results
let licenseFields = [
  { name: 'opsdirId', label: 'Id' },
  { name: 'ts', type: 'date', label: 'Timestamp' },
  { name: 'opsdirDtstart', type: 'date', label: 'Date Start' },
  { name: 'opsdirDtend', type: 'date', label: 'Date End' },
  { name: 'opsdirHostId', label: 'Host ID' },
  { name: 'opsdirClusterId', label: 'Cluster ID' },
  { name: 'opsdirHostName', label: 'Host Name' },
  { name: 'opsdirClusterName', label: 'Cluster Name' },
  { name: 'environment', label: 'Environment' },
  { name: 'licensee', label: 'Licensee' },
  { name: 'licenseKey', label: 'License Key' },
  { name: 'licenseKeyCpus', label: 'License CPUs' },
  { name: 'licenseKeyOption', label: 'License Options' },
  { name: 'licenseKeySize', label: 'License Key Size' },
  { name: 'licenseKeyCores', label: 'License Cores' },
  { name: 'licenseKeyExpires', label: 'License Expires' },
  { name: 'edition', label: 'License Edition' }
];

let paramDefs = {
  defaultSortField: 'opsdirDtstart',
  defaultSortOrder: 'descending',
  wordField: "*",
  fields: [
    {
      name: 'tsFrom',
      field: 'ts',
      dir: '>=',
      type: 'jsonPropertyRangeQuery'
    }, {
      name: 'tsTo',
      field: 'ts',
      dir: '<=',
      type: 'jsonPropertyRangeQuery'
    }, {
      name: 'dtstartFrom',
      field: 'opsdirDtstart',
      dir: '>=',
      type: 'jsonPropertyRangeQuery'
    }, {
      name: 'dtstartTo',
      field: 'opsdirDtstart',
      dir: '<=',
      type: 'jsonPropertyRangeQuery'
    }, {
      name: 'id',
      field: 'resourceId',
      type: 'jsonPropertyValueQuery'
    }, {
      name: 'clusterId',
      field: 'opsdirClusterId',
      type: 'jsonPropertyValueQuery'
    }, {
      name: 'hostId',
      field: 'opsdirHostId',
      type: 'jsonPropertyValueQuery'
    }, {
      name: 'clusterName',
      field: 'opsdirClusterName',
      type: 'jsonPropertyValueQuery'
    }, {
      name: 'hostName',
      field: 'opsdirHostName',
      type: 'jsonPropertyValueQuery'
    }
  ]
};

let projections = [];

const DEFAULT_SORT_FIELD = 'ts';

//List of fields for which indexing is not done but sorting on them make sense to end user
const NON_INDEXED_FIELDS = [
  'licensee',
  'environment',
  'licenseKeyCpus',
  'licenseKeyCores',
  'licenseKeySize',
  'cpus',
  'cores',
  'coreThreads',
  'edition',
  'platform',
  'architecture',
  'cpus',
  'licenseKeyCapacity',
  'licenseKeyExpires'
];

class EP extends endpoint.Endpoint {
  POST() {
    return this.query(this.getRequestBody());
  }

  GET() {
    return this.query(this.getRequestParams());
  }

  query(body) {
    xdmp.securityAssert("http://marklogic.com/xdmp/privileges/opsdir-license-admin",
                        "execute");

    let parameters = parseParams.parseParams(fields, paramDefs, body);
    let sortField = null;
    let index = NON_INDEXED_FIELDS.indexOf(body.sortField);
    if(index >= 0) {
      sortField = paramDefs.defaultSortField;
    } else {
      sortField = parameters.sortField;
    }

    let results = search.search({
      criteria: [...parameters.criteria, ...baseCriteria],
      // don't change slice values because pagination is done programmatically
      // in code below.
      slice: [0, 1000000],
      primaryKey: primaryKey,
      sort: [[sortField, parameters.direction]],
      fields: fields,
      projections: projections
    });

    results = this.licenseFilter(body, results);

    let fieldDecl = [];
    let filename = "";
    if (body.aggregate === "edition") {
      results = this.aggregateEdition(results);
      if (body.format === "csv") {
        filename = "opsdir-license-editions";
        fieldDecl = [{"name": "edition", "header": "License Edition"},
                     {"name": "licenseKeyCores", "header": "Licensed Cores"},
                     {"name": "cores", "header": "Used Cores"},
                     {"name": "platform", "header": "Platform"},
                     {"name": "environment", "header": "Environment"}];
      }
    } else if (body.aggregate === "licenseKey") {
      results = this.aggregateLicenseKey(results);
      if (body.format === "csv") {
        filename = "opsdir-license-keys";
        fieldDecl = [{"name": "licensee", "header": "Licensee"},
                     {"name": "hosts", "header": "Hosts"},
                     {"name": "clusters", "header": "Clusters"},
                     {"name": "environment", "header": "Environment"},
                     {"name": "groups", "header": "Groups"},
                     {"name": "architecture", "header": "Architecture"},
                     {"name": "cpus", "header": "CPU"},
                     {"name": "cores", "header": "Cores"},
                     {"name": "coreThreads", "header": "Threads"},
                     {"name": "licenseKeyCores", "header": "Licensed Cores"},
                     {"name": "licenseKeyOption", "header": "Options"},
                     {"name": "licenseKeyExpires", "header": "License Expires"},
                     {"name": "licenseKey", "header": "License Key"}];
      }
    } else if (body.aggregate !== undefined) {
      let msg = "Bad request: invalid aggregate: " + body.aggregate;
      throw new endpoint.EndpointError(400, msg);
    } else {
      if (body.format === "csv") {
        filename = "opsdir-licenses";
        fieldDecl = [{"name": "opsdirHostName", "header": "Host"},
                     {"name": "opsdirClusterName", "header": "Cluster"},
                     {"name": "environment", "header": "Environment"},
                     {"name": "opsdirGroupName", "header": "Group"},
                     {"name": "architecture", "header": "Architecture"},
                     {"name": "cpus", "header": "CPU"},
                     {"name": "cores", "header": "Cores"},
                     {"name": "coreThreads", "header": "Threads"},
                     {"name": "licenseKeyCores", "header": "Licensed Cores"},
                     {"name": "licenseKeyOption", "header": "Options"},
                     {"name": "licenseKeyExpires", "header": "Expiration"},
                     {"name": "licenseKey", "header": "License Key"},
                     {"name": "licensee", "header": "Licensee"}];
      }
    }

    //apply sorting on data which is filtered already
    if(index >= 0) {
      sortNonIndexed(body.sortField, body.direction, results.items);
    }
    if (body.format === undefined || body.format === "json") {
      //apply pagination on sorted data.
      let start = parameters.slice[0];
      let length = parameters.slice[1];
      results.items = results.items.slice(start, start + length);

      return jslib.gzipReply(results);
    } else if (body.format === "csv") {
      return this.download(filename, fieldDecl, results.items);
    } else {
      let msg = "Bad request: invalid format: " + body.format;
      throw new endpoint.EndpointError(400, msg);
    }
  }

  licenseFilter(body, results) {
    let licenseKeys = ["licenseKeySize", "licenseKeyCpus", "licenseKeyCores",
                       "licensee", "licenseKeyExpires", "licenseKey", "edition",
                       "environment", "cpus", "cores", "architecture", "coreThreads",
                       "platform"];
    let filtered = [];
    for (var index = 0; index < results.items.length; index++) {
      let item = results.items[index];
      let detail = undefined;

      if (item.statusProperties && item.statusProperties.statusDetail) {
        detail = item.statusProperties.statusDetail;
        delete item.statusProperties;
      }

      if (detail && detail.licenseKeyOption) {
        let options = detail.licenseKeyOption;
        item["licenseKeyOption"] = [];
        for (var pos = 0; pos < options.length; pos++) {
          item["licenseKeyOption"].push(options[pos].value);
        }
      }

      for (let pos = 0; pos < licenseKeys.length; pos++) {
        let name = licenseKeys[pos];
        if (detail && detail[name]) {
          if (detail[name] instanceof Object && ("value" in detail[name])) {
            item[name] = detail[name].value;
          } else {
            item[name] = detail[name];
          }
        }
      }

      let match = (body.edition === undefined) && (item.licenseKey !== undefined);
      if (body.edition !== undefined) {
        if (Array.isArray(body.edition)) {
          for (let index = 0; index < body.edition.length && !match; index++) {
            match = body.edition[index] === item.edition;
          }
        } else {
          match = (body.edition === item.edition);
        }
      }
      if (match) {
        filtered.push(item);
      }
    }

    results.items = filtered;
    results.estimate = filtered.length;
    return results;
  }

  aggregateEdition(licenses) {
    let temp = { };
    let results = { items: [], estimate: 0 };

    for (var index = 0; index < licenses.items.length; index++) {
      let license = licenses.items[index];
      if (temp[license.edition] === undefined) {
        temp[license.edition] = { };
        temp[license.edition].cores = license.cores;
        temp[license.edition].licenseKeyCores = license.licenseKeyCores;
        temp[license.edition].platform = { };
        temp[license.edition].platform[license.platform] = 1;
      } else {
        temp[license.edition].cores += license.cores;
        temp[license.edition].licenseKeyCores += license.licenseKeyCores;
        temp[license.edition].platform[license.platform] = 1;
      }
    }

    for (let key in temp) {
      let result = { };
      result.edition = key;
      result.licenseKeyCores = temp[key].licenseKeyCores;
      result.cores = temp[key].cores;
      result.platform = [];
      for (let plat in temp[key].platform) {
        result.platform.push(plat);
      }
      result.environment =  getEnvCountByLicenseEdition(licenses, result.edition);
      results.items.push(result);
      results.estimate = results.estimate + 1;
    }
    return results;
  }

  aggregateLicenseKey(licenses) {
    let temp = { };
    let results = { items: [], estimate: 0 };

    for (var index = 0; index < licenses.items.length; index++) {
      let license = licenses.items[index];
      let key = license.licenseKey || "unlicensed";
      if (temp[key] === undefined) {
        temp[key] = { };
        temp[key].edition = license.edition;
        temp[key].hosts = [license.opsdirHostName];
        temp[key].clusters = [license.opsdirClusterName];
        temp[key].groups = [license.opsdirGroupName];
        temp[key].architecture = [license.architecture];
        temp[key].cpus = license.cpus;
        temp[key].cores = license.cores;
        temp[key].coreThreads = license.coreThreads;
        temp[key].licensee = license.licensee;
        temp[key].licenseKeyCpus = license.licenseKeyCpus;
        temp[key].licenseKeyCores = license.licenseKeyCores;
        temp[key].licenseKeySize = license.licenseKeySize;
        temp[key].licenseKeyOption = license.licenseKeyOption;
        temp[key].licenseKeyExpires = license.licenseKeyExpires;
      } else {
        temp[key].hosts.push(license.opsdirHostName);
        if (temp[key].clusters.indexOf(license.opsdirClusterName) < 0) {
          temp[key].clusters.push(license.opsdirClusterName);
        }
        if (temp[key].groups.indexOf(license.opsdirGroupName) < 0) {
          temp[key].groups.push(license.opsdirGroupName);
        }
        if (temp[key].architecture.indexOf(license.architecture) < 0) {
          temp[key].architecture.push(license.architecture);
        }
        temp[key].cpus += license.cpus;
        temp[key].cores += license.cores;
      }
    }

    for (let key in temp) {
      let result = temp[key];
      result.licenseKey = key;
      result.environment = getEnvCountByLicenseKey(licenses, result.licenseKey);
      results.items.push(result);
      results.estimate = results.estimate + 1;
    }

    return results;
  }
}

/**
 * Function to sort data based in fields for which support is not present in search API
 * In other words sort data based on fields which are not indexed in database.
 * @param {any} sortField - field name by which data needs to be sorted
 * @param {any} direction - order in which data needs to be sorted.
 * @param {any} data - actual array of data objects
 * @returns Array of data objects sorted on fields given by sortField and in order described by direction.
 */
function sortNonIndexed(sortField, direction, data) {
  //validate if data is an instance of array
  if(data !== undefined && data !== null && Array.isArray(data)){
    return data.sort(compareFunction(sortField, direction));
  } else {
    return data;
  }
}

/**
 * Function that returns a comparator function to be used in Array.sort method
 * @param {any} sortField - field name by which data needs to be sorted
 * @param {any} direction - order in which data needs to be sorted.
 *                          valid values are "asc" or "desc", default is "asc"
 * @returns {function} sort function to be used as compareFunction in Array.sort.
 */
function compareFunction(sortField, direction) {
    let multiplier = (direction === 'desc' || direction === 'descending') ? -1 : 1;
    return (a,b) => {
        // check whether sortfield is actually a property of objects passed in.
        // some properties are not relevant depending what value is set in aggregate parameter.
        if(sortField in a && sortField in b) {
            let val1 = a[sortField], val2 = b[sortField];
            if(typeof a[sortField] == 'string') {
                val1 = val1.toLowerCase();
                val2 = val2.toLowerCase();
            }
            return multiplier * ((val1 >= val2) ? 1 : -1);
        } else {
            let msg = "Bad request: invalid sort field: " + sortField
            throw new endpoint.EndpointError(400, msg)
        }
  };
}

function getEnvCountByLicenseKey(licenses, licenseKey) {
  let environment = {};
  licenses.items.filter(license => license.licenseKey === licenseKey)
    .forEach(item => {
      environment[item.environment] = environment[item.environment] ? environment[item.environment] + 1: 1;
    });
  return environment;
}


function getEnvCountByLicenseEdition(licenses, edition) {
  let environment = {};
  licenses.items.filter(license => license.edition === edition)
    .forEach(item => {
      environment[item.environment] = environment[item.environment] ? environment[item.environment] + 1: 1;
    });
  return environment;
}

endpoint.Endpoint.respond(EP);
