"use strict";

/* Copyright 2002-2018 MarkLogic Corporation. All Rights Reserved. */

const util_1 = require('./util');
var jsearch = require('/MarkLogic/jsearch.sjs');
const resourceUtil = require('./resourceUtil');
const opsdir = require('/common/opsdir.xqy');
let rColl = 'http://marklogic.com/opsdirector/v1/resource';
function recurseResource(map, resource, todo, onlyMissing = false) {
    if (!resource.resourceId) {
        return map;
    }
    // don't do the todo if we are set to only Missing and
    if (!(onlyMissing && map[resource.resourceId])) {
        map = todo(map, resource);
    }
    if (resource.children) {
        for (let resourceType of Object.keys(resource.children)) {
            let resourcesOfType = resource.children[resourceType];
            for (let child of resourcesOfType) {
                map = recurseResource(map, child, todo, onlyMissing);
            }
        }
    }
    if (resource.associations) {
        for (let resourceType of Object.keys(resource.associations)) {
            let resourcesOfType = resource.associations[resourceType];
            for (let child of resourcesOfType) {
                map = recurseResource(map, child, todo, onlyMissing);
            }
        }
    }
    return map;
}
function recurseResources(map, todo, onlyMissing = false) {
    return Object.keys(map).reduce((map, resourceId) => {
        return recurseResource(map, map[resourceId], todo, onlyMissing);
    }, map);
}
function resourcesToResource(resources) {
    return resources.substr(0, resources.length - 1);
}
function resourceResultToMapItem(resourceResult) {
    let doc = resourceResult.document;
    let id = doc.opsdirId.toString();
    let errorResponse = doc.errorResponse;
    if (errorResponse) {
        let err = 'Found configuration document with an error condition:\n\n' + JSON.stringify(doc);
        throw (err);
    }
    let newItem = {
        resourceType: resourcesToResource(doc.opsdirResourceType.toString()),
        resourceName: doc.opsdirName.toString(),
        ts: doc.ts.toString(),
        resourceId: id,
        clusterId: doc.opsdirClusterId.toString(),
        ancestors: []
    };
    switch (newItem.resourceType) {
        case 'cluster':
            break; // empty array is fine
        // this reducer is used on the way out, so this is where we are
        // straightening out the group vs. groupName vs. id issue
        case 'server':
            newItem.altParentId = doc.groupName.toString();
            // we need to come back to this,
            newItem.ancestors = ['altId', doc.opsdirClusterId.toString()];
            break;
        // the rest are all children of a cluster
        case 'group':
            newItem.ancestors = [doc.opsdirClusterId.toString()];
            break;
        case 'host':
            newItem.ancestors = [doc.opsdirClusterId.toString()];
            newItem.opsdirGroupName = doc.opsdirGroupName.toString();
            break;
        case 'database':
            newItem.ancestors = [doc.opsdirClusterId.toString()];
            break;
        case 'forest':
            newItem.ancestors = [doc.opsdirClusterId.toString()];
            break;
        default:
            throw new Error(`Unexpected resource type ("${newItem.resourceType}") on resource ` + newItem.resourceId);
    }
    return newItem;
}
function reduceToTreeMap(params = {}) {
    return (acc, item) => {
        let newItem;
        try {
            newItem = resourceResultToMapItem(item);
        }
        catch (err) {
            util_1.LOG('error', 'Internal Server Error', err);
        }
        if (newItem) {
            acc[newItem.resourceId] = acc[newItem.resourceId] ?
                Object.assign({}, acc[newItem.resourceId], newItem) :
                newItem;
        }
        return acc;
    };
}
function findGroupForServer(resource, items) {
    return items
        .filter(potentialGroup => potentialGroup.resourceType === 'group' &&
        potentialGroup.resourceName === resource.altParentId &&
        potentialGroup.ancestors[0] === resource.ancestors[1])[0];
}
function fixServerGroupRelationship(map, items, resource) {
    if (resource.resourceType === 'server' && resource.ancestors[0] === 'altId') {
        let group = findGroupForServer(resource, items);
        if (group) {
            resource.ancestors[0] = group.resourceId;
            delete resource.altParentId;
            return true;
        }
        else {
            return false;
        }
    }
    else {
        return true;
    }
}
function addMissingAncestors(map, missingAncestors, resource) {
    return Object.assign({}, missingAncestors, resource.ancestors.reduce((acc, ancestorId) => {
        if (ancestorId !== 'altId') {
            if (!map[ancestorId]) {
                acc[ancestorId] = true;
            }
        }
        return acc;
    }, {}));
}
function getMap(...extraCriteria) {
  let map = jsearch.documents()
      .where(cts.andQuery([
        resourceUtil.configDocs(),
        ...extraCriteria,
      ]))
      .slice(0, 50000)
      .reduce((acc, item) => {
        // TODO: get what we need into indexes and switch to tuples
        let newItem = resourceResultToMapItem(item);
        if (newItem) {
          acc[newItem.resourceId] = acc[newItem.resourceId] ?
            Object.assign({}, acc[newItem.resourceId], newItem) :
          newItem;
        }
        return acc;
      }, {})
      .result().results || {};

  // Now do a little patching; put opsdirGroupId in host entries.
  // Make groups an ancestor of hosts. (This has to be a patch because
  // the resources in the database don't have opsdirGroupId everywhere
  // that they would be useful. Because they aren't provided by the
  // Management API.)
  let groups = {};
  for (let id in map) {
    if (map[id].resourceType === "group") {
      groups[id] = map[id];
    }
  }
  for (let id in map) {
    if (map[id].resourceType === "host") {
      let groupId = undefined;
      for (let gid in groups) {
        if (groups[gid].resourceName === map[id].opsdirGroupName
            && groups[gid].clusterId === map[id].clusterId) {
          groupId = gid;
        }
      }
      if (groupId) {
        map[id].opsdirGroupId = groupId;
        // It feels like appending the groupId to the host ancestors
        // should make the hosts children of the group, but it
        // doesn't. And despite the fact that several other resources
        // have multiple ancestors, they appear only to get one
        // parent. I'm giving up.
        // let ancestors = map[id].ancestors || [];
        // ancestors.push(groupId);
        // map[id].ancestors = ancestors;
      }
    }
  }

  return map;
}
function resourceTree(params) {
    /**
     * Goals:
     * 1. fix up altId situations
     * 2. ensure we have ancestors for all resources in the set, up through (enterprise)
     *
     * We are not going after children or associations or attributes here, that stage is over
     * We should be able to do this in one pass (plus a return to the data for the ancestor extraction)
     */
    var map = getMap(params.constraints);
    let items = Object.keys(map).reduce((arr, key) => { arr.push(map[key]); return arr; }, []);
    let missingGroups = [];
    var missingAncestors;
    map = recurseResources(map, (acc, resource) => {
        missingAncestors = addMissingAncestors(map, missingAncestors, resource);
        if (!fixServerGroupRelationship(map, items, resource)) {
            // tried and failed... missing group
            let clusterId = resource.ancestors[1];
            missingGroups.push({ groupName: resource.altParentId, clusterId: clusterId });
            // for good measure, ensure we also fetch the cluster foh the group
            missingAncestors[clusterId] = true;
        }
        acc[resource.resourceId] = resource;
        return acc;
    }, false);
    missingAncestors = missingAncestors || {};
    let missingSpecs = missingGroups.reduce((acc, item) => {
        let groupConstraint = cts.andQuery([
            cts.jsonPropertyValueQuery('opsdirResourceType', 'groups'),
            cts.jsonPropertyValueQuery('opsdirName', item.groupName),
            cts.jsonPropertyValueQuery('opsdirClusterId', item.clusterId),
        ]);
        acc.push(groupConstraint);
        return acc;
    }, []);
    if (Object.keys(missingAncestors).length) {
        missingSpecs.push(cts.jsonPropertyValueQuery('opsdirId', Object.keys(missingAncestors)));
    }
    if (missingSpecs.length) {
        let newDocs = Object.assign({}, map, getMap(cts.andQuery([resourceUtil.configDocs(), cts.orQuery(missingSpecs)])));
        let newItems = Object.keys(newDocs).reduce((acc, key) => { acc.push(newDocs[key]); return acc; }, []);
        Object.keys(newDocs).map(resourceId => {
            fixServerGroupRelationship(newDocs, newItems, newDocs[resourceId]);
        });
        return newDocs;
    }
    else {
        return map;
    }
}
exports.resourceTree = resourceTree;
function trimTree(treeMap) {
    // we don't need ancestors, we need children
    // we need small payload
    let newMap = recurseResources(treeMap, (acc, resource) => {
        let parentId = resource.ancestors && resource.ancestors.shift();
        let parent = parentId ? treeMap[parentId] : null;
        if (parent) {
            let kidsAddress = resource.resourceType + 's';
            let chillun = parent.children || (parent.children = {});
            let kids = parent.children[kidsAddress] || [];
            resource.parentId = parentId;
            kids.push(resource.resourceId);
            chillun[kidsAddress] = kids;
        }
        delete resource.ancestors;
        acc[resource.resourceId] = resource;
        return acc;
    }, false);
    newMap = recurseResources(treeMap, (acc, resource) => {
        let newResource=null;
        if (resource.resourceType === "database"){
         let db = opsdir.managedResource(resource.clusterId, "databases", resource.resourceId);
         let forests = [];
         let forestNames= db.forest;
         if(forestNames){
            for (var i = 0; i < forestNames.length; i++) {
                let forestId = opsdir.managedResourceId(resource.clusterId,"forests",forestNames[i]);
                if(forestId){
                    forests.push(forestId);
                }
            }
         }
         newResource = {
            rid: resource.resourceId,
            rn: resource.resourceName,
            rt: resource.resourceType,
            pid: resource.parentId,
            clusterId: resource.clusterId,
            c: resource.children,
            assn: resource.associations,
            a: resource.attributes,
            ts: resource.ts,
            forest: forests,
            opsdirGroupName: resource.opsdirGroupName,
            opsdirGroupId: resource.opsdirGroupId
         };
        }else{
         newResource = {
            rid: resource.resourceId,
            rn: resource.resourceName,
            rt: resource.resourceType,
            pid: resource.parentId,
            clusterId: resource.clusterId,
            c: resource.children,
            assn: resource.associations,
            a: resource.attributes,
            ts: resource.ts,
            opsdirGroupName: resource.opsdirGroupName,
            opsdirGroupId: resource.opsdirGroupId
         };
        }
        acc[resource.resourceId] = newResource;
        return acc;
    }, false);
    return newMap;
}
exports.trimTree = trimTree;
