"use strict";

/* Copyright 2002-2018 MarkLogic Corporation. All Rights Reserved. */

// See https://wiki.marklogic.com/x/nojqAw

const endpoint = require("../lib/endpoint");
const search   = require("../lib/search");
const jslib    = require('/common/jslib.xqy');
const sd       = require("../lib/systemDatabases.sjs");
const rg       = require('/common/resource-group.xqy');
const rs       = require('/common/rscopes.xqy');
const opsdir   = require("/common/opsdir.xqy");

const ALERTS_COLLECTION = "http://marklogic.com/opsdirector/v1/alerts";

const fieldDefs = {
  "filter": "array",
  "query": "array",
  "text": "string",
  "start":  { type: "number", default: 0 },
  "length": { type: "number", default: 50 },
  "sort": { type: "string", default: "ts" },
  "order": { type: ["ascending", "descending"], default: "ascending" },
  "format": { type: ["json","csv"], default: "json" },
  "headers": "array",
  "filename": "string",

  // These are really just shortcuts for queries; but for ease of
  // transition, I'm supporting them.
  "state": "array",
  "status": "string",
  "acknowledged": { type: ["yes", "no"] },
  "type": "array",
  "tsFrom": "string",
  "tsTo": "string",
  "resourceId": "array",

  // These are special
  "resourceGroupId": "array",
  "excludeAncillary": "boolean"
};

class EP extends endpoint.Endpoint {
  POST() {
    return this.query(this.getRequestBody());
  }

  GET() {
    return this.query(this.getRequestParams());
  }

  DELETE() {
    throw this.clientError("Method Not Allowed", 405);
  }

  query(body) {
    let _ = fn.trace({"method": xdmp.getRequestMethod(),
                      "uri": xdmp.getOriginalUrl(),
                      "body": body},
                     "opsdir-app-endpoint");

    body = this.validShape(body, fieldDefs);

    if (!body.query) {
      body.query = [];
    }

    let vacuous = false;
    for (let name of ["state","status","acknowledged","resourceId"]) {
      if (name in body) {
        let obj = {};
        obj[name] = body[name];
        body.query.push(obj);

        // Special cases that should return no results
        vacuous = vacuous
          || ((name === "state" || name === "resourceId")
              && body[name].length === 0);

        delete body[name];
      }
    }
    if (body.type) {
      body.query.push({"resourceType": body.type});
      delete body["type"];
    }
    if (body.tsFrom) {
      body.query.push({"ts": {"op": ">=", "value": body.tsFrom}});
      delete body["tsFrom"];
    }
    if (body.tsTo) {
      body.query.push({"ts": {"op": "<=", "value": body.tsTo}});
      delete body["tsTo"];
    }

    // Now catch the queries that are vacuous
    for (let query of body.query) {
      if ("state" in query) {
        if (Array.isArray(query["state"]) && query["state"].length == 0) {
          vacuous = true;
        }
      }
    }

    let baseCriteria = [ cts.collectionQuery([ALERTS_COLLECTION]) ];

    // Filter by resourceGroupId
    if (body.resourceGroupId){
      let typeArr = [];
      for (let t of opsdir.SERVERRESOURCETYPES) {
        typeArr.push(t);
      }

      for (let id of body.resourceGroupId) {
        let resourceGroup = rg.get(id);
        let resources = [];
        if (resourceGroup) {
          for (var i = 0; i < resourceGroup.resource.length; i++) {
            let id = resourceGroup.resource[i];
            resources.push(id);

            let rsrc = opsdir.managedResource(null, resourceGroup.type, id);
            if (rsrc) {
              for (let inh of rs.inheritedResources(rsrc)) {
                let rtype = "";
                if (inh.opsdirResourceType) {
                  rtype = inh.opsdirResourceType.toString();
                }
                if (typeArr.indexOf(rtype) >= 0) {
                  resources.push(inh.opsdirId);
                }
              }
            }
          }
          if (resources) {
            baseCriteria.push(
              cts.jsonPropertyValueQuery('resourceId',resources));}
        }
      }
    }

    // Filter out system databases
    if (body.excludeAncillary && body.excludeAncillary.toString() === "true") {
      baseCriteria.push(
        cts.notQuery(
            cts.andQuery([
                cts.jsonPropertyValueQuery('resourceName', sd.systemDatabases()),
                cts.jsonPropertyValueQuery('resourceType', ['database','forest'])]
            )));
    }

    let results = {
      "types": [ "alerts" ]
    };

    if (vacuous) {
      results["alerts"] = { "estimate": 0, "start": 0, "count": 0, "items": [] };
    } else {
      body.fieldName = "alerts-search";
      results["alerts"] = search.search(body, baseCriteria);
    };

    if (body.format == "csv") {
      let filename = this.csvFilename(body.filename, "alerts");
      return this.download(filename, body.headers, results["alerts"]["items"]);
    } else {
      return jslib.gzipReply(results);
    }
  }
}

endpoint.Endpoint.respond(EP);
