"use strict";

/* Copyright 2002-2018 MarkLogic Corporation. All Rights Reserved. */

// See https://wiki.marklogic.com/x/lYHqAw

const endpoint = require("../lib/endpoint");
const search   = require("../lib/search");
const jslib    = require('/common/jslib.xqy');
const opsdir   = require("/common/opsdir.xqy");

const listFieldDefs = {
  "sort": { type: ["opsdirName", "opsdirId", "hostCount", "ts"], "required": false },
  "order": { type: ["ascending", "descending"], default: "ascending" },
  "text": "string",
  "start":  { type: "number", default: 0 },
  "length": { type: "number", default: 50 },
  "unknown": "boolean",
  "format": { type: ["json","csv"], default: "json" },
  "headers": "array",
  "filename": "string"
};

const discFieldDefs = {
  "clusterIds": "array"
};

class EP extends endpoint.Endpoint {
  POST() {
    let body = this.getRequestBody();
    if (body === Object(body)) {
      return this.disconnect(body);
    } else {
      throw this.clientError("Payload must be an object");
    }
  }

  GET() {
    return this.query(this.getRequestParams());
  }

  query(body) {
    let _ = fn.trace({"method": xdmp.getRequestMethod(),
                      "uri": xdmp.getOriginalUrl(),
                      "body": body},
                     "opsdir-app-endpoint");

    body = this.validShape(body, listFieldDefs);

    // We can't use search here because the 'connected' status comes
    // from a join. But these documents are small and we only need one
    // per cluster, so it should be safe.

    let results = {
      "type": [ "clusters" ],
      "clusters": {
        "estimate": 0,
        "start": 0,
        "count": 0,
        "unknownClusterCount": 0
      }
    };

    let items = [];
    let discCount = 0;
    for (let cid of opsdir.managedResourceIds("clusters")) {
      let rsrc = opsdir.managedResource(cid, "clusters", cid).toObject();
      let disconnected = true;
      let ts = undefined;
      for (let hid of rsrc.hostIds) {
        let doc = cts.doc("/config/host-status/" + cid + "/" + hid + "/snapshot.json");
        doc = doc.root.toObject();

        if (ts === undefined || doc.ts > ts) {
          ts = doc.ts;
        }

        disconnected = disconnected && (doc.status === "error");
      }

      if (disconnected) {
        discCount++;
      }

      if (!body.unknown || disconnected) {
        let item = {
          "opsdirId": cid,
          "opsdirName": rsrc.opsdirName,
          "hostCount": rsrc.hostCount,
          "osVersion": rsrc.osVersion,
          "disconnected": disconnected,
          "ts": ts
        };
        items.push(item);
        results.clusters.estimate++;
      }
    }

    // Implement a stupid search
    if (body.text) {
      results.clusters.estimate = 0;
      let match = [];
      for (let item of items) {
        if (xdmp.quote(item).indexOf(body.text) >= 0) {
          match.push(item);
          results.clusters.estimate++;
        }
      }
      items = match;
    }

    // Implement a stupid sort
    if (body.sort) {
      items = items.sort(compareFunction(body.sort, body.order));
    }

    if (body.start > 0) {
      results.clusters.start = body.start;
      items = items.slice(body.start, items.length);
    }

    if (body.length < items.length) {
      items = items.slice(0, body.length);
    }

    results.clusters.unknownClusterCount = discCount;
    results.clusters.count = items.length;
    results.clusters.items = items;

    if (body.format == "csv") {
      let filename = this.csvFilename(body.filename, "disconnected");
      return this.download(filename, body.headers, results["clusters"]["items"]);
    } else {
      return jslib.gzipReply(results);
    }
  }

  disconnect(body) {
    let _ = fn.trace({"method": xdmp.getRequestMethod(),
                      "uri": xdmp.getOriginalUrl(),
                      "body": body},
                     "opsdir-app-endpoint");

    body = this.validShape(body, discFieldDefs);

    for (let cid of body.clusterIds) {
      opsdir.disconnectCluster(cid);
    }

    let results = {
      "message": "ok",
      "disconnected": body.clusterIds
    };

    return jslib.gzipReply(results);
  }
}

// Return a comparator function for the specified sortField and direction
function compareFunction(sortField, order) {
  let multiplier = (order === 'ascending') ? 1 : -1;
  return (a,b) => {
    let val1 = a[sortField];
    let val2 = b[sortField];
    if (typeof val1 == 'string') {
      val1 = val1.toLowerCase();
      val2 = val2.toLowerCase();
    }
    return multiplier * ((val1 >= val2) ? 1 : -1);
  };
}

endpoint.Endpoint.respond(EP);
