"use strict";

/* Copyright 2002-2018 MarkLogic Corporation. All Rights Reserved. */

// ************************************************************
// NOTE: THIS ENDPOINT IS NOT CALLED FOR /v2/licenses. THE OLD
// /v1/ ENDPOINT REMAINS IN USE FOR OPSDIR 1.x (and 2.0.0 at the moment)
// ************************************************************

const endpoint = require("../lib/endpoint");
const search = require('../lib/search');
const jslib = require('/common/jslib.xqy');

const fieldDefs = {
  "filter": "array",
  "query": "array",
  "text": "string",
  "start":  { type: "number", default: 0 },
  "length": { type: "number", default: 50 },
  "sort": "string",
  "order": { type: ["ascending", "descending"], default: "ascending" },
  "format": { type: ["json","csv"], default: "json" },
  "headers": "array",
  "filename": "string",

  "aggregate": { "type": [ "edition", "licenseKey" ], default: "licenseKey" },

  "clusterId": "array",
  "hostId": "array",
  "clusterName": "array",
  "hostName": "array",
  "edition": "string"
};

// Per funcspec:
// Aggregate by edition:
// Searchable: edition, platform, environment
// Sortable: licensedCores, usedCores, platform, environment
// Aggregate by licenseKey:
// Searchable: host, cluster, group,
// Sortable: host, cluster, environment, group, architecture, cpu, cores, threads,
//           licensedCores, expiration, licenseKey, licensee

let baseQuery = [
  cts.collectionQuery(['http://marklogic.com/opsdirector/v1/resource']),
  cts.jsonPropertyValueQuery('opsdirResourceType', 'host-status')
];

let primaryKey = 'opsdirHostId';

class EP extends endpoint.Endpoint {
  POST() {
    return this.query(this.getRequestBody());
  }

  GET() {
    return this.query(this.getRequestParams());
  }

  query(body) {
    let _ = fn.trace({"method": xdmp.getRequestMethod(),
                      "uri": xdmp.getOriginalUrl(),
                      "body": body},
                     "opsdir-app-endpoint");

    xdmp.securityAssert("http://marklogic.com/xdmp/privileges/opsdir-license-admin",
                        "execute");

    body = this.validShape(body, fieldDefs);

    if (body.hostId) {
      baseQuery.push(cts.jsonPropertyValueQuery("opsdirHostId", body.hostId));
    }
    if (body.hostName) {
      baseQuery.push(cts.jsonPropertyValueQuery("opsdirResourceName", body.hostName));
    }
    if (body.clusterId) {
      baseQuery.push(cts.jsonPropertyValueQuery("opsdirClusterId", body.clusterId));
    }
    if (body.clusterName) {
      baseQuery.push(cts.jsonPropertyValueQuery("opsdirClusterName", body.clusterName));
    }

    let sbody = {};
    sbody.start = 0;
    sbody.length = Number.MAX_SAFE_INTEGER;
    sbody.query = body.query;

    let results = search.search(sbody, baseQuery);
    let licenses = [];
    for (let item of results.items) {
      let obj = {
        "opsdirClusterId": item.opsdirClusterId,
        "opsdirClusterName": item.opsdirClusterName,
        "opsdirHostId": item.opsdirHostId,
        "opsdirHostName": item.opsdirResourceName,
        "opsdirGroupName": item.opsdirGroupName
      };

      if (item.statusProperties && item.statusProperties.statusDetail) {
        let detail = item.statusProperties.statusDetail;
        for (let prop of [ "licenseKeySize", "licenseKeyCpus", "licenseKeyCores",
                           "licenseKeyExpires", "licenseKey", "licenseKeyOption",
                           "licensee", "edition",
                           "environment", "cpus", "cores", "architecture", "coreThreads",
                           "platform" ]) {
          if (prop in detail) {
            if (prop === "licenseKeyOption") {
              // This one is special; an array of objects
              obj[prop] = [];
              for (let opt of detail[prop]) {
                obj[prop].push(opt.value);
              }
            } else {
              if (Object(detail[prop]) === detail[prop]) {
                obj[prop] = detail[prop]["value"];
              } else {
                obj[prop] = detail[prop];
              }
            }
          }
        }
      }

      if (!body.edition || (body.edition === obj.edition)) {
        licenses.push(obj);
      }
    }

    let aggr = [];
    if (body.aggregate === "edition") {
      aggr = this.aggregateEdition(licenses);
    } else {
      aggr = this.aggregateLicenseKey(licenses);
    }

    if (body.text) {
      let searched = [];
      if (body.aggregate === "edition") {
        // Searchable: edition, platform, environment
        for (let lic of aggr) {
          let match = (lic.edition.indexOf(body.text) >= 0);
          for (let plat of lic.platform) {
            match = match || (plat.indexOf(body.text) >= 0);
          }
          for (let env in lic.environment) {
            match = match || (env.indexOf(body.text) >= 0);
          }
          if (match) {
            searched.push(lic);
          }
        }
      } else {
        // Searchable: host, cluster, group,
        for (let lic of aggr) {
          let match = false;
          for (let key of ["hosts", "clusters", "groups"]) {
            for (let val of lic[key]) {
              match = match || (val.indexOf(body.text) >= 0);
            }
          }
          if (match) {
            searched.push(lic);
          }
        }
      }

      aggr = searched;
    }

    if (body.sort) {
      let sortMap = {
        "hosts": "opsdirHostName",
        "clusters": "opsdirClusterName",
        "groups": "opsdirGroupName"
      };
      let sort = body.sort;
      if (sort in sortMap) {
        sort = sortMap[sort];
      }

      aggr = sortNonIndexed(sort, body.order, aggr);
    }

    if (body.filter) {
      let filtered = [];
      for (let lic of aggr) {
        let obj = {};
        for (let prop of body.filter) {
          if (prop in lic) {
            obj[prop] = lic[prop];
          }
        }
        filtered.push(obj);
      }
      aggr = filtered;
    }

    let xresults = {
      "types": [ "licenses" ],
      "licenses": {
        "estimate": aggr.length,
        "start": body.start,
        "count": Math.min(body.length, aggr.length),
        "items": aggr.slice(body.start, body.start + body.length)
      }
    };

    if (body.format == "csv") {
      let filename = this.csvFilename(body.filename, "licenses-" + body.aggregate);
      return this.download(filename, body.headers, xresults.licenses.items);
    } else {
      return jslib.gzipReply(xresults);
    }
  }

  aggregateEdition(licenses) {
    let temp = { };
    let results = [];

    for (var index = 0; index < licenses.length; index++) {
      let license = licenses[index];

      if (temp[license.edition] === undefined) {
        temp[license.edition] = { };
        temp[license.edition].cores = license.cores;
        temp[license.edition].licenseKeyCores = license.licenseKeyCores;
        temp[license.edition].platform = { };
        temp[license.edition].platform[license.platform] = 1;
        temp[license.edition].opsdirClusterId = [ license.opsdirClusterId ];
        temp[license.edition].opsdirClusterName = [ license.opsdirClusterName ];
        temp[license.edition].opsdirHostId = [ license.opsdirHostId ];
        temp[license.edition].opsdirHostName = [ license.opsdirHostName ];
      } else {
        temp[license.edition].cores += license.cores;
        temp[license.edition].licenseKeyCores += license.licenseKeyCores;
        temp[license.edition].platform[license.platform] = 1;

        temp[license.edition].opsdirClusterId.push(license.opsdirClusterId);
        temp[license.edition].opsdirClusterName.push(license.opsdirClusterName);
        temp[license.edition].opsdirHostId.push(license.opsdirHostId);
        temp[license.edition].opsdirHostName.push(license.opsdirHostName);
      }
    }

    for (let key in temp) {
      let result = { };
      result.edition = key;
      result.licenseKeyCores = temp[key].licenseKeyCores;
      result.cores = temp[key].cores;
      result.platform = temp[key].platform;
      result.environment = getEnvCountByLicenseEdition(licenses, result.edition);

      // Use sets to get distinct values
      result.opsdirClusterId = [...new Set(temp[key].opsdirClusterId)];
      result.opsdirClusterName = [...new Set(temp[key].opsdirClusterName)];
      result.opsdirHostId = [...new Set(temp[key].opsdirHostId)];
      result.opsdirHostName = [...new Set(temp[key].opsdirHostName)];

      results.push(result);
    }

    return results;
  }

  aggregateLicenseKey(licenses) {
    let temp = { };
    let results = [];

    for (var index = 0; index < licenses.length; index++) {
      let license = licenses[index];
      let key = license.licenseKey || "unlicensed";
      if (temp[key] === undefined) {
        temp[key] = { };
        temp[key].edition = license.edition;
        temp[key].hosts = [license.opsdirHostName];
        temp[key].clusters = [license.opsdirClusterName];
        temp[key].groups = [license.opsdirGroupName];
        temp[key].architecture = [license.architecture];
        temp[key].cpus = license.cpus;
        temp[key].cores = license.cores;
        temp[key].coreThreads = license.coreThreads;
        temp[key].licensee = license.licensee;
        temp[key].licenseKeyCpus = license.licenseKeyCpus;
        temp[key].licenseKeyCores = license.licenseKeyCores;
        temp[key].licenseKeySize = license.licenseKeySize;
        temp[key].licenseKeyOption = license.licenseKeyOption;
        temp[key].licenseKeyExpires = license.licenseKeyExpires;
        temp[key].opsdirClusterId = [ license.opsdirClusterId ];
        temp[key].opsdirClusterName = [ license.opsdirClusterName ];
        temp[key].opsdirHostId = [ license.opsdirHostId ];
        temp[key].opsdirHostName = [ license.opsdirHostName ];
      } else {
        temp[key].hosts.push(license.opsdirHostName);
        if (temp[key].clusters.indexOf(license.opsdirClusterName) < 0) {
          temp[key].clusters.push(license.opsdirClusterName);
        }
        if (temp[key].groups.indexOf(license.opsdirGroupName) < 0) {
          temp[key].groups.push(license.opsdirGroupName);
        }
        if (temp[key].architecture.indexOf(license.architecture) < 0) {
          temp[key].architecture.push(license.architecture);
        }
        temp[key].cpus += license.cpus;
        temp[key].cores += license.cores;
        temp[key].opsdirClusterId.push(license.opsdirClusterId);
        temp[key].opsdirClusterName.push(license.opsdirClusterName);
        temp[key].opsdirHostId.push(license.opsdirHostId);
        temp[key].opsdirHostName.push(license.opsdirHostName);
      }
    }

    for (let key in temp) {
      let result = temp[key];
      result.licenseKey = key;
      result.environment = getEnvCountByLicenseKey(licenses, result.licenseKey);

      // Use sets to get distinct values
      result.opsdirClusterId = [...new Set(result.opsdirClusterId)];
      result.opsdirClusterName = [...new Set(result.opsdirClusterName)];
      result.opsdirHostId = [...new Set(result.opsdirHostId)];
      result.opsdirHostName = [...new Set(result.opsdirHostName)];

      results.push(result);
    }

    return results;
  }
}

/**
 * The aggregate data isn't conveniently indexed in the database, so
 * we have to sort by hand.
 * @param {any} sortField - field name by which data needs to be sorted
 * @param {any} direction - order in which data needs to be sorted.
 * @param {any} data - actual array of data objects
 * @returns Array of data objects sorted on fields given by sortField
 *          and in order described by direction.
 */
function sortNonIndexed(sortField, direction, data) {
  // validate if data is an instance of array

  if (data !== undefined && data !== null && Array.isArray(data)) {
    return data.sort(compareFunction(sortField, direction));
  } else {
    return data;
  }
}

/**
 * Function that returns a comparator function to be used in Array.sort method
 * @param {any} sortField - field name by which data needs to be sorted
 * @param {any} direction - order in which data needs to be sorted.
 * @returns {function} sort function to be used as compareFunction in Array.sort.
 */
function compareFunction(sortField, direction) {
  let multiplier = (direction === 'descending') ? -1 : 1;
  return (a,b) => {
    // Check whether sortfield is actually a property of objects passed in.
    if(sortField in a && sortField in b) {
      let val1 = a[sortField];
      let val2 = b[sortField];

      // If the sort fields are objects, then sort on the names of
      // the properties. If the sort fields are arrays, sort on the
      // value of the largest or smallest item in the array (per
      // the sort order).

      if (Array.isArray(val1)) {
        val1 = sortStrings(val1, multiplier)[0];
      }

      if (Object(val1) === val1) {
        val1 = Object.keys(val1);
      }

      if (Array.isArray(val2)) {
        val2 = sortStrings(val2, multiplier)[0];
      }

      if (Object(val2) === val2) {
        val2 = Object.keys(val2);
      }

      if(typeof a[sortField] == 'string') {
        val1 = val1.toLowerCase();
        val2 = val2.toLowerCase();
      }

      // console.log("%j: %j <=> %j", multiplier * ((val1 >= val2) ? 1 : -1), val1, val2);

      return multiplier * ((val1 >= val2) ? 1 : -1);
    } else {
      return new endpoint.EndpointError(400, "Bad request: invalid sort field: " + sortField);
    }
  };
}

function sortStrings(data, multiplier) {
  return data.sort(function (a, b) {
    return multiplier * (a.toLowerCase() >= b.toLowerCase() ? 1 : - 1);
  });
}

function getEnvCountByLicenseKey(licenses, licenseKey) {
  let environment = {};
  licenses.filter(license => license.licenseKey === licenseKey)
    .forEach(item => {
      environment[item.environment]
        = environment[item.environment] ? environment[item.environment] + 1: 1;
    });
  return environment;
}


function getEnvCountByLicenseEdition(licenses, edition) {
  let environment = {};
  licenses.filter(license => license.edition === edition)
    .forEach(item => {
      environment[item.environment]
        = environment[item.environment] ? environment[item.environment] + 1: 1;
    });
  return environment;
}

endpoint.Endpoint.respond(EP);
