"use strict";

/* Copyright 2002-2018 MarkLogic Corporation. All Rights Reserved. */

// See https://wiki.marklogic.com/x/oIjqAw

const endpoint = require("../lib/endpoint");
const search   = require("../lib/search");
const jslib    = require('/common/jslib.xqy');
const rg       = require('/common/resource-group.xqy');
const rs       = require('/common/rscopes.xqy');
const opsdir   = require("/common/opsdir.xqy");

const LOGS_COLLECTION = 'http://marklogic.com/opsdirector/v1/logs';

const fieldDefs = {
  "filter": "array",
  "query": "array",
  "start":  { type: "number", default: 0 },
  "length": { type: "number", default: 50 },
  "sort": { type: "string", default: "ts" },
  "text": "string",
  "order": { type: ["ascending", "descending"], default: "ascending" },
  "format": { type: ["json","csv"], default: "json" },
  "headers": "array",
  "filename": "string",

  // These are really just shortcuts for queries; but for ease of
  // transition, I'm supporting them.
  "hostId": "array",
  "clusterId": "array",
  "resourceGroupId": "array",
  "level": "array",
  "tsFrom": "string",
  "tsTo": "string"
};

class EP extends endpoint.Endpoint {
  POST() {
    return this.query(this.getRequestBody());
  }

  GET() {
    return this.query(this.getRequestParams());
  }

  query(body) {
    let _ = fn.trace({"method": xdmp.getRequestMethod(),
                      "uri": xdmp.getOriginalUrl(),
                      "body": body},
                     "opsdir-app-endpoint");

    body = this.validShape(body, fieldDefs);

    if (!body.query) {
      body.query = [];
    }

    let vacuous = false;
    let rgHostIds = [];
    if (body.resourceGroupId) {
      vacuous = true;
      let rgHostMap = {};
      for (let id of body.resourceGroupId) {
        let resourceGroup = rg.get(id).toObject();
        if (resourceGroup.type === "hosts" || resourceGroup.type === "clusters") {
          for (var i = 0; i < resourceGroup.resource.length; i++) {
            let id = resourceGroup.resource[i];
            if (resourceGroup.type === "hosts") {
              rgHostMap[id] = "hosts";
              vacuous = false;
            } else {
              let rsrc = opsdir.managedResource(null, resourceGroup.type, id);
              if (rsrc) {
                for (let inh of rs.inheritedResources(rsrc)) {
                  let rtype = "";
                  if (inh.opsdirResourceType) {
                    rtype = inh.opsdirResourceType.toString();
                  }
                  if (rtype === "hosts") {
                    rgHostMap[inh.opsdirId] = rtype;
                    vacuous = false;
                  }
                }
              }
            }
          }
        }
      }

      for (let id in rgHostMap) {
        rgHostIds.push(id);
      }

      delete body.resourceGroupId;
    }

    if (body.hostId) {
      vacuous = body.hostId.length == 0; // N.B. not vacuous || because RG doesn't matter
      body.query.push({"opsdirHostId": body.hostId});
      delete body["hostId"];
    } else {
      if (rgHostIds.length > 0) {
        body.query.push({"opsdirHostId": rgHostIds});
      }
    }
    if (body.clusterId) {
      vacuous = vacuous || body.clusterId.length == 0;
      body.query.push({"opsdirClusterId": body.clusterId});
      delete body["clusterId"];
    }
    if (body.level) {
      vacuous = vacuous || body.level.length == 0;
      body.query.push({"level": body.level});
      delete body["level"];
    }
    if (body.tsFrom) {
      body.query.push({"ts": {"op": ">=", "value": body.tsFrom}});
      delete body["tsFrom"];
    }
    if (body.tsTo) {
      body.query.push({"ts": {"op": "<=", "value": body.tsTo}});
      delete body["tsTo"];
    }

    let results = {
      "types": [ "logs" ]
    };

    if (vacuous) {
      results["logs"] = { "estimate": 0, "start": 0, "count": 0, "items": [] };
    } else {
      // This terrible hack is for backwards compatibility
      let search_results = search.search(body, cts.collectionQuery([LOGS_COLLECTION]));
      let items = [];
      for (let item of search_results.items) {
        let obj = {};
        for (let key in item) {
          obj[key] = item[key].toString();
        }
        obj["logFile"] = "ErrorLog.txt";
        items.push(obj);
      }

      results["logs"] = {
        "estimate": search_results.estimate,
        "start": search_results.start,
        "count": search_results.count,
        "items": items
      };
    }

    if (body.format == "csv") {
      let filename = this.csvFilename(body.filename, "logs");
      return this.download(filename, body.headers, results["logs"]["items"]);
    } else {
      return jslib.gzipReply(results);
    }
  }
}

endpoint.Endpoint.respond(EP);
