xquery version "1.0-ml";

(: Copyright 2002-2018 MarkLogic Corporation. All Rights Reserved. :)

(: See https://wiki.marklogic.com/x/vAOxB :)

import module namespace manage= "http://marklogic.com/manage"
at "/MarkLogic/manage/manage.xqy";

import module namespace opsdir= "http://marklogic.com/v1/opsdirector"
at "/common/opsdir.xqy";

import module namespace jslib= "http://marklogic.com/v1/opsdirector/jslib"
at "/common/jslib.xqy";

import module namespace rg="http://marklogic.com/v1/opsdirector/resource-groups"
at "/common/resource-group.xqy";

declare namespace meters="http://marklogic.com/manage/meters";
declare namespace zip="xdmp:zip";

declare option xdmp:mapping "false";

declare variable $OneDay := xs:dayTimeDuration("P1D");
declare variable $KINDS := ("host", "database", "forest", "server", "group", "cluster");

declare function local:kind(
  $kind as xs:string
) as xs:string
{
  if ($kind = $KINDS)
  then $kind
  else error((), "OD-BADTYPE", "Unexpected resource type: " || $kind)
};

declare function local:period(
  $period as xs:string
) as xs:string
{
  if ($period = ("raw", "hour", "day"))
  then $period
  else error((), "OD-BADPERIOD", "Unexpected period: " || $period)
};

declare function local:dateTime(
  $dt as xs:string
) as xs:dateTime
{
  if ($dt castable as xs:dateTime)
  then $dt cast as xs:dateTime
  else error((), "OD-BADDATETIME", "Invalid datetime: " || $dt)
};

declare function local:boolean(
  $bool as xs:string
) as xs:boolean
{
  if ($bool = ("true","false"))
  then ($bool = "true")
  else error((), "OD-BADBOOL", "Invalid boolean: " || $bool)
};

declare function local:metrics(
  $metrics as xs:string*
) as xs:string*
{
  $metrics
};

declare function local:format(
  $format as xs:string
) as xs:string
{
  if ($format = ("json","xml","csv","csvdebug"))
  then $format
  else error((), "OD-BADFORMAT", "Invalid format: " || $format)
};

declare function local:ids(
  $ids as xs:string*
) as xs:unsignedLong*
{
  for $id in $ids
  return
    if ($id castable as xs:unsignedLong)
    then $id cast as xs:unsignedLong
    else error((), "OD-BADID", "Invalid ID: " || $id)
};

declare function local:csv-package(
  $format as xs:string,
  $parts as element(zip:part)+,
  $tables as element(table)+,
  $relations as element()*
)
{
  let $manifest := <parts xmlns="xdmp:zip">{ $parts }</parts>
  return
    if (count($tables) = 1)
    then
      let $_ := xdmp:set-response-content-type("text/plain")
      let $_ := if ($format = "csv")
                then xdmp:add-response-header('Content-Disposition',
                        'attachment; filename=' || $manifest/zip:part)
                else ()
      return
        (string($tables[1]))
    else
      if ($format = "csvdebug")
      then
        <csvdebug><zip>{$manifest}{$tables}</zip>{$relations}</csvdebug>
      else
        let $_ := xdmp:set-response-content-type("application/zip")
        return
          xdmp:zip-create($manifest, $tables ! text { string(.) })
};

declare function local:csv-parts(
  $meters as element()*
) as element(zip:parts)
{
  let $meta := $meters[1]
  let $prop := $meters[2]
  let $relations := $meters[3]
  let $_ := $meters//meters:errors//meters:error-type ! xdmp:log("Metrics error: " || .)
  return
    <parts xmlns="xdmp:zip">
      { for $metric in $relations//(meters:summary|meters:detail)/../../*
        let $kind := if ($metric/ancestor::meters:database-metrics)
                     then "database"
                     else if ($metric/ancestor::meters:host-metrics)
                     then "host"
                     else if ($metric/ancestor::meters:server-metrics)
                     then "server"
                     else if ($metric/ancestor::meters:forest-metrics)
                     then "forest"
                     else
                       error((), "OD-BADTYPE", "Cannot determine resource type")
        where local-name($metric) ne "errors"
        return
          <part>{$kind}/{local-name($metric)}.csv</part>
      }
    </parts>
};

declare function local:csv-tables(
  $meters as element()*
) as element(table)*
{
  let $meta := $meters[1]
  let $prop := $meters[2]
  let $relations := $meters[3]
  return
    if ($relations//meters:summary)
    then
      for $metric in $relations//meters:summary/../../*
      where local-name($metric) ne "errors"
      return
        <table>
          <row>"Timestamp","{string($metric/meters:units)}"&#10;</row>
          { for $entry in $metric//meters:summary//meters:entry
            return
              <row>&#x22;{string($entry/meters:dt)}
                   {'&#x22;,'}
                   {string($entry/meters:value)}&#10;</row>
          }
        </table>
    else
      for $kind in $relations/*
      (: Databases have an extra master/??? wrapper :)
      for $metric in $kind//meters:detail/..
      let $dts := distinct-values($metric/meters:detail/meters:data/meters:entry/meters:dt)
      let $headers := ("Timestamp", $metric/meters:detail/meters:name/string())
      return
        <table metric="{local-name($kind)}/{local-name($metric)}">
          <row>
            { concat('"', string-join($headers, '","'), '"') }
            { "&#10;" }
          </row>
          { for $dt in $dts
            let $data := for $detail in $metric/meters:detail
                         let $value := $detail/meters:data/meters:entry[meters:dt = $dt]
                         let $value := $value/meters:value/string()
                         return
                           if (empty($value))
                           then '""'
                           else $value
            return
              <row>&#x22;{string($dt)}&#x22;,{string-join($data, ',')}&#10;</row>
          }
        </table>
};

declare function local:csv(
  $format as xs:string,
  $kind as xs:string,
  $meters as element()*
)
{
  let $manifest  := local:csv-parts($meters)
  let $parts     := local:csv-tables($meters)
  return
    if (empty($manifest/zip:part))
    then
      xdmp:set-response-code(204, "No metrics")
    else
      local:csv-package($format, $manifest/zip:part, $parts, $meters[3])
};

declare function local:get-post-body(
) as object-node() {
  if (contains(xdmp:get-request-header("content-type"), 'json'))
  then
    xdmp:get-request-body("json")/object-node()
  else
    error((), "OD-PAYLOAD",
          "Unexpected payload type: " || xdmp:get-request-header("content-type"))
};

declare function local:create-post-body(
) as object-node() {
  let $map := map:map()
  let $_   := for $name in xdmp:get-request-field-names()
              let $value := xdmp:get-request-field($name)
              return
                map:put($map, $name, tokenize($value, "\s*,\s*"))
  return
    xdmp:to-json($map)/object-node()
};

declare function local:validate(
  $body as object-node()
) as object-node()
{
  if ($body/kind = "multi")
  then
    local:validate-multi($body)
  else
    local:validate-single($body)
};

declare function local:validate-single(
  $body as object-node()
) as object-node()
{
  let $yesterday := string(current-dateTime() - $OneDay)
  let $now       := string(current-dateTime())

  let $map := map:map()
  let $_   := map:put($map, "kind", local:kind($body/kind))
  let $_   := map:put($map, "period", local:period(($body/period, "raw")[1]))
  let $_   := map:put($map, "start", local:dateTime(($body/start, $yesterday)[1]))
  let $_   := map:put($map, "end", local:dateTime(($body/end, $now)[1]))
  let $_   := map:put($map, "summary", local:boolean(($body/summary/string(), "true")[1]))
  let $_   := map:put($map, "detail", local:boolean(($body/detail/string(), "false")[1]))
  let $_   := map:put($map, "metrics", local:metrics($body/metrics))
  let $_   := map:put($map, "format", local:format(($body/format, "json")[1]))
  let $_   := map:put($map, "cluster", array-node { local:ids($body/cluster) } )
  let $_   := map:put($map, "resource", array-node { local:ids($body/resource) } )
  let $_   := map:put($map, "host", array-node { local:ids($body/host) } )
  let $_   := map:put($map, "resourceGroupId",
                            array-node { local:ids($body/resourceGroupId) } )
  return
    xdmp:to-json($map)/object-node()
};

declare function local:validate-multi(
  $body as object-node()
) as object-node()
{
  let $yesterday := string(current-dateTime() - $OneDay)
  let $now       := string(current-dateTime())

  let $period   := local:period(($body/period, "raw")[1])
  let $start    := string(local:dateTime(($body/start, $yesterday)[1]))
  let $end      := string(local:dateTime(($body/end, $now)[1]))
  let $summary  := string(local:boolean(($body/summary/string(), "true")[1]))
  let $detail   := string(local:boolean(($body/detail/string(), "false")[1]))
  let $format   := string(local:format(($body/format, "json")[1]))
  let $cluster  := array-node { local:ids($body/cluster) }
  let $resource := array-node { local:ids($body/resource) }
  let $host     := array-node { local:ids($body/host) }
  let $rgroup   := array-node { local:ids($body/resourceGroupId) }

  let $map     := map:map()
  let $_       := map:put($map, "kind", "multi")
  let $_       := map:put($map, "format", $format)
  let $metrics := map:map()

  let $_   := for $kind in $KINDS
              let $obj := $body/metrics/*[local-name(.) = $kind]
              where $obj
              return
                let $map := map:map()
                let $_   := map:put($map, "kind", local:kind($kind))
                let $_   := map:put($map, "period", local:period(($obj/period, $period)[1]))
                let $_   := map:put($map, "start", local:dateTime(($obj/start, $start)[1]))
                let $_   := map:put($map, "end", local:dateTime(($obj/end, $end)[1]))
                let $_   := map:put($map, "summary",
                                    local:boolean(($obj/summary/string(), $summary)[1]))
                let $_   := map:put($map, "detail",
                                    local:boolean(($obj/detail/string(), $detail)[1]))
                let $_   := map:put($map, "metrics", local:metrics($obj/metrics))
                let $_   := map:put($map, "format", $format)
                let $_   := if ($obj/cluster)
                            then
                              map:put($map, "cluster", array-node { local:ids($obj/cluster) } )
                            else
                              map:put($map, "cluster", $cluster)
                let $_   := if ($obj/resource)
                            then
                              map:put($map, "resource",
                                            array-node { local:ids($obj/resource) } )
                            else
                              map:put($map, "resource", $resource)
                let $_   := if ($obj/host)
                            then
                              map:put($map, "host", array-node { local:ids($obj/host) } )
                            else
                              map:put($map, "host", $host)
                let $_   := if ($obj/resourceGroupId)
                            then
                              map:put($map, "resourceGroupId",
                                            array-node { local:ids($obj/resourceGroupId) } )
                            else
                              map:put($map, "resourceGroupId", $rgroup)
                return
                  map:put($metrics, $kind, $map)
  let $_ := map:put($map, "metrics", $metrics)
  return
    xdmp:to-json($map)/object-node()
};

declare function local:multi-metrics(
  $multibody as object-node()
)
{
  let $database   := "OpsDirector"
  let $collection := "meters"

  let $map   := map:map()
  let $_     := for $kind in $KINDS
                let $body := $multibody/metrics/*[local-name(.) = $kind]
                where $body
                return
                  let $cid        := $body/cluster ! xs:unsignedLong(.)
                  let $rid        := $body/resource ! xs:unsignedLong(.)
                  let $hid        := $body/host ! xs:unsignedLong(.)
                  let $resourceGroupIds := $body/resourceGroupId ! xs:unsignedLong(.)

                  (: retrieve all server ids this user has access to :)
                  let $resource-ids := opsdir:managed-resource-ids($kind || "s")

                  (: if single rid, check it is accessible:)
                  let $rids := if (exists($rid))
                               then $rid[xs:string(.) = $resource-ids]
                               else $resource-ids

                  (: if resourceGroupId supplied, match up resources :)
                  let $rids := if ($resourceGroupIds)
                               then
                                 let $resources := for $rgid in $resourceGroupIds
                                                   return rg:get(xs:string($rgid))/resource
                                 return distinct-values($rids[.=$resources])
                               else $rids
                  return
                    if (empty($rids))
                    then ()
                    else
                      map:put($map, $kind, jslib:resource-metrics(
                          $body/kind,
                          $rids,
                          $body/period,
                          $body/start,
                          $body/end,
                          $body/summary,
                          $body/detail,
                          $body/metrics,
                          "xml",
                          $database,
                          $collection,
                          $cid,
                          $hid)/node())

  (: N.B. The <root> wrappers are necessary because the relative order of
     different documents is unpredictable. Without the wrappers, the parts
     and the tables can wind up in different orders...which is bad. :)

  let $parts := <root>
                  { for $kind in $KINDS
                    where map:contains($map, $kind)
                    return
                      local:csv-parts(map:get($map, $kind))
                  }
                </root>
  let $tabs  := <root>
                  { for $kind in $KINDS
                    where map:contains($map, $kind)
                    return
                      local:csv-tables(map:get($map, $kind))
                  }
                </root>
  let $rels  := <root>
                  { for $kind in $KINDS
                    where map:contains($map, $kind)
                    return
                      map:get($map, $kind)[3]
                  }
                </root>
  return
    if (empty($parts/*/zip:part))
    then
      xdmp:set-response-code(204, "No metrics")
    else
      local:csv-package($multibody/format, $parts/*/zip:part, $tabs/*, $rels)
};

let $body := if (xdmp:get-request-method() = "POST")
             then local:get-post-body()
             else local:create-post-body()
let $body := local:validate($body)

let $_    := if ($body/kind = "multi")
             then
               if ($body/format = "json")
               then error((), "OD-BADMULTI", "Multiple metrics requires CSV output")
               else ()
             else
               ()

return
  if ($body/kind = "multi")
  then
    local:multi-metrics($body)
  else
    let $database   := "OpsDirector"
    let $collection := "meters"
    let $cid        := $body/cluster ! xs:unsignedLong(.)
    let $rid        := $body/resource ! xs:unsignedLong(.)
    let $hid        := $body/host ! xs:unsignedLong(.)
    let $resourceGroupIds := $body/resourceGroupId ! xs:unsignedLong(.)

    (: retrieve all server ids this user has access to :)
    let $resource-ids := opsdir:managed-resource-ids($body/kind || "s")

    (: if single rid, check it is accessible:)
    let $rids := if (exists($rid))
                 then $rid[xs:string(.) = $resource-ids]
                 else $resource-ids

    (: if resourceGroupId supplied, match up resources :)
    let $rids := if ($resourceGroupIds)
                 then
                   let $resources := for $rgid in $resourceGroupIds
                                     return rg:get(xs:string($rgid))/resource
                   return distinct-values($rids[.=$resources])
                 else $rids
    return
      if (empty($rids))
      then ()
      else
        if ($body/format = "csv" or $body/format = "csvdebug")
        then
          local:csv($body/format, $body/kind, jslib:resource-metrics(
                      $body/kind,
                      $rids,
                      $body/period,
                      $body/start,
                      $body/end,
                      $body/summary,
                      $body/detail,
                      $body/metrics,
                      "xml",
                      $database,
                      $collection,
                      $cid,
                      $hid)/node())
        else
          jslib:gzip-reply(
                jslib:resource-metrics(
                    $body/kind,
                    $rids,
                    $body/period,
                    $body/start,
                    $body/end,
                    $body/summary,
                    $body/detail,
                    $body/metrics,
                    $body/format,
                    $database,
                    $collection,
                    $cid,
                    $hid)/node())
