"use strict";

/* Copyright 2002-2018 MarkLogic Corporation. All Rights Reserved. */

// See ...

declareUpdate();

const endpoint = require("../lib/endpoint");
const search   = require("../lib/search");
const jslib    = require('/common/jslib.xqy');
const rs       = require("/common/rscopes.xqy");

const NOTIFICATIONS_COLLECTION = "http://marklogic.com/opsdirector/v1/notifications";
const THRESHOLDS = ["critical", "at-risk", "all"];
const VALIDIDS = [ ...rs.resourceGroupIds() ];

const listFieldDefs = {
  "filter": "array",
  "query": "array",
  "text": "string",
  "start":  { type: "number", default: 0 },
  "length": { type: "number", default: 50 },
  "sort": "string",
  "order": { type: ["ascending", "descending"], default: "ascending" },
  "format": { type: ["json","csv"], default: "json" },
  "headers": "array",
  "filename": "string",

  "type": { type: "string", default: "email" }
};

const fieldDefs = {
  "name": { type: "string", required: true },
  "description": "string",
  "scopes": { type: "array", required: true },
  "threshold": { type: "string", required: true },
  "email": { type: "string", required: true },
  "enabled": { type: "boolean", required: true },
  "includeJson": { type: "boolean", default: false },
  "id": "string",
  "type": { type: "string", default: "email" }
};

class EP extends endpoint.Endpoint {
  GET() {
    return this.query(this.getRequestParams());
  }

  POST() {
    let body = this.getRequestBody();
    if (body === Object(body)) {
      return this.query(this.getRequestBody());
    } else {
      throw this.clientError("Payload must be an object");
    }
  }

  PUT() {
    let body = this.getRequestBody();
    if (body === Object(body)) {
      return this.query(this.getRequestBody());
    } else {
      throw this.clientError("Payload must be an object");
    }
  }

  DELETE() {
    return this.query(this.getRequestParams());
  }

  query(body) {
    let _ = fn.trace({"method": xdmp.getRequestMethod(),
                      "uri": xdmp.getOriginalUrl(),
                      "body": body},
                     "opsdir-app-endpoint");

    xdmp.securityAssert("http://marklogic.com/xdmp/privileges/opsdir-admin",
                        "execute");

    if (body.operation !== undefined) {
      throw this.clientError("Illegal body property: operation");
    }

    let op = xdmp.getRequestField("operation");

    // The rewriter guarantees that an id will be passed in for those
    // operations that require an id.
    // If an ID *is* passed in, make sure it's a legitimate id
    let id = xdmp.getRequestField("id");
    if (id !== null && !id.match(/^\d+(,\d+)*$/)) {
      throw this.clientError("Invalid notification id: " + id);
    }

    switch (op) {
    case "create":
      return this.create(body);
      break;
    case "list":
      return this.list(body);
      break;
    case "read":
      return this.read(body, id);
      break;
    case "update":
      return this.update(body, id);
      break;
    case "delete":
      return this.del(body, id);
      break;
    default:
      throw this.clientError("Illegal operation: " + op);
    }
  }

  list(body) {
    body = this.validShape(body, listFieldDefs);

    let baseCriteria = [ cts.collectionQuery([NOTIFICATIONS_COLLECTION]) ];

    let results = {
      "types": [ "email" ]
    };

    //body.fieldName = "email-notification-search";
    results["email"] = search.search(body, baseCriteria);

    // What if some of the resource scopes have been deleted...
    for (let idx in results.email.items) {
      results.email.items[idx] = this.cleanupData(results.email.items[idx]);
    }

    if (body.format == "csv") {
      let filename = this.csvFilename(body.filename, "notifications");
      return this.download(filename, body.headers, results["email"]["items"]);
    } else {
      return jslib.gzipReply(results);
    }
  }

  create(body) {
    body = this.validShape(body, fieldDefs);

    if (body.name === "") {
      throw new endpoint.EndpointError(400, "The 'name' is required.");
    }

    if (!body.description) {
      body.description = "";
    }

    if (THRESHOLDS.indexOf(body.threshold) < 0) {
      throw new endpoint.EndpointError(400, "Invalid threshold: " + body.threshold);
    }

    for (let idx in body.scopes) {
      let id = body.scopes[idx];
      if (VALIDIDS.indexOf(id) < 0) {
        throw new endpoint.EndpointError(400, "Invalid resource scope: " + id);
      }
    }

    const id = xdmp.random().toString();

    let result = {
      "id": id,
      "name": body.name,
      "description": body.description,
      "threshold": body.threshold,
      "email": body.email,
      "scopes": body.scopes,
      "enabled": body.enabled ? "true" : "false",
      "includeJson": body.includeJson ? "true" : "false",
      "type": "email"
    };

    xdmp.documentInsert(`/config/notifications/${id}.json`, result,
                        { permissions: [xdmp.permission('opsdir-admin', 'read'),
                                        xdmp.permission('opsdir-admin', 'update')],
                          collections: [ NOTIFICATIONS_COLLECTION ] });

    xdmp.setResponseCode(201, "Created");
    return result;
  }

  read(body, id) {
    if (body.id !== undefined && body.id !== id) {
      throw this.clientError("ID mismatch");
    }

    let alert = cts.doc(`/config/notifications/${id}.json`);

    if (alert === null) {
      xdmp.setResponseCode(404, "Not Found");
      return null;
    } else {
      alert = this.cleanupData(alert.toObject());
      return alert;
    }
  }

  update(body, id) {
    if (body.id !== undefined && body.id !== id) {
      throw this.clientError("ID mismatch");
    }

    body = this.validShape(body, fieldDefs);

    if (body.name === "") {
      throw new endpoint.EndpointError(400, "The 'name' is required.");
    }

    if (!body.description) {
      body.description = "";
    }

    if (THRESHOLDS.indexOf(body.threshold) < 0) {
      throw new endpoint.EndpointError(400, "Invalid threshold: " + body.threshold);
    }

    for (let idx in body.scopes) {
      let id = body.scopes[idx];
      if (VALIDIDS.indexOf(id) < 0) {
        throw new endpoint.EndpointError(400, "Invalid resource scope: " + id);
      }
    }

    let result = {
      "id": id,
      "name": body.name,
      "description": body.description,
      "threshold": body.threshold,
      "email": body.email,
      "scopes": body.scopes,
      "enabled": body.enabled ? "true" : "false",
      "includeJson": body.includeJson ? "true" : "false",
      "type": "email"
    };

    xdmp.documentInsert(`/config/notifications/${id}.json`, result,
                        { permissions: [xdmp.permission('opsdir-admin', 'read'),
                                        xdmp.permission('opsdir-admin', 'update')],
                          collections: [ NOTIFICATIONS_COLLECTION ] });

    xdmp.setResponseCode(204, "No content");
    return null;
  }

  del(body, id) {
    if (body.id !== undefined && body.id !== id) {
      throw this.clientError("ID mismatch");
    }

    for (let nid of id.split(",")) {
      let uri = `/config/notifications/${nid.trim()}.json`;
      if (fn.docAvailable(uri)) {
        xdmp.documentDelete(uri);
      }
    }
    xdmp.setResponseCode(204, "No content");
    return null;
  }

  cleanupData(body) {
    // Remove any scopes that no longer exist;
    let newScopes = [];
    for (let idx in body.scopes) {
      let id = body.scopes[idx];
      if (VALIDIDS.indexOf(id) >= 0) {
        newScopes.push(id);
      }
    }
    body.scopes = newScopes;

    // Convert string "enabled" into proper boolean. It's stored
    // in the database as a string so that we can sort on it.
    body.enabled = (body.enabled != "false");
    body.includeJson = (body.includeJson != "false");

    return body;
  }
}

endpoint.Endpoint.respond(EP);
