"use strict";

/* Copyright 2002-2018 MarkLogic Corporation. All Rights Reserved. */

const endpoint = require("../lib/endpoint");
const rg       = require("/common/resource-group.xqy");
const rs       = require("/common/rscopes.xqy");
const jslib    = require('/common/jslib.xqy');
const search   = require('../lib/search');

const fieldDefs = {
  "filter": "array",
  "text": "string",
  "start":  { type: "number", default: 0 },
  "length": { type: "number", default: 50 },
  "sort": "string",
  "order": { type: ["ascending", "descending"], default: "ascending" },
  "format": { type: ["json","csv"], default: "json" },
  "headers": "array",
  "filename": "string",

  "role": "array",
  "resource": "array",
  "id": "array",
  "type": "string"
};

class EP extends endpoint.Endpoint {
  GET() {
    return this.query(this.getRequestParams());
  }

  POST() {
    let body = this.getRequestBody();
    if (body === Object(body)) {
      return this.query(this.getRequestBody());
    } else {
      throw this.clientError("Payload must be an object");
    }
  }

  query(body) {
    let _ = fn.trace({"method": xdmp.getRequestMethod(),
                      "uri": xdmp.getOriginalUrl(),
                      "body": body},
                     "opsdir-app-endpoint");

    body = this.validShape(body, fieldDefs);

    let op = xdmp.getRequestField("operation");
    switch (op) {
    case "list":
      return this.list(body);
      break;
    default:
      throw this.clientError("Illegal operation: " + op, 400);
    }
  }

  list(body) {
    let result = {};

    let roleids = [];
    let invalidRole = false;
    if (body.role) {
      for (let role of body.role) {
        try {
          let rids = jslib.rolesAndSubroles(role);
          for (let rid of rids) {
            roleids.push(rid);
          }
        } catch (err) {
          if (err.name === "SEC-ROLEDNE") {
            invalidRole = true;
          } else {
            throw err;
          }
        }
      }
    }

    let resourceids = [];
    if (body.resource) {
      resourceids = body.resource;
    }

    let match = undefined;
    if (body.id) {
      match = body.id;
    }

    let type = body.type;
    let groups = [];
    for (let group of rg.list(type).toObject().resourceGroups) {
      let ok = false;
      ok = (!body.text || group.name.indexOf(body.text) >= 0);
      if (ok) {
        groups.push(group);
      }
    }

    if (match !== undefined) {
      let filter = [];
      for (let idx = 0; idx < groups.length; idx++) {
        if (match.indexOf(groups[idx].id.toString()) >= 0) {
          filter.push(groups[idx]);
        }
      }
      groups = filter;
    }

    if (roleids.length > 0) {
      let filter = [];
      for (let gidx = 0; gidx < groups.length; gidx++) {
        let rg = rs.resourceGroup(groups[gidx].id);
        let perms = xdmp.documentGetPermissions(xdmp.nodeUri(rg));
        let found = false;
        for (let idx = 0; !found && idx < perms.length; idx++) {
          let perm = perms[idx];
          found = found || (roleids.indexOf(perm.roleId) >= 0);
        }
        if (found) {
          filter.push(groups[gidx]);
        }
      }
      groups = filter;
    } else if (invalidRole) {
      // If all of the specified roles were invalid, then there can be no
      // resource groups in any of those roles. This is distinct from the
      // case where no roles are specified.
      groups = [];
    }

    if (resourceids.length > 0) {
      let filter = [];
      for (let gidx = 0; gidx < groups.length; gidx++) {
        let found = false;
        for (let idx = 0; !found && idx < groups[gidx].resource.length; idx++) {
          const dbid = groups[gidx].resource[idx].toString();
          found = found || resourceids.indexOf(dbid) >= 0;
        }
        if (found) {
          filter.push(groups[gidx]);
        }
      }
      groups = filter;
    }

    if (body.filter) {
      let filtered = [];
      for (let group of groups) {
        let obj = {};
        for (let prop of body.filter) {
          if (prop in group) {
            obj[prop] = group[prop];
          }
        }
        filtered.push(obj);
      }
      groups = filtered;
    }

    result = {
      "type": [ "resourceGroups" ],
      "resourceGroups": {
        "estimate": groups.length,
        "start": body.start,
        "count": Math.min(body.length, groups.length),
        "items": groups.slice(body.start, body.start + body.length)
      }
    };

    if (body.format == "csv") {
      let filename = this.csvFilename(body.filename, "licenses-" + body.aggregate);
      return this.download(filename, body.headers, result.resourceGroups.items);
    } else {
      return jslib.gzipReply(result);
    }
  }
}

endpoint.Endpoint.respond(EP);
