"use strict";

/* Copyright 2002-2018 MarkLogic Corporation. All Rights Reserved. */

declareUpdate();

const endpoint = require("../lib/endpoint");
const search   = require("../lib/search");
const t        = require("/common/opsdir.xqy");
const ts       = require("/common/opsec.xqy");
const rg       = require("/common/resource-group.xqy");
const jslib    = require('/common/jslib.xqy');

const RG_COLLECTION = 'http://marklogic.com/opsdirector/v1/resource-group';

const listFieldDefs = {
  "filter": "array",
  "query": "array",
  "text": "string",
  "start":  { type: "number", default: 0 },
  "length": { type: "number", default: 50 },
  "sort": "string",
  "order": { type: ["ascending", "descending"], default: "ascending" },
  "format": { type: ["json","csv"], default: "json" },
  "headers": "array",
  "filename": "string",

  // Query shortcuts
  "role": "array",
  "group": "array",
  "resource": "array",
  "type": "string",

  // For backwards compatibility, allow 'id' as a synonym of group
  "id": "array"
};

const accessFieldDefs = {
  "sort": "string",
  "order": { type: ["ascending", "descending"], default: "ascending" },
  "format": { type: ["json","csv"], default: "json" },
  "headers": "array",
  "filename": "string"
};

const updateFieldDefs = {
  "name": "string",
  "description": "string",
  "role": "array",
  "resource": "array",
  "roleOperation": { type: [ "add", "remove" ] },

  // For backwards compatibility, allow 'id' and 'type' in the body
  "id": "string",
  "type": "string"
};

const statusFieldDefs = {
  "format": { type: ["json"], default: "json" }
};

const createFieldDefs = {
  "name": { type: "string", required: true },
  "description": { type: "string", default: "" },
  "type": { type: ["databases","servers","hosts","clusters"], required: true },
  "resources": "array"
};

class EP extends endpoint.Endpoint {
  GET() {
    const headers = [ xdmp.addResponseHeader("Cache-Control", "no-cache"),
    xdmp.addResponseHeader("Cache-Control", "no-store") ];
    return this.query(this.getRequestParams());
  }

  POST() {
    let body = this.getRequestBody();
    if (body === Object(body)) {
      return this.query(this.getRequestBody());
    } else {
      throw this.clientError("Payload must be an object");
    }
  }

  PUT() {
    let body = this.getRequestBody();
    if (body === Object(body)) {
      return this.query(this.getRequestBody());
    } else {
      throw this.clientError("Payload must be an object");
    }
  }

  DELETE() {
    return this.query(this.getRequestParams());
  }

  query(body) {
    let _ = fn.trace({"method": xdmp.getRequestMethod(),
                      "uri": xdmp.getOriginalUrl(),
                      "body": body},
                     "opsdir-app-endpoint");

    if (body.operation !== undefined) {
      throw this.clientError("Illegal body property: operation");
    }

    let op = xdmp.getRequestField("operation");

    // The rewriter guarantees that an id will be passed in for those
    // operations that require an id.
    // If an ID *is* passed in, make sure it's a legitimate id
    let id = xdmp.getRequestField("id");
    if (id !== null && !id.match(/^\d+$/)) {
      throw this.clientError("Invalid resource group id: " + id);
    }

    switch (op) {
    case "create":
      return this.create(body);
      break;
    case "list":
      return this.list(body);
      break;
    case "access":
      return this.access(body);
      break;
    case "status":
      return this.status(body);
      break;
    case "read":
      return this.read(body, id);
      break;
    case "update":
      return this.update(body, id);
      break;
    case "delete":
      return this.del(body, id);
      break;
    default:
      throw this.clientError("Illegal operation: " + op);
    }
  }

  list(body) {
    body = this.validShape(body, listFieldDefs);

    if (!body.query) {
      body.query = [];
    }

    let roleQuery = [];
    if (body.role) {
      let roleids = [];
      for (let id of body.role) {
        try {
          let rids = jslib.rolesAndSubroles(id);
          roleids = [ ...roleids, ...rids ];
        } catch (err) {
          if (err.name !== "SEC-ROLEDNE") {
            throw err;
          }
        }
      }

      roleQuery = [ cts.fieldValueQuery("rg-roles", roleids) ];
    }

    let resourceQuery = [];
    if (body.resource) {
      resourceQuery = [ cts.fieldValueQuery("rg-resources", body.resource) ];
    }

    let idQuery = [];
    if (body.group) {
      idQuery = [ cts.jsonPropertyValueQuery("opsdirId", body.group) ];
    }
    if (body.id) {
      idQuery = idQuery.concat(cts.jsonPropertyValueQuery("opsdirId", body.id));
    }

    let typeQuery = [];
    if (body.type) {
      typeQuery = [ cts.jsonPropertyValueQuery("opsdirResourceType", body.type) ];
    }

    let textQuery = [];
    if (body.text && body.text.trim().length) {
      let textq = body.text.trim();
      if ((textq.length >= 3) && (textq.indexOf("*") < 0)) {
        textq = textq + "*";
      }
      textQuery = [ cts.fieldWordQuery("rg-search", textq) ];
      delete body["text"];
    }

    let baseQuery = [ cts.collectionQuery([RG_COLLECTION]), ...roleQuery,
                      ...resourceQuery, ...idQuery, ...typeQuery, ...textQuery ];

    // Unfortunately, for backwards compatibility, we expose different
    // property names in the payload than actually exist in the documents.
    // To make that work, we have to do the filtering *after* we convert
    // the payload. C'est la vie.
    let filter = body.filter;
    delete body.filter;

    // As a consequence of the hack below wrt field names, we need a hack
    // here as well.
    switch (body.sort) {
      case "id":
         body.sort = "opsdirId";
         break;
      case "type":
         body.sort = "opsdirResourceType";
         break;
      default:
         // nop
    }

    // Hack to reformat the results
    let results = search.search(body, baseQuery);
    let estimate = results.estimate;
    let groups = [];
    for (let doc of results.items) {
      // Note: a little hacky. Instead of complicated tests for filtering;
      // we just normalize the names in the group and then delete the ones
      // that are filtered out. Except for resourceNames because that's
      // potentially expensive.

      let group = { "id": doc.opsdirId, "type": doc.opsdirResourceType };
      for (let prop of ["name", "description", "size" ]) {
        if (prop in doc) {
          group[prop] = doc[prop];
        }
      }

      if ("resource" in doc) {
        let idlist = [];
        for (let id of doc.resource) {
          let obj = t.managedResource(null, null, id);
          if (obj) {
            idlist.push(id);
          }
        }

        group["resourceIds"] = idlist;

        if (!filter || filter.indexOf("resourceNames") >= 0) {
          group["resourceNames"] = [];
          // Why can't I iterate 'for (let id of doc.resource)'?
          for (let id of idlist) {
            let rsrc = t.managedResource(null, doc.opsdirResourceType, id).toObject();
            group["resourceNames"].push(rsrc.opsdirName);
          }
        }
      }

      if (!filter || filter.indexOf("roles") >= 0) {
        group["roles"] = [];
        // Why can't I iterate 'for (let id of doc.resource)'?
        for (let pos = 0; pos < doc.groupPermissions.length; pos++) {
          let role = ts.getRoleNames(doc.groupPermissions[pos].role).toObject();
          if (role.length != 0) {
            role = role[0].toString();
            if (role.indexOf("opsdir-rg-internal-") != 0) {
              group["roles"].push(role);
            }
          }
        }
      }

      // Remove the keys we don't need
      if (filter) {
        for (let key in group) {
          if (filter.indexOf(key) == -1) {
            delete group[key];
          }
        }
      }

      groups.push(group);
    }

    results = {
      "types": [ "resourceGroups" ],
      "resourceGroups": {
        "estimate": estimate,
        "start": body.start,
        "count": groups.length,
        "items": groups
      }
    };

    if (body.format == "csv") {
      let filename = this.csvFilename(body.filename, "resourceGroups");
      return this.download(filename, body.headers, results["resourceGroups"]["items"]);
    } else {
      return jslib.gzipReply(results);
    }
  }

  access(body) {
    body = this.validShape(body, listFieldDefs);

    let textQuery = [];
    if (body.text && body.text.trim().length) {
      let textq = body.text.trim();
      if ((textq.length >= 3) && (textq.indexOf("*") < 0)) {
        textq = textq + "*";
      }
      textQuery = [ cts.fieldWordQuery("rg-search", textq) ];
      delete body["text"];
    }

    let baseQuery = [ cts.collectionQuery([RG_COLLECTION]), ...textQuery ];

    // Hack to reformat the results
    let results = search.search(body, baseQuery);
    let groups = [];
    for (let doc of results.items) {
      let group = { "name": doc.name };
      group["roles"] = [];
      // Why can't I iterate 'for (let id of doc.resource)'?
      for (let pos = 0; pos < doc.groupPermissions.length; pos++) {
        let role = ts.getRoleNames(doc.groupPermissions[pos].role);
        role = role.toObject()[0].toString();
        if (role.indexOf("opsdir-rg-internal-") != 0) {
          group["roles"].push(role);
        }
      }
      groups.push(group);
    }

    results = {
      "types": [ "resourceGroups" ],
      "resourceGroups": {
        "estimate": groups.length,
        "start": body.start,
        "count": groups.length,
        "items": groups
      }
    };

    if (body.format == "csv") {
      let filename = this.csvFilename(body.filename, "resourceGroups");
      return this.download(filename, body.headers, results["resourceGroups"]["items"]);
    } else {
      return jslib.gzipReply(results);
    }
  }

  status(body) {
    body = this.validShape(body, listFieldDefs);

    let docs_collection = "http://marklogic.com/opsdirector/v1/join";
    if (xdmp.traceEnabled("opsdir-joins-test")) {
      docs_collection = "http://marklogic.com/opsdirector/v1/join/test/data";
    }

    let alerts_collection = "http://marklogic.com/opsdirector/v1/alerts";
    if (xdmp.traceEnabled("opsdir-joins-test")) {
      alerts_collection = "http://marklogic.com/opsdirector/v1/alerts/test/data";
    }

    // Return the documents that have interesting permissions. The fact that
    // a document is readable/updateable by "opsdir-data-internal" and/or
    // "opsdir-admin" isn't interesting.
    let dresult = {};
    let aresult = {};
    try {
      xdmp.securityAssert("http://marklogic.com/xdmp/privileges/opsdir-admin",
                          "execute");

      for (let uri of cts.uris(null, null, cts.collectionQuery(docs_collection))) {
        let join = cts.doc(uri).toObject();
        let obj = {
          "opsdirId": join.opsdirId,
          "opsdirClusterId": join.opsdirClusterId,
          "opsdirName": join.opsdirName,
          "opsdirResourceType": join.opsdirResourceType
        };

        if (join.opsdirResourceType === "tasks") {
          obj["path"] = join.path;
          obj["hostIds"] = join.hostIds;
          obj["hostNames"] = join.hostNames;
        }

        let perms = xdmp.documentGetPermissions(uri);
        for (let perm of perms) {
          let role = t.getRoleName(perm.roleId).toString();
          if ((role !== "opsdir-data-internal") && (role !== "opsdir-admin")) {
            let arr = obj[perm.capability + "Roles"] || [];
            arr.push(role);
            obj[perm.capability + "Roles"] = arr;
          }
        }
        dresult[uri] = obj;
      }

      for (let uri of cts.uris(null, null, cts.collectionQuery(alerts_collection))) {
        let join = cts.doc(uri).toObject();
        let obj = {
          "opsdirId": join.opsdirId,
          "opsdirClusterId": join.opsdirClusterId,
          "resourceName": join.resourceName,
          "resourceId": join.resourceId,
          "resourceType": join.resourceType,
          "status": join.status,
          "acknowledged": join.acknowledged
          //,"raw": join
        };
        let perms = xdmp.documentGetPermissions(uri);
        for (let perm of perms) {
          let role = t.getRoleName(perm.roleId).toString();
          if ((role !== "opsdir-data-internal") && (role !== "opsdir-admin")) {
            let arr = obj[perm.capability | "Roles"] || [];
            arr.push(role);
            obj[perm.capability + "Roles"] = arr;
          }
        }
        aresult[uri] = obj;
      }
    } catch (err) {
      // Nevermind; if you aren't opsdir-admin, then no data for you.
      console.log("Only opsdir-admin can see resource group status");
      console.log("%j", err);
    }

    let baseQuery = [ cts.collectionQuery([RG_COLLECTION]) ];

    // Hack to reformat the results
    let results = search.search(body, baseQuery);
    let groups = [];
    for (let doc of results.items) {
      // Note: a little hacky. Instead of complicated tests for filtering;
      // we just normalize the names in the group and then delete the ones
      // that are filtered out. Except for resourceNames because that's
      // potentially expensive.

      let group = { "id": doc.opsdirId, "type": doc.opsdirResourceType };
      for (let prop of ["name", "description", "size" ]) {
        if (prop in doc) {
          group[prop] = doc[prop];
        }
      }
      if ("resource" in doc) {
        group["resourceIds"] = doc.resource;
      }

      group["resourceNames"] = [];
      // Why can't I iterate 'for (let id of doc.resource)'?
      for (let pos = 0; pos < doc.resource.length; pos++) {
        let id = doc.resource[pos];
        let rsrc = t.managedResource(null, doc.opsdirResourceType, id).toObject();
        group["resourceNames"].push(rsrc.opsdirName);
      }

      group["roles"] = [];
      // Why can't I iterate 'for (let id of doc.resource)'?
      for (let pos = 0; pos < doc.groupPermissions.length; pos++) {
        let role = ts.getRoleNames(doc.groupPermissions[pos].role);
        role = role.toObject()[0].toString();
        if (role.indexOf("opsdir-rg-internal-") != 0) {
          group["roles"].push(role);
        }
      }

      groups.push(group);
    }

    results = {
      "types": [ "resourceGroups", "resources", "alerts" ],
      "resourceGroups": {
        "estimate": groups.length,
        "start": body.start,
        "count": groups.length,
        "items": groups
      },
      "resources": dresult,
      "alerts": aresult
    };

    return jslib.gzipReply(results);
  }

  create(body) {
    xdmp.securityAssert("http://marklogic.com/xdmp/privileges/opsdir-admin",
                        "execute");

    body = this.validShape(body, createFieldDefs);

    // For backwards compatability with the front end, allow resource=[]
    if (body.resource && body.resource.length == 0) {
      delete body.resource;
    }

    // You cannot create a resource group with roles or resources;
    // those are known limitations
    if (body.resource) {
      throw this.clientError("Invalid property: resource");
    }

    let name = body.name;
    let desc = body.description;
    let type = body.type;

    if (name === null || name === undefined) {
      throw new endpoint.EndpointError(400, "The 'name' is required.");
    }

    if (name === null || desc === undefined) {
      desc = "";
    }

    if (type === null || type === undefined) {
      throw new endpoint.EndpointError(400, "The 'type' is required.");
    }

    let result = rg.create(name, desc, type);

    xdmp.setResponseCode(201, "Created");
    return result;
  }

  read(body, id) {
    if (body.id !== undefined) {
      throw this.clientError("Conflicting id specifications");
    }

    let result = rg.get(id);
    if (result === null) {
      xdmp.setResponseCode(404, "Not found");
    }

    return result;
  }

  update(body, id) {
    xdmp.securityAssert("http://marklogic.com/xdmp/privileges/opsdir-admin",
                        "execute");

    body = this.validShape(body, updateFieldDefs);

    // Why is this allowed in either place?
    let roleOperation = xdmp.getRequestField("roleOperation");
    if (body.roleOperation) {
      if (roleOperation) {
        if (roleOperation !== body.roleOperation) {
          throw this.clientError("Conflicting role operations");
        }
      }
      roleOperation = body.roleOperation;
    }

    if (body.id !== undefined && body.id !== id) {
      throw this.clientError("ID mismatch");
    }

    // You can't change the type, so just ignore it.
    if (body.type) {
      delete body.type;
    }

    if (roleOperation) {
      rg.update(id, body, roleOperation);
    } else {
      rg.update(id, body);
    }

    xdmp.setResponseCode(204, "No content");
    return null;
  }

  del(body, id) {
    xdmp.securityAssert("http://marklogic.com/xdmp/privileges/opsdir-admin",
                        "execute");

    if (body.id !== undefined && body.id !== id) {
      throw this.clientError("ID mismatch");
    }

    rg.delete(id);
    xdmp.setResponseCode(204, "No content");
    return null;
  }

  arrContains(srcArray, targetArray) {
    let found = false;
    for (let value of srcArray) {
      found = found || (targetArray.indexOf(value) >= 0);
    }
    return found;
  }
}

endpoint.Endpoint.respond(EP);
