"use strict";

/* Copyright 2002-2018 MarkLogic Corporation. All Rights Reserved. */

// See https://wiki.marklogic.com/x/lYHqAw

const endpoint = require("../lib/endpoint");
const search   = require("../lib/search");
const jslib    = require('/common/jslib.xqy');
const sd       = require("../lib/systemDatabases.sjs");
const rg       = require('/common/resource-group.xqy');
const rs       = require('/common/rscopes.xqy');
const opsdir   = require("/common/opsdir.xqy");

const ALERTS_COLLECTION = "http://marklogic.com/opsdirector/v1/alerts";
const HEALTH_STATUS_RANK = {
  'info': 0,
  'healthy': 1,
  'maintenance': 2,
  'offline': 3,
  'unknown': 4,
  'at-risk': 5,
  'critical': 6
};

const resourceFields = {
  "clusters": "clusters-search",
  "hosts": "hosts-search",
  "servers": "servers-search",
  "databases": "databases-search",
  "forests": "forests-search",
  "tasks": "tasks-search",
  "groups": null
};

const RESOURCE_TYPES = Object.keys(resourceFields);

const fieldDefs = {
  "type": { "type": [ ...RESOURCE_TYPES, "logs" ] },
  "filter": "array",
  "query": "array",
  "start":  { type: "number", default: 0 },
  "length": { type: "number", default: 50 },
  "sort": "string",
  "text": "string",
  "order": { type: ["ascending", "descending"], default: "ascending" },
  "format": { type: ["json","csv"], default: "json" },
  "headers": "array",
  "filename": "string",

  "view": { type: ["tree", "list"], default: "list" },
  "resourceGroupId": "array",
  "excludeAncillary": "boolean"
  // N.B. clusterId and databaseId are no longer allowed;
  // for clusterId use [{"opsdirClusterId": "cluster_id"}] as a query.
  // for databaseId use [{"databaseIds": "database_id"}] as a query.
};

class EP extends endpoint.Endpoint {
  POST() {
    let body = this.getRequestBody();
    if (body === Object(body)) {
      return this.query(this.getRequestBody());
    } else {
      throw this.clientError("Payload must be an object");
    }
  }

  GET() {
    return this.query(this.getRequestParams());
  }

  query(body) {
    let _ = fn.trace({"method": xdmp.getRequestMethod(),
                      "uri": xdmp.getOriginalUrl(),
                      "body": body},
                     "opsdir-app-endpoint");

    body = this.validShape(body, fieldDefs);

    // By default, all the resource types. That includes logs if we're doing CSV output
    let rtypes = RESOURCE_TYPES;
    if (body.format === "csv") {
      rtypes = [ ...RESOURCE_TYPES, "logs" ];
    }

    if (body.type) {
      if (Array.isArray(body.type)) {
        rtypes = body.type;
      } else {
        rtypes = [ body.type ];
      }
    }

    if (rtypes.indexOf("logs") >= 0 && body.format !== "csv") {
      throw this.clientError("Logs can only be specified for csv output");
    }

    let results = {
      "view": body.view,
      "types": rtypes
    };

    if (body.view == "tree") {
      results.info = {};
    }

    // Filtering has to be done after the search because we need
    // at least the opsdirId property in order to inject the health
    let propFilter = body.filter;
    delete body.filter;

    // If resource groups are specified, work out the constraints
    let extraCriteria = null;
    if (body.resourceGroupId) {
      let rgResourceIds = {};
      for (let id of body.resourceGroupId) {
        let resourceGroup = rg.get(id);
        if(!resourceGroup) {
          throw this.clientError("Payload contains invalid resourceGroupId");
        }
        for (var i = 0; i < resourceGroup.resource.length; i++) {
          let id = resourceGroup.resource[i];
          rgResourceIds[id] = resourceGroup.type;

          let rsrc = opsdir.managedResource(null, resourceGroup.type, id);
          if (rsrc) {
            for (let inh of rs.inheritedResources(rsrc)) {
              let rtype = "";
              if (inh.opsdirResourceType) {
                rtype = inh.opsdirResourceType.toString();
              }
              if (RESOURCE_TYPES.indexOf(rtype) >= 0) {
                rgResourceIds[inh.opsdirId] = rtype;
              }
            }
          }
        }
      }

      let ids = [];
      for (let id in rgResourceIds) {
        ids.push(id);
      }
      extraCriteria = cts.jsonPropertyValueQuery('opsdirId',ids);
    }

    // I could be more specific, but the server relies on globally unique
    // IDs so I will too.

    // Get all of the non-info alerts that are open; we do this once
    // and iterate over the result instead of doing it once for each
    // resource type.
    let aCriteria = [ cts.collectionQuery([ALERTS_COLLECTION]) ];
    let aBody = { "start": 0,
                  "length": 9999,
                  "sort": "ts",
                  "order": "descending",
                  "filter": [ "ts", "status", "opsdirId", "code", "message", "state",
                              "acknowledged", "resourceId" ],
                  "query": [ { "status": "open" },
                             { "state": { "op": "!=", "value": "info" } } ] };
    let alerts = search.search(aBody, aCriteria);

    // Cache for unreachable clusters
    this.clusterAlertCache = {};

    for (let rtype of rtypes) {
      if (rtype === "logs") {
        results[rtype] = this.getLogs(body);
        console.log(results[rtype]);
      } else {
        let baseQuery = cts.collectionQuery("http://marklogic.com/opsdirector/v1/join");

        if (xdmp.traceEnabled("opsdir-joins-test")) {
          baseQuery = cts.collectionQuery("http://marklogic.com/opsdirector/v1/join/test/data");
        }

        let typeQuery = cts.jsonPropertyValueQuery('opsdirResourceType', rtype);
        if ((rtype == "databases") && body.excludeAncillary) {
          // Filter out system databases
          let exclQuery = cts.notQuery(
            cts.jsonPropertyValueQuery('opsdirName', sd.systemDatabases()));
          baseQuery = cts.andQuery([baseQuery, typeQuery, exclQuery, extraCriteria]);
        } else {
          baseQuery = cts.andQuery([baseQuery, typeQuery, extraCriteria]);
        }

        body.fieldName = resourceFields[rtype];
        let typeResult = search.search(body, baseQuery);

        results[rtype] = {};
        for (let key in typeResult) {
          if (key == "items") {
            results[rtype][key] = [];
            for (let item of typeResult[key]) {
              delete item["statusInfo"];
              results[rtype][key].push(this.checkHealth(alerts, item));

              if (rtype === "tasks" && body.format === "csv") {
                if (item.taskHost === null) {
                  item.hostNames = "All hosts in group " + item.groupName;
                }
              }
            }
          } else {
            results[rtype][key] = typeResult[key];
          }

          if (propFilter && (key === "items")) {
            for (let item of results[rtype][key]) {
              for (let prop in item) {
                if (propFilter.indexOf(prop) < 0) {
                  delete item[prop];
                }
              }
            }
          }
        }

        if (body.view == "tree") {
          let ids = [];
          for (let item of results[rtype]["items"]) {
            ids.push(item.opsdirId);
            results.info[item.opsdirId] = item;
          }
          results[rtype] = ids;
        }
      }
    }

    if (body.format == "csv") {
      if (rtypes.length == 1) {
        let filename = this.csvFilename(body.filename, rtypes[0]);
        return this.download(filename, body.headers, results[rtypes[0]]["items"]);
      } else {
        // Because we're abusing the headers property in this case, it's
        // going to be an array containing an object; get the object.
        let headers = body.headers[0];
        if (body.headers.length != 1 || (headers !== Object(headers))) {
          throw this.clientError("The headers property must be a single object");
        }

        let manifest = [];
        let nodes = [];

        for (let rtype of rtypes) {
          manifest.push({"path": rtype + ".csv"});

          let data = this.formatCsv(headers[rtype], results[rtype]["items"]);
          let builder = new NodeBuilder();
          builder.addText(data);
          nodes.push(builder.toNode());
        }

        let filename = this.csvFilename(null, "manage");
        if (body.filename) {
          filename = this.csvFilename(body.filename, null);
        }

        if (! xdmp.traceEnabled("opsdir-debug-csv")) {
          xdmp.addResponseHeader('Content-Disposition',
                                 'attachment; filename=' + filename);
          xdmp.setResponseContentType("application/zip");
        }
        return xdmp.zipCreate(manifest, nodes);
      }
    } else {
      return jslib.gzipReply(results);
    }
  }

  checkHealth(results, item) {
    // Is the cluster that this item is from currently unreachable?
    let unreachable = null;
    if (item.opsdirClusterId in this.clusterAlertCache) {
      unreachable = this.clusterAlertCache[item.opsdirClusterId];
    } else {
      for (let alert of results.items) {
        if ((alert.opsdirId == item.opsdirClusterId)
            && (alert.status == "open")
            && (alert.code == "OPSDIR-CLUSTER-HOSTS-UNREACHABLE")) {
          unreachable = alert;
        }
      }
      this.clusterAlertCache[item.opsdirClusterId] = unreachable;
    }

    let rank = 0;
    let health = null;
    for (let alert of results.items) {
      if (alert.resourceId != item.opsdirId) {
        continue;
      }

      if (health == null) {
        health = alert;
        rank = HEALTH_STATUS_RANK[health.state];
      }

      let newrank = HEALTH_STATUS_RANK[alert.state];
      if (newrank > rank) {
        health = alert;
        rank = newrank;
      }
    }

    if (unreachable) {
      item["health"] = Object.assign({}, unreachable);
      item["health"]["resourceId"] = item.opsdirId;
    } else {
      if (health && (rank > 0)) {
        item["health"] = health;
      } else {
        item["health"] = { "state": "healthy", "message": "healthy" };
      }
    }

    return item;
  }

  getLogs(body) {
    const LOGS_COLLECTION = 'http://marklogic.com/opsdirector/v1/logs';
    let search_results = search.search(body, cts.collectionQuery([LOGS_COLLECTION]));
    let result = {};
    result.estimate = search_results.estimate;
    result.start = search_results.start;
    result.count = search_results.count;
    result.items = search_results.items;
    return result;
  }
}

endpoint.Endpoint.respond(EP);
