"use strict";

/* Copyright 2002-2018 MarkLogic Corporation. All Rights Reserved. */

const endpoint = require("../lib/endpoint");
const search   = require("../lib/search");
const rbac     = require("/common/rbac.xqy");
const t        = require("/common/opsdir.xqy");
const rs       = require("/common/rscopes.xqy");
const ts       = require("/common/opsec.xqy");
const jslib    = require('/common/jslib.xqy');

const listFieldDefs = {
  "filter": "array",
  "regex": "string",
  "start":  { type: "number", default: 0 },
  "length": { type: "number", default: 50 },
  "sort": "string",
  "order": { type: ["ascending", "descending"], default: "ascending" },
  "roles": "array",
  "groupIds": "array",
  "showRolesWithNoResourceAccess": { type: "boolean", default: "true" },
  "format": { type: ["json","csv"], default: "json" },
  "headers": "array",
  "filename": "string"
};

const createFieldDefs = {
  "name": { type: "string", required: true },
  "description": { type: "string", default: "" },
  "subroles": "array"
};

class EP extends endpoint.Endpoint {
  GET() {
    return this.query(this.getRequestParams());
  }

  POST() {
    let body = this.getRequestBody();
    if (body === Object(body)) {
      return this.query(this.getRequestBody());
    } else {
      throw this.clientError("Payload must be an object");
    }
  }

  PUT() {
    let body = this.getRequestBody();
    if (body === Object(body)) {
      return this.query(this.getRequestBody());
    } else {
      throw this.clientError("Payload must be an object");
    }
  }

  DELETE() {
    return this.query(this.getRequestParams());
  }

  query(body) {
    let _ = fn.trace({"method": xdmp.getRequestMethod(),
                      "uri": xdmp.getOriginalUrl(),
                      "body": body},
                     "opsdir-app-endpoint");

    let op = xdmp.getRequestField("operation");
    switch (op) {
    case "create":
      return this.create(body);
      break;
    case "list":
      return this.read(body, null);
      break;
    case "read":
      return this.read(body, xdmp.getRequestField("name"));
      break;
    case "update":
      return this.update(body);
      break;
    case "delete":
      return this.del(body);
      break;
    default:
      throw new endpoint.EndpointError(400, "Illegal operation: " + op);
    }
  }

  create(body) {
    xdmp.securityAssert("http://marklogic.com/xdmp/privileges/opsdir-admin",
                        "execute");

    body = this.validShape(body, createFieldDefs);

    let name  = body.name;
    let desc  = body.description;
    let roles = body.subroles;

    if (!desc) {
      desc = "";
    }

    if (!roles) {
      roles = [];
    }

    try {
      let result = rbac.createRole(name, roles, desc);
      xdmp.setResponseCode(201, "Created");
      return result;
    } catch (error) {
      if (error.name !== "SEC-ROLEEXISTS") {
        throw error;
      }
      // This should be idempotent, but for backwards compatibility...
      throw this.clientError("Role exists");
    }
  }

  read(body, name) {
    body = this.validShape(body, listFieldDefs);

    let showGroups = (!body.filter
                      || body.groupIds
                      || body.filter.indexOf("resourceGroupNames") >= 0
                      || body.filter.indexOf("resourceGroupIds") >= 0
                      || body.sort == "resourceGroupNames"
                      || body.sort == "resourceGroupIds");

    let groups = [];
    if (showGroups) {
      // Get the resource groups
      for (let doc of rs.resourceGroups().toObject()) {
        let group = {};
        group["name"] = doc.name.toString();
        group["id"] = doc.opsdirId.toString();
        group["roles"] = [];
        // Why can't I iterate 'for (let id of doc.resource)'?
        for (let pos = 0; pos < doc.groupPermissions.length; pos++) {
          let role = ts.getRoleNames(doc.groupPermissions[pos].role).toObject();
          if (role.length != 0) {
            role = role[0].toString();
            if (role.indexOf("opsdir-rg-internal-") != 0) {
              group["roles"].push(role);
            }
          }
        }
        groups.push(group);
      }
    }

    let roleList = [];
    if (name) {
      let role = rbac.getRole(name);
      if (role !== null) {
        roleList = [ role ];
      }
      if (roleList.length == 0) {
        throw new endpoint.EndpointError(404);
      }
    } else {
      roleList = rbac.listRoles(body.regex).toObject();
    }

    let roles = [];
    for (let role of roleList) {
      let name = role["name"].toString();

      if (!body.roles || body.roles.indexOf(name) >= 0) {
        let obj = {};
        obj["name"] = name;
        obj["description"] = role["description"].toString();
        obj["subroles"] = role["subroles"].toObject();
        obj["resourceGroupNames"] = [];
        obj["resourceGroupIds"] = [];
        obj["inheritedResourceGroups"] = [];
        for (let group of groups) {
          let found = (group["roles"].indexOf(name) >= 0);
          let foundInheritedOnly = false;
          for (let subrole of obj["subroles"]) {
            if (!found) {
              foundInheritedOnly = (group["roles"].indexOf(subrole) >= 0);
            }
            found = found || foundInheritedOnly;
          }

          if (found) {
            obj["resourceGroupNames"].push(group["name"]);
            obj["resourceGroupIds"].push(group["id"]);
          }

          if (foundInheritedOnly) {
            obj["inheritedResourceGroups"].push({name: group["name"], id: group["id"]});
          }
        }

        if (!body.groupIds || this.arrContains(obj["resourceGroupIds"], body.groupIds)) {
          if (obj["resourceGroupNames"].length > 0
              || body.showRolesWithNoResourceAccess) {
            roles.push(obj);
          }
        }
      }
    }

    if (body.sort) {
      roles = this.sort(body, roles);
    }

    let totroles = roles.length;
    roles = roles.slice(body.start, body.start + body.length);

    // Remove the keys we don't need (after sorting!)
    if (body.filter) {
      for (let obj of roles) {
        for (let key in obj) {
          if (body.filter.indexOf(key) == -1) {
            delete obj[key];
          }
        }
      }
    }

    let results = {
      "types": [ "roles" ],
      "roles": {
        "estimate": totroles,
        "start": body.start,
        "count": roles.length,
        "items": roles
      }
    };

    if (body.format == "csv") {
      let filename = this.csvFilename(body.filename, "roles");
      return this.download(filename, body.headers, results["roles"]["items"]);
    } else {
      return jslib.gzipReply(results);
    }
  }

  sort(body, roles) {
    switch (body.sort) {
      case "name":
        return roles.sort(function(aobj,bobj) {
          let a = (aobj["name"] || "").toLowerCase();
          let b = (bobj["name"] || "").toLowerCase();
          if (body.order == "ascending") {
            return a < b ? -1 : (b < a ? 1 : 0);
          } else {
            return a > b ? -1 : (b > a ? 1 : 0);
          }
        });
      case "description":
        return roles.sort(function(aobj,bobj) {
          let a = (aobj["description"] || "").toLowerCase();
          let b = (bobj["description"] || "").toLowerCase();
          if (body.order == "ascending") {
            return a < b ? -1 : (b < a ? 1 : 0);
          } else {
            return a > b ? -1 : (b > a ? 1 : 0);
          }
        });
      case "subroles":
        return roles.sort(function(aobj,bobj) {
          let a = (aobj["subroles"][0] || "").toLowerCase();
          let b = (bobj["subroles"][0] || "").toLowerCase();
          if (body.order == "ascending") {
            return a < b ? -1 : (b < a ? 1 : 0);
          } else {
            return a > b ? -1 : (b > a ? 1 : 0);
          }
        });
      case "resourceGroupNames":
        return roles.sort(function(aobj,bobj) {
          let a = (aobj["resourceGroupNames"][0] || "").toLowerCase();
          let b = (bobj["resourceGroupNames"][0] || "").toLowerCase();
          if (body.order == "ascending") {
            return a < b ? -1 : (b < a ? 1 : 0);
          } else {
            return a > b ? -1 : (b > a ? 1 : 0);
          }
        });
      case "resourceGroupIds":
        return roles.sort(function(aobj,bobj) {
          let a = (aobj["resourceGroupIds"][0] || "").toLowerCase();
          let b = (bobj["resourceGroupIds"][0] || "").toLowerCase();
          if (body.order == "ascending") {
            return a < b ? -1 : (b < a ? 1 : 0);
          } else {
            return a > b ? -1 : (b > a ? 1 : 0);
          }
        });
      default:
        throw this.clientError("Cannot sort on field: " + body.sort);
    }
  }

  update(body) {
    xdmp.securityAssert("http://marklogic.com/xdmp/privileges/opsdir-admin",
                        "execute");

    let name = xdmp.getRequestField("name");

    if (body.name && body.name !== name) {
      throw new endpoint.EndpointError(400, "Name mismatch.");
    }

    body.name = name;

    try {
      rbac.updateRole(body);
      xdmp.setResponseCode(204, "No content");
      return null;
    } catch (error) {
      if (error.name == "OD-BADROLE") {
        throw new endpoint.EndpointError(400, "Invalid payload");
      } else if (error.name == "OD-NOROLE") {
        throw new endpoint.EndpointError(404, "No such role");
      } else {
        throw error;
      }
    }
  }

  del(body) {
    xdmp.securityAssert("http://marklogic.com/xdmp/privileges/opsdir-admin",
                        "execute");

    let name = xdmp.getRequestField("name");

    if (body.name && body.name !== name) {
      throw new endpoint.EndpointError(400, "Name mismatch.");
    }

    rbac.deleteRole(name);
    xdmp.setResponseCode(204, "No content");
    return null;
  }

  arrContains(srcArray, targetArray) {
    let found = false;
    for (let value of srcArray) {
      found = found || (targetArray.indexOf(value) >= 0);
    }
    return found;
  }
}

endpoint.Endpoint.respond(EP);
