"use strict";

/* Copyright 2002-2018 MarkLogic Corporation. All Rights Reserved. */

// See https://wiki.marklogic.com/???

const endpoint = require("../lib/endpoint");
const search   = require("../lib/search");
const jslib    = require('/common/jslib.xqy');
const sd       = require("../lib/systemDatabases.sjs");
const opsdir   = require("/common/opsdir.xqy");

const RESOURCE_TYPES = [ "clusters", "hosts", "servers", "databases", "forests" ];

// These are pseudo-v1 field defs; they aren't really the same,
// but they're the ones actually used by the /v1/ calls in the
// app.
const v1fieldDefs = {
  "statusType": { "type": [ "local-cluster-status",
                            "server-status", "forest-status", "database-status",
                            "host-status", "counts" ] },
  "start":  { type: "number", default: 0 },
  "length": { type: "number", default: 50 },
  "sort": "string",
  "order": { type: ["ascending", "descending"], default: "ascending" },
  "id": "array",
  "hostId": "array",
  "clusterId": "array",
  "format": { "type": [ "json" ] }
};

const fieldDefs = {
  "statusType": { "type": [ "cluster-status",
                            "server-status", "forest-status", "database-status",
                            "host-status", "counts" ] },
  "start":  { type: "number", default: 0 },
  "length": { type: "number", default: 50 },
  "sort": "string",
  "order": { type: ["ascending", "descending"], default: "ascending" },
  "opsdirId": "array",
  "format": { "type": [ "json" ] },
  "v1payload": { "type": "boolean", "default": false }
};

class EP extends endpoint.Endpoint {
  POST() {
    let body = this.getRequestBody();
    if (body === Object(body)) {
      return this.query(this.getRequestBody());
    } else {
      throw this.clientError("Payload must be an object");
    }
  }

  GET() {
    return this.query(this.getRequestParams());
  }

  query(body) {
    let _ = fn.trace({"method": xdmp.getRequestMethod(),
                      "uri": xdmp.getOriginalUrl(),
                      "body": body},
                     "opsdir-app-endpoint");

    let v1Type = undefined;

    if (("id" in body) || ("hostId" in body) || ("clusterId" in body)) {
      body = this.validShape(body, v1fieldDefs);

      v1Type = body.statusType;
      let v2 = { "v1payload": true };
      for (let key in body) {
        switch (key) {
        case "statusType":
          if (body.statusType === "local-cluster-status") {
            v2.statusType = "cluster-status";
          } else {
            v2.statusType = body.statusType;
          }
          break;
        case "id":
          if (body.id.length > 1
              || (body.id.length === 1 && body.id[0] !== "")) {
            v2.opsdirId = body.id;
          }
          break;
        case "hostId":
          if (body.hostId.length > 1
              || (body.hostId.length === 1 && body.hostId[0] !== "")) {
            v2.opsdirId = body.hostId;
          }
          break;
        case "clusterId":
          if (body.clusterId.length > 1
              || (body.clusterId.length === 1 && body.clusterId[0] !== "")) {
            v2.opsdirId = body.clusterId;
          }
          break;
        default:
          v2[key] = body[key];
        }
      }

      body = v2;
    }

    body = this.validShape(body, fieldDefs);

    let statusPropName = undefined;
    let rtypes = RESOURCE_TYPES;
    switch (body.statusType) {
    case "server-status":
      rtypes = [ "servers" ];
      statusPropName = "statusProperties";
      break;
    case "forest-status":
      rtypes = [ "forests" ];
      statusPropName = "statusProperties";
      break;
    case "database-status":
      rtypes = [ "databases" ];
      statusPropName = "statusProperties";
      break;
    case "host-status":
      rtypes = [ "hosts" ];
      statusPropName = "statusProperties";
      break;
    case "counts":
      rtypes = [ "databases" ];
      statusPropName = "countProperties";
      break;
    default:
      // "local-cluster-status"
      rtypes = [ "clusters" ];
      statusPropName = "statusProperties";
      break;
    }

    let results = {
      "types": rtypes
    };

    for (let rtype of rtypes) {
      let queries = [ cts.collectionQuery("http://marklogic.com/opsdirector/v1/join") ];

      if (xdmp.traceEnabled("opsdir-joins-test")) {
        queries =
          [ cts.collectionQuery("http://marklogic.com/opsdirector/v1/join/test/data") ];
      }

      if (body.opsdirId !== undefined) {
        queries.push(cts.jsonPropertyValueQuery('opsdirId', body.opsdirId));
      }

      queries.push(cts.jsonPropertyValueQuery('opsdirResourceType', rtype));

      if ((rtype == "databases") && body.excludeAncillary) {
        // Filter out system databases
        queries.push(cts.notQuery(
            cts.jsonPropertyValueQuery('opsdirName', sd.systemDatabases())));
      }

      let baseQuery = cts.andQuery(queries);
      let typeResult = search.search(body, baseQuery);
      results[rtype] = {};
      for (let key in typeResult) {
        if (key == "items") {
          results[rtype][key] = [];
          for (let item of typeResult[key]) {
            let stats = {};
            for (let prop of [ "opsdirId", "opsdirClusterId", "opsdirName",
                               "opsdirClusterName", "opsdirResourceType" ]) {
              stats[prop] = item[prop];
            }

            if (item.statusInfo) {
              if (body.statusType) {
                stats[statusPropName] = item.statusInfo[statusPropName];
              } else {
                for (let prop in item.statusInfo) {
                  stats[prop] = item.statusInfo[prop];
                }
              }
            } else {
              stats[statusPropName] = {};
            }

            results[rtype][key].push(stats);
          }
        } else {
          results[rtype][key] = typeResult[key];
        }
      }

      if (body.statusType == "counts") {
        let fCounts = [];
        for (let item of results.databases.items) {
          let rsrc = opsdir.managedResource(null, "databases", item.opsdirId).toObject();
          for (let fid of rsrc.forestIds) {
            let queries = [ cts.collectionQuery("http://marklogic.com/opsdirector/v1/join") ];
            if (xdmp.traceEnabled("opsdir-joins-test")) {
              queries =
                [ cts.collectionQuery("http://marklogic.com/opsdirector/v1/join/test/data") ];
            }
            queries.push(cts.jsonPropertyValueQuery('opsdirId', fid));
            queries.push(cts.jsonPropertyValueQuery('opsdirResourceType', "forests"));
            let baseQuery = cts.andQuery(queries);
            let searchResult = search.search(body, baseQuery);
            if (searchResult.items && searchResult.items[0]) {
              let fitem = searchResult.items[0];
              if (fitem.statusInfo && fitem.statusInfo.countProperties) {
                fCounts.push({
                  "opsdirId": rsrc.opsdirId,
                  "opsdirClusterId": rsrc.opsdirClusterId,
                  "opsdirName": rsrc.opsdirName,
                  "opsdirClusterName": rsrc.opsdirClusterName,
                  "opsdirResourceType": "forests",
                  "countProperties": fitem.statusInfo.countProperties
                });
              }
            }
          }
        }

        results.forests = {
          "estimate": fCounts.length,
          "start": 0,
          "count": fCounts.length,
          "items": fCounts
        };
      }
    }

    if (body.statusType == "counts") {
      // Can't do this in the loop above because we're iterating over the types!
      results.types.push("forests");
    }

    if (body.v1payload) {
      let rtype = results.types[0];
      let obj = results[rtype];
      results = obj;

      if (results.items) {
        for (let item of results.items) {
          item.opsdirResourceType = v1Type;
          item.opsdirKind = "stats";
          item.opsdirResourceName = item.opsdirName;
          delete item.opsdirName;

          let cluster = opsdir.managedResource(null, "clusters", item.opsdirClusterId);
          if (cluster && cluster.opsdirName) {
            item.opsdirClusterName = cluster.opsdirName;
          } else {
            item.opsdirClusterName = "???";
          }
        }
      }
    } else {
      let newTypes = [];

      for (let rtype of results.types) {
        for (let item of results[rtype].items) {
          let cluster = opsdir.managedResource(null, "clusters", item.opsdirClusterId);
          item.opsdirClusterName = cluster.opsdirName;
        }

        let newType = rtype.substring(0, rtype.length - 1);

        if (body.statusType === "counts") {
          newType += "Counts";
        } else {
          newType += "Status";
        }
        newTypes.push(newType);

        results[newType] = results[rtype];
        delete results[rtype];
      }

      results.types = newTypes;
    }

    return jslib.gzipReply(results);
  }
}

endpoint.Endpoint.respond(EP);
