"use strict";

/* Copyright 2002-2018 MarkLogic Corporation. All Rights Reserved. */

class EndpointError {
  constructor(statusCode, response) {
    this.name = "EndpointError";
    this.statusCode = statusCode;
    this.response = response;
    this.auth = this.getAuthType();    
    switch (statusCode) {
    case 400:
      this.message = 'Bad Request';
      break;
    case 401:
      this.message = 'Unauthorized';
      break;
    case 403:
      this.message = 'Forbidden';
      break;
    case 404:
      this.message = 'Not Found';
      break;
    case 405:
      this.message = 'Method Not Allowed';
      break;
    default:
      this.statusCode = 500;
      this.message = 'Internal Server Error';
      break;
    }
  }
  toString() {
    return this.response;
  }
  getAuthType(){
    const jslib = require('/common/jslib.xqy');
    try {
      return jslib.serverAuthentication();
    } catch (err) {
      return "unknown";
    }
  }
}

class Endpoint {
  static handleErrors(responseFunc) {
    try {
      return responseFunc() || {};
    } catch (err) {
      if (err.statusCode && err.statusCode < 500) {
        xdmp.setResponseCode(err.statusCode, err.message);
        return err;
      } else {
        throw err;
      }
    }
  }

  getRequestBody() {
    try {
      return xdmp.getRequestBody().toObject();
    }
    catch (err) {
      throw new EndpointError(400, err.stack);
    }
  }

  validShape(body, fieldDefs) {
    // We assume that fieldDefs is defined correctly
    for (let name in fieldDefs) {
      let def = fieldDefs[name];
      if (def === Object(def) && body[name] === undefined) {
        if (def.required) {
          throw new EndpointError(400, "Required field is absent: " + name);
        }
        if ("default" in def) {
          body[name] = def.default;
        }
      }
    }

    let valid = {};
    for (let name in body) {
      if (fieldDefs[name] === undefined) {
        throw new EndpointError(400, "Unexpected property: " + name);
      }

      let propType = undefined;
      let propCard = undefined;
      let value = body[name];

      if (value === null) {
        continue; // null is always ok.
      }

      if (fieldDefs[name] === Object(fieldDefs[name])) {
        propType = fieldDefs[name].type;
        propCard = fieldDefs[name].cardinality;
      } else {
        propType = fieldDefs[name];
      }

      if (Array.isArray(propType)) {
        let values = [ value ];
        if (Array.isArray(value)) {
          values = value;
        }

        if (propCard) {
          if (values.length !== propCard) {
            throw new EndpointError(400, "Invalid cardinality for "
                                         + name + ": " + values.length);
          }
        }

        for (let vitem of values) {
          if (propType.indexOf(vitem) < 0) {
            throw new EndpointError(400, "Invalid value for property "
                                         + name + ": " + vitem);
          }
        }
      } else {
        if (propType !== null) {
          switch (propType) {
          case "array":
            if (!Array.isArray(value)) {
              value = [ value ];
            }
            if (propCard) {
              if (value.length !== propCard) {
                throw new EndpointError(400, "Invalid cardinality for "
                                             + name + ": " + value.length);
              }
            }
            break;
          case "object":
            if (value !== Object(value)) {
              throw new EndpointError(400, "Property value must be an object: " + name);
            }
            break;
          case "number":
            // Yes, we allow "0" to become 0. On purpose.
            if (typeof value == "string") {
              value = Number.parseFloat(value);
            }
            if (Number(parseFloat(value)) !== value) {
              throw new EndpointError(400, "Property value must be a number: " + name);
            }
            break;
          case "boolean":
            if (value.toString() !== "true" && value.toString() !== "false") {
              throw new EndpointError(400, "Property value must be a boolean: " + name);
            }
            value = (value.toString() === "true");
            break;
          case "string":
            if (typeof value != "string") {
              throw new EndpointError(400, "Property value must be a string: " + name);
            }
            break;
          default:
            throw new EndpointError(500, "Invalid property type: " + propType);
          }
        }
      }

      valid[name] = value;
    }

    if ("format" in fieldDefs && valid["format"] != "csv") {
      if (("filename" in valid) || ("headers" in valid)) {
        throw new EndpointError(400, "The filename and headers properties only apply to CSV");
      }
    }

    return valid;
  }

  clientError(message, code) {
    if (code == null) {
      code = 400;
    }

    return new EndpointError(code, message);
  }

  // http://stackoverflow.com/questions/8486099/
  //        how-do-i-parse-a-url-query-parameters-in-javascript
  // Debugged by nwalsh
  getJsonFromUrl(query) {
    var result = {};
    query.split("&").forEach(function (part) {
      if (part) {
        // replace every + with space, regexp-free version
        part = part.split("+").join(" ");
        var eq = part.indexOf("=");
        var key = eq > -1 ? part.substr(0, eq) : part;
        var val = eq > -1 ? decodeURIComponent(part.substr(eq + 1)) : "";
        var from = key.indexOf("[");
        if (from == -1) {
          key = decodeURIComponent(key);
          if (result[key] === undefined) {
            result[key] = val;
          } else if (Array.isArray(result[key])) {
            result[key].push(val);
          } else {
            let arr = [ result[key], val ];
            result[key] = arr;
          }
        } else {
          var to = key.indexOf("]");
          var index = decodeURIComponent(key.substring(from + 1, to));
          key = decodeURIComponent(key.substring(0, from));
          if (!result[key])
            result[key] = [];
          if (!index)
            result[key].push(val);
          else
            result[key][index] = val;
        }
      }
    });
    return result;
  }

  getRequestParams() {
    let origUrl = xdmp.getOriginalUrl();
    let result = {};

    try {
      if (origUrl.includes("?")) {
        result = this.getJsonFromUrl(origUrl.replace(/^[^\?]*\?/, ''));
      }
    }
    catch (err) {
      throw new EndpointError(400, "Invalid parameters");
    }

    // Handle the magic property "payload"
    if (result.payload) {
      for (let key in result) {
        if (key != "payload") {
          throw new EndpointError(400, "Invalid parameters with payload");
        }
      }

      try {
        return fn.head(xdmp.unquote(result.payload, null, ["format-json"])).toObject();
      } catch (err) {
        throw new EndpointError(400, "Parse error in payload parameter");
      }
    }

    return result;
  }

  static LOG(msg) {
    return xdmp.log('OpsDirector error: ' + msg + " (in " + xdmp.getRequestPath() + ")");
  }

  static securityAssert() {
    xdmp.securityAssert("http://marklogic.com/xdmp/privileges/opsdir-user",
                        "execute");
  }

  GET() {
    Endpoint.LOG("GET() not implemented");
    xdmp.setResponseCode(404, 'Not Found');
  }

  PUT() {
    Endpoint.LOG("PUT() not implemented");
    xdmp.setResponseCode(404, 'Not Found');
  }

  POST() {
    Endpoint.LOG("POST() not implemented");
    xdmp.setResponseCode(404, 'Not Found');
  }

  DELETE() {
    Endpoint.LOG("DELETE() not implemented");
    xdmp.setResponseCode(404, 'Not Found');
  }

  MISSING(reqMethod) {
    Endpoint.LOG(reqMethod + "() not implemented");
    xdmp.setResponseCode(404, 'Not Found');
  }

  static respond(t) {
    // We have to let the session endpoint run w/o auth because that's how we login
    if (xdmp.getRequestPath() == "/app/v2/endpoints/session.sjs"
        || xdmp.getRequestPath() == "/app/v2/endpoints/motd.sjs"
        || xdmp.getRequestPath() == "/app/v2/endpoints/email.sjs") {
      // Nevermind, we don't want to do auth on these...
    } else {
      // This doesn't get caught by handleErrors so we have to do it specially
      try {
        Endpoint.securityAssert();
      } catch (error) {
        if (error.name !== "SEC-PRIV") {
          throw error;
        }
        xdmp.setResponseCode(401, "Unauthorized");
        return {
          "name": "EndpointError",
          "statusCode": 401,
          "message": "Unauthorized"
        };
      }
    }

    var reqMethod = xdmp.getRequestMethod();

    // Register endpoints with feature metrics
    var feature = xdmp.getRequestPath();
    if (feature.endsWith(".sjs")) {
      feature = feature.substring(0, feature.length - 4);
    }
    if (feature.startsWith("/app/v2/endpoints/")) {
      feature = feature.substring(15);
    }
    feature = feature.replace("/", ".");
    feature = feature.replace("-", "_");
    feature = "opsdir.v1." + reqMethod.toLowerCase() + "." + feature;
    let fid = xdmp.featureMetricRegister(feature);
    xdmp.featureMetricIncrement(fid);

    var i = new t();
    switch (reqMethod) {
    case 'GET':
      return Endpoint.handleErrors(() => i.GET());
    case 'PUT':
      return Endpoint.handleErrors(() => i.PUT());
    case 'POST':
      return Endpoint.handleErrors(() => i.POST());
    case 'DELETE':
      return Endpoint.handleErrors(() => i.DELETE());
    default:
      return Endpoint.handleErrors(() => i.MISSING(reqMethod));
    }
  }

  // Common filename munging
  csvFilename(filename, rtype) {
    if (filename) {
      return filename;
    }

    let dt = fn.currentDateTime().toString();
    if (dt.indexOf(".") > 0) {
      dt = dt.substring(0, dt.indexOf("."));
    }
    dt = dt.replace(/:/g, "-");

    return "opsdir-" + rtype + "-" + dt + ".csv";
  }

  // headers is an array of objects mapping property names to column names.
  //        [ { "property": "prop1", "header": "Header 1" },
  //          { "property": "prop2": "header": "Header 2" } ... ]
  // items is an array of objects; all of the properties identified
  //       in the fields object will be output as CSV.
  //
  formatCsv(headers, items) {
    let fields = [];
    if (headers) {
      fields = headers;
    } else {
      let item = items[0];
      for (let key in item) {
        fields.push({"header": key, "property": key});
      }
    }

    let acc = [];
    for (var index = 0; index < fields.length; index++) {
      let field = fields[index];
      if (field.label) {
        acc.push(field.label);
      } else if(field.header) {
        acc.push(field.header);
      } else {
        acc.push(field.name);
      }
    }
    let header = acc.map(field => `"${field.replace('"', '""')}"`).join(',') + '\r\n';

    let rows = items.map(item => fields.reduce((acc, field) => {
      let fieldInItem = item[field.property];
      if (fieldInItem !== undefined && fieldInItem !== null) {
        let value = fieldInItem;

        // What's the right answer here?
        if (xdmp.describe(fieldInItem).startsWith("fn.head")) {
          fieldInItem = fieldInItem.toObject();
        }

        if (Array.isArray(fieldInItem)) {
          value = fieldInItem.join(", ");
        } else if (typeof fieldInItem == "object") {
          value=fieldInItem.toString();
        }

        let fieldStr = value.replace ? `"${value.replace('"', '""')}"` : value;
        acc.push(fieldStr);
      } else {
        acc.push('""');
      }
      return acc;
    }, [])).join('\r\n') + '\r\n';
    return header + rows;
  }

  download(filename, headers, items) {
    if (! xdmp.traceEnabled("opsdir-debug-csv")) {
      xdmp.addResponseHeader('Content-Disposition',
                             'attachment; filename=' + filename);
    }
    return this.formatCsv(headers,items);
  }
}

exports.Endpoint = Endpoint;
exports.EndpointError = EndpointError;
