"use strict";

/* Copyright 2002-2018 MarkLogic Corporation. All Rights Reserved. */

const jsearch = require('/MarkLogic/jsearch');

const propertyTypes = {
  "created": "dateTime",
  "hostTs": "dateTime",
  "lastBackup": "dateTime",
  "lastStartup": "dateTime",
  "opsdirDtend": "dateTime",
  "opsdirDtstart": "dateTime",
  "timestamp": "dateTime",
  "ts": "dateTime",

  "activeFragmentCount": "unsignedLong",
  "databaseCount": "unsignedLong",
  "deletedFragmentCount": "unsignedLong",
  "diskSize": "unsignedLong",
  "diskSpace": "unsignedLong",
  "documentCount": "unsignedLong",
  "encryptedDiskSize": "unsignedLong",
  "encryptedSize": "unsignedLong",
  "fastDataMaxSize": "unsignedLong",
  "forestCount": "unsignedLong",
  "fragmentCount": "unsignedLong",
  "freeSpace": "unsignedLong",
  "groupCount": "unsignedLong",
  "hostCount": "unsignedLong",
  "largeDataSize": "unsignedLong",
  "period": "unsignedLong",
  "port": "unsignedLong",
  "serverCount": "unsignedLong",
  "size": "unsignedLong",
  "standCount": "unsignedLong"
};

function search(body, baseQuery) {
  // This function assumes that body has the standard structure:
  // {
  //   "filter": ["prop1", "prop2", ... ],
  //   "query": { "prop1": "value",
  //              "prop2": [ "value1", "value2", ... ] },
  //   "text": "textsearch",
  //   "order": "ascending", // or "descending"
  //   "sort": "sortProp",
  //   "start": first,
  //   "length": length
  // }

  let filter = undefined;
  if (body.filter) {
    if (Array.isArray(body.filter)) {
      filter = body.filter;
    } else {
      filter = [ body.filter ];
    }
  }

  let query = [ baseQuery ];
  if (Array.isArray(baseQuery)) {
    query = baseQuery;
  }

  if (body.query) {
    for (let qry of body.query) {
      for (let prop in qry) {
        let ctsq = undefined;
        let param = [ qry[prop] ];
        if (Array.isArray(qry[prop])) {
          param = qry[prop];
        }

        let values = [];
        for (let q of param) {
          if (q === Object(q)) {
            ctsq = cts.jsonPropertyRangeQuery(prop, q.op, typedValue(prop, q.value));
            query.push(ctsq);
          } else {
            values.push(typedValue(prop, q));
          }
        }

        if (values.length > 0) {
          if (prop in propertyTypes) {
            ctsq = cts.jsonPropertyValueQuery(prop, values);
          } else {
            // string
            ctsq = cts.jsonPropertyValueQuery(prop, values, "exact");
          }
          query.push(ctsq);
        }
      }
    }
  }

  let filterReq = false;
  // Wait, what if the baseQuery includes a fieldWordQuery or a wordQuery?
  if (baseQuery) {
    if (Array.isArray(baseQuery)) {
      for (let query of baseQuery) {
        filterReq = filterReq
          || (query instanceof cts.fieldWordQuery)
          || (query instanceof cts.wordQuery);
      }
    } else {
      filterReq = filterReq
        || (baseQuery instanceof cts.fieldWordQuery)
        || (baseQuery instanceof cts.wordQuery);
    }
  }

  // Deal with the text query case...
  if (body.text && body.text.trim().length) {
    filterReq = true;
    let textQuery = body.text.trim();
    if ((textQuery.length >= 3) && (textQuery.indexOf("*") < 0)) {
      textQuery = textQuery + "*";
    }
    let opts = ['case-insensitive', 'punctuation-insensitive'];

    if (body.fieldName) {
      query.push(cts.fieldWordQuery(body.fieldName, textQuery, opts));
    } else {
      query.push(cts.wordQuery(textQuery, opts));
    }
  }

  let order = body.order || "ascending";
  let sort  = undefined;
  if (body.sort) {
    try {
      sort = cts.jsonPropertyReference(body.sort, [ "type=string" ]);
    } catch (err) {
      try {
        sort = cts.jsonPropertyReference(body.sort, [ "type=unsignedLong" ]);
      } catch (err) {
        sort = cts.jsonPropertyReference(body.sort, [ "type=dateTime" ]);
      }
    }
  }

  // Note: cannot nest fn.trace inside where because it returns the wrong thing.
  let _ = fn.trace(cts.andQuery(query), "opsdir-search");

  if (xdmp.traceEnabled("opsdir-search-plan")) {
    let _ = fn.trace(cts.plan(cts.andQuery(query)), "opsdir-search-plan");
  }

  let docs = jsearch.documents().where(cts.andQuery(query));

  if (sort) {
    let _ = fn.trace(cts.indexOrder(sort, order), "opsdir-search");
    docs = docs.orderBy(cts.indexOrder(sort, order));
  }

  if (filterReq) {
    docs = docs.filter();
  }

  docs = docs.slice(body.start, body.start + body.length)
    .reduce((accResults, { uri, document }) =>
            { let obj = document.toObject();
              if (filter) {
                let props = filter.reduce((acc, field) => {
                  if (field in obj) {
                    acc[field] = obj[field];
                  }
                  return acc;
                }, { });
                accResults.push(props);
              } else {
                accResults.push(obj);
              }
              return accResults;
            }, []);

  let { results, estimate } = docs.result();

  return { estimate,
           "start": body.start,
           "count": results ? results.length : 0,
           "items": results || [] };
}

function typedValue(prop, value) {
  if (prop in propertyTypes) {
    switch (propertyTypes[prop]) {
    case "dateTime":
      return xs.dateTime(value);
    case "unsignedLong":
      return xs.unsignedLong(value);
    default:
      return value;
    }
  } else {
    return value;
  }
}

exports.search = search;
