xquery version "1.0-ml";

(: Copyright 2015-2018 MarkLogic Corporation.  All Rights Reserved. :)

module namespace a="http://marklogic.com/v1/opsdirector/alerts";

import module namespace t="http://marklogic.com/v1/opsdirector"
    at "/common/opsdir.xqy";

import module namespace p="http://marklogic.com/v1/opsdirector/polling"
    at "/common/polling.xqy";

import module namespace cc="http://marklogic.com/v1/opsdirector/camel-case"
    at "/common/camel-case.xqy";

import module namespace io="http://marklogic.com/v1/opsdirector/cluster-io"
    at "/common/cluster-io.xqy";

import module namespace n="http://marklogic.com/v1/opsdirector/notifications"
    at "/common/notifications.xqy";

declare default function namespace "http://www.w3.org/2005/xpath-functions";

declare namespace host="http://marklogic.com/xdmp/status/host";

declare option xdmp:mapping "false";

declare %private variable $a:ALERT-POLL := "/config/alert-poll.xml";
declare %private variable $a:TXN-NAME   := "opsdir-alert-poll";

declare function a:get-health-reports(
  $host as object-node()
) as object-node()?
{
  let $_    := trace("Getting health from " || $host/hostName, $t:trace-debug)
  let $reports := io:get-body($host/opsdirClusterId,
                              $host/hostName,
                              "/manage/v2/?view=health")
  let $_ := trace("Got " || count($reports/cluster-health-report)
                  || " reports from " || $host/hostName, $t:trace-debug-alerts)
  return
    if ($reports/error or $reports/errorResponse)
    then
      let $_ := trace("System alerts query failed: " || $host/hostName, $t:trace-status)
      return
        ()
    else
      $reports
};

declare function a:get-reports(
  $cluster as xs:unsignedLong?,
  $status as xs:string*
) as object-node()*
{
  let $collection := if (exists($cluster))
                     then "http://marklogic.com/opsdirector/v1/alerts/" || $cluster
                     else "http://marklogic.com/opsdirector/v1/alerts"
  for $report in collection($collection)/node()
  where empty($status) or $report/status = $status
  return
    $report
};

declare function a:open-reports(
  $cluster as xs:unsignedLong?
) as object-node()*
{
  let $collection := if (exists($cluster))
                     then "http://marklogic.com/opsdirector/v1/alerts/" || $cluster
                     else "http://marklogic.com/opsdirector/v1/alerts"
  for $report in collection($collection)/node()
  where $report/status != 'closed'
  return
    $report
};

declare function a:delete-reports(
  $cluster as xs:unsignedLong?
) as empty-sequence()
{
  let $directory := if (exists($cluster))
                    then "/alerts/" || $cluster || "/"
                    else "/alerts/"
  let $_ := trace("Deleting " || $directory, $t:trace-status)
  return
    xdmp:directory-delete($directory)
};

declare %private function a:alert-code(
  $alert as object-node()
) as xs:string
{
  (: An alert is uniquely identified by its cluster, state,
     resource-type, resource-name, and code. Note that its
     host is explicitly excluded because many hosts generate
     the same alerts. :)
  let $string := string-join(($alert/opsdirClusterId, $alert/state,
                              $alert/resourceType, string($alert/resourceId),
                              $alert/code, $alert/message), "|")
  return xdmp:sha1($string)
};

declare function a:patch-reports(
  $report-obj as object-node()*,
  $cluster-id as xs:string,
  $host-id as xs:string
) as object-node()*
{
  (: The health reports come from the Management API with a set of
     related resources. Keeping track of those relationships,
     translating from the random ID values in each report returned
     from the Management API to stable values in the OpsDirector
     database, is hard. Or it would be. Since the current UX doesn't
     use them, I'm just going to leave them out for the time being. :)
  let $cluster := t:managed-resource((), "clusters", $cluster-id)
  let $reports := for $report in $report-obj/cluster-health-report
                  let $report := cc:to-camel-case(xdmp:to-json($report)/node())
                  let $id     := a:alert-code($report)
                  return
                    map:new((
                      (map:entry("opsdirId", $id || "-" || xdmp:random()),
                       map:entry("opsdirResourceType", "alerts"),
                       map:entry("alertCode", $id),
                       map:entry("acknowledged", "no"),
                       map:entry("opsdirClusterId", $cluster-id),
                       map:entry("opsdirClusterName", string($cluster/opsdirName)),
                       map:entry("opsdirHostId", $host-id),
                       for $key in $report/*
                       return map:entry(
                         local-name-from-QName(node-name($key)),
                         $report/*[node-name(.) = node-name($key)]))))
  return
    ($reports ! xdmp:to-json(.))/object-node()
};

declare function a:update-reports(
  $report as object-node()*
)
{
  let $cluster-id := $report/cluster-id
  let $host-id    := $report/host-id
  let $timestamp  := if (exists($report/timestamp))
                     then string($report/timestamp)
                     else string(current-dateTime())
  where exists($report) and exists($cluster-id) and exists($host-id)
  return
    let $reports := a:patch-reports($report, $cluster-id, $host-id)
    let $abort   := for $report in $reports
                    where ($report/code = "HEALTH-HOST-ERROR"
                           and $report/resourceId = $host-id)
                    return
                      $report
    return
      if (exists($abort))
      then
        xdmp:log("Ignoring health reports from damaged host: " || $host-id)
      else
        let $_       := trace("Update " || count($reports)
                              || " reports for cluster " || $cluster-id,
                              $t:trace-debug-alerts)
(:
        let $_       := for $report in $reports
                        return trace($report/opsdirId || " " || $report/resourceName,
                                     $t:trace-debug-alerts)
:)
        let $coll  := ("http://marklogic.com/opsdirector/v1/alerts/" || $cluster-id,
                       "http://marklogic.com/opsdirector/v1/alerts")
        let $all   := collection($coll[1])/object-node()
        let $open  := a:close-reports($timestamp,
                                      $all[not(status = 'closed')],
                                      $reports)
        let $new   := a:new-reports($open, $reports)

        let $_     := trace("Store " || count($new) || " new reports",
                            $t:trace-debug-alerts)

        for $report in $new
        let $id  := string(xdmp:random())
        let $obj := object-node {
                      "uid": $id,
                      "ts": string(current-dateTime()),
                      "opsdirDtstart": $timestamp,
                      "status": "open"
                    } + $report

        let $uri   := "/alerts/" || $cluster-id || "/" || $id || ".json"
        let $alert := xdmp:to-json($obj)/node()
        
        (: Send email for the new alerts :)
        let $_ := try {
          n:email-notifications($alert)
        } catch ($e) {
          xdmp:log(("Failed to send email for alert:",
                    $alert, $e))
        }

        let $_   := trace("Ins: " || $uri, $t:trace-debug-uri)
        let $p   := if ($alert/resourceId)
                    then
                      let $r := t:managed-resource((), (), $alert/resourceId)
                      return
                        if (exists($r))
                        then
                          xdmp:document-get-permissions(xdmp:node-uri($r[1]))
                        else
                          ()
                    else
                      ()
        let $p   := if (empty($p)) then $t:DBPERMS else $p
        return
          xdmp:document-insert($uri, $alert, $p, $coll)
};

declare %private function a:close-reports(
  $timestamp as xs:string,
  $open as object-node()*,
  $reports as object-node()*
)
{
  let $_ := trace(("New reports: " || count($reports)), $t:trace-debug-alerts)
(:
  let $_ := for $report in $reports
            return trace($report/opsdirId || " " || $report/resourceName,
                         $t:trace-debug-alerts)
:)
  let $_ := trace(("Open reports: " || count($open)), $t:trace-debug-alerts)
(:
  let $_ := for $report in $open
            return trace($report/opsdirId || " " || $report/resourceName,
                         $t:trace-debug-alerts)
:)
  let $still-open := (for $report in $open
                      where $report/alertCode = $reports/alertCode
                            or $report[local = true()]
                      return
                        $report)

  let $_ := trace(("Still open:" || count($still-open)), $t:trace-debug-alerts)
(:
  let $_ := for $report in $still-open
            return trace($report/opsdirId || " " || $report/resourceName,
                         $t:trace-debug-alerts)
:)
  let $close := $open except $still-open
  let $_     := for $report in $close
                (: Oh for the love of ... this is a hack :)
                let $map := map:new(for $key in $report/* except ($report/status)
                                    return map:entry(
                                      local-name-from-QName(node-name($key)),
                                      $report/*[node-name(.) = node-name($key)]))
                let $duration := xs:dateTime($timestamp) - xs:dateTime($report/dtstart)
                let $obj := object-node { "opsdirDtend": $timestamp,
                                          "duration": string($duration),
                                          "status": "closed" }
                return
                  xdmp:node-replace($report, xdmp:to-json($map + $obj)/node())

  let $_     := trace("Close " || count($close) || " reports", $t:trace-debug-alerts)
  let $_     := trace("Still open " || count($still-open) || " reports", $t:trace-debug-alerts)
  return
    $still-open
};

declare %private function a:new-reports(
  $open as object-node()*,
  $reports as object-node()*
)
{
  for $report in $reports
  where not($report/alertCode = $open/alertCode)
  return
    $report
};

declare function a:poll-clusters(
)
{
  if (xdmp:trace-enabled($t:trace-alert-disable))
  then
    let $_ := trace("Health alerts disabled by trace event", $t:trace-alert-disable)
    return
      ()
  else
    p:poll-clusters("alerts", a:poll-cluster#1)
};

declare function a:poll-cluster(
  $host as object-node()
)
{
  let $reports := a:get-health-reports($host)
  let $_       := a:update-reports($reports)
  return
    ()
};
