xquery version "1.0-ml";

(: Copyright 2015-2018 MarkLogic Corporation.  All Rights Reserved. :)

module namespace io="http://marklogic.com/v1/opsdirector/cluster-io";

import module namespace t="http://marklogic.com/v1/opsdirector"
    at "/common/opsdir.xqy";

import module namespace cc="http://marklogic.com/v1/opsdirector/camel-case"
    at "/common/camel-case.xqy";

import module namespace sess="http://marklogic.com/v1/opsdirector/session"
    at "/common/session.xqy";

import module namespace config="http://marklogic.com/v1/opsdirector/config"
    at "/common/config.xqy";

declare default function namespace "http://www.w3.org/2005/xpath-functions";

declare namespace xdmp-http="xdmp:http";

declare option xdmp:mapping "false";

declare %private variable $io:timeout :=
   try {
         let $value := config:get-atomic-property("http-timeout", 300)
         return
           xs:unsignedLong($value)
       } catch ($e) {
         300
       };

declare function io:get-properties(
  $host as xs:string,
  $resource-type as xs:string,
  $id as xs:string
) as node()?
{
  io:get-properties((), $host, $resource-type, $id)
};

declare function io:get-properties(
  $cluster as xs:string?,
  $host as xs:string,
  $resource-type as xs:string,
  $id as xs:string
) as node()?
{
  let $rsrc := t:managed-resource($cluster, "hosts", $host)
  let $_    := if (empty($rsrc))
               then error((), "No such " || $resource-type || ": " || $id)
               else ()

  let $session := sess:get($rsrc/opsdirClusterId, $rsrc/opsdirId)
  let $json    := io:get-body($rsrc/opsdirClusterId, $rsrc/hostName,
                              "/manage/v2/" || $resource-type || "/" || $id || "/properties")
  where exists($json)
  return
    cc:to-camel-case($json)
};

declare function io:get(
  $cluster-id as xs:string,
  $host as xs:string,
  $path as xs:string
)
{
  io:get($cluster-id, $host, $path, "application/json")
};

declare function io:get(
  $cluster-id as xs:string,
  $host as xs:string,
  $path as xs:string,
  $accept as xs:string
)
{
  let $config  := doc("/config/cluster/"||$cluster-id||".json")/object-node()
  let $port    := $config/port

  let $internal   := doc-available("/certs/in/" || $cluster-id)

  (: Try to avoid the exception. In the internal case, opsdir-9999 will exist
     and will succeed. In the external case, if there's a distinct credential
     for each managed host, then the else branch will succeed. If it fails,
     then the only possible successful case left is that there is a single
     credential for every host because the hosts are behind a load balancer
     and SSL terminates at the load balancer.
  :)
  let $credential := try {
                       if ($internal)
                       then
                         xdmp:credential-id("opsdir-" || $cluster-id)
                       else
                         xdmp:credential-id("opsdir-" || $cluster-id || "-" || $host)
                      } catch ($e) {
                        if ($e/error:code eq "XDMP-NOCRED")
                        then
                          xdmp:credential-id("opsdir-" || $cluster-id)
                        else
                          xdmp:rethrow()
                      }
  let $uri     := "https://" || $host || ":" || $port || $path
  let $opts    := <options xmlns="xdmp:http">
                    <headers>
                      <accept>{$accept}</accept>
                    </headers>
                    <verify-cert>false</verify-cert>
                    <credential-id>
                      { $credential }
                    </credential-id>
                    <timeout>{$io:timeout}</timeout>
                  </options>
  return
    try {
      let $_      := trace("GET " || $uri, $t:trace-debug-configs)
      let $resp   := xdmp:http-get($uri, $opts)
      let $status := $resp[1]/xdmp-http:code/data()
      return
        if ($status lt 300)
        then
          $resp
        else
          let $_   := trace("OpsDirector: Got error response: " || $uri, $t:trace-status)
          let $_   := trace($resp, $t:trace-debug-io)
          (: Error message, error message, who can find the error message? :)
          let $msg := if ($resp[2]/*:html/*:body/*:h1)
                      then string($resp[2]/*:html/*:body/*:h1)
                      else
                        let $text := xdmp:quote($resp[2])
                        let $msg  := replace($text, "^.*<h1>(.*?)<.*$", "$1", "s")
                        return
                          if ($msg = "")
                          then ()
                          else $msg
          return
            ("ERROR",
             document {
               object-node {
                 "error": string($resp[1]/*:code),
                 "message": if ($msg) then $msg else xdmp:quote($resp[2]),
                 "detail": xdmp:quote($resp[2])
               }
             })
    } catch ($e) {
      let $_ := trace("OpsDirector: Failed to contact: " || $uri, $t:trace-status)
      let $_ := trace($e, $t:trace-debug-io)
      return
        ("ERROR",
         document {
           object-node {
             "error": string($e/error:code),
             "message": string($e/error:format-string)
           }
         })
    }
};

declare function io:get-body(
  $cluster-id as xs:string,
  $host as xs:string,
  $path as xs:string
)
{
  io:get-body($cluster-id, $host, $path, "application/json")
};

declare function io:get-body(
  $cluster-id as xs:string,
  $host as xs:string,
  $path as xs:string,
  $accept as xs:string
)
{
  io:get($cluster-id, $host, $path, $accept)[2]/node()
};

declare function io:head(
  $cluster-id as xs:string,
  $host as xs:string,
  $path as xs:string
) as xs:boolean
{
  let $config   := doc("/config/cluster/"||$cluster-id||".json")/object-node()
  let $port     := $config/port

  let $internal   := doc-available("/certs/in/" || $cluster-id)

  (: Try to avoid the exception. In the internal case, opsdir-9999 will exist
     and will succeed. In the external case, if there's a distinct credential
     for each managed host, then the else branch will succeed. If it fails,
     then the only possible successful case left is that there is a single
     credential for every host because the hosts are behind a load balancer
     and SSL terminates at the load balancer.
  :)
  let $credential := try {
                       if ($internal)
                       then
                         xdmp:credential-id("opsdir-" || $cluster-id)
                       else
                         xdmp:credential-id("opsdir-" || $cluster-id || "-" || $host)
                      } catch ($e) {
                        if ($e/error:code eq "XDMP-NOCRED")
                        then
                          xdmp:credential-id("opsdir-" || $cluster-id)
                        else
                          xdmp:rethrow()
                      }
  let $uri     := "https://" || $host || ":" || $port || $path
  let $opts    := <options xmlns="xdmp:http">
                    <verify-cert>false</verify-cert>
                    <credential-id>
                      { $credential }
                    </credential-id>
                    <timeout>{$io:timeout}</timeout>
                  </options>
  return
    try {
      let $_      := trace("HEAD " || $uri, $t:trace-debug-configs)
      let $resp   := xdmp:http-head($uri, $opts)
      let $status := $resp[1]/xdmp-http:code/data()
      return
        $status lt 400
    } catch ($e) {
      false()
    }
};

declare function io:post(
  $cluster-id as xs:string,
  $host as xs:string,
  $path as xs:string,
  $payload as object-node()
)
{
  io:post-or-put("POST", $cluster-id, $host, $path, $payload)
};

declare function io:put(
  $cluster-id as xs:string,
  $host as xs:string,
  $path as xs:string,
  $payload as object-node()
)
{
  io:post-or-put("PUT", $cluster-id, $host, $path, $payload)
};

declare function io:post-or-put(
  $method as xs:string,
  $cluster-id as xs:string,
  $host as xs:string,
  $path as xs:string,
  $payload as object-node()
)
{
  let $config   := doc("/config/cluster/"||$cluster-id||".json")/object-node()
  let $port     := $config/port

  let $internal   := doc-available("/certs/in/" || $cluster-id)

  (: Try to avoid the exception. In the internal case, opsdir-9999 will exist
     and will succeed. In the external case, if there's a distinct credential
     for each managed host, then the else branch will succeed. If it fails,
     then the only possible successful case left is that there is a single
     credential for every host because the hosts are behind a load balancer
     and SSL terminates at the load balancer.
  :)
  let $credential := try {
                       if ($internal)
                       then
                         xdmp:credential-id("opsdir-" || $cluster-id)
                       else
                         xdmp:credential-id("opsdir-" || $cluster-id || "-" || $host)
                      } catch ($e) {
                        if ($e/error:code eq "XDMP-NOCRED")
                        then
                          xdmp:credential-id("opsdir-" || $cluster-id)
                        else
                          xdmp:rethrow()
                      }
  let $uri     := "https://" || $host || ":" || $port || $path
  let $opts    := <options xmlns="xdmp:http">
                    <headers>
                      <accept>application/json</accept>
                      <content-type>application/json</content-type>
                    </headers>
                    <verify-cert>false</verify-cert>
                    <credential-id>
                      { $credential }
                    </credential-id>
                    <timeout>{$io:timeout}</timeout>
                  </options>
  return
    try {
      let $_      := trace($method || " " || $uri, $t:trace-debug-configs)
      let $resp   := if ($method = "POST")
                     then xdmp:http-post($uri, $opts, $payload)
                     else if ($method = "PUT")
                          then xdmp:http-put($uri, $opts, $payload)
                          else error((), "OD-BADMETHOD", "Unknown method: " || $method)
      let $status := $resp[1]/xdmp-http:code/data()
      return
        if ($status lt 300)
        then
          $resp
        else
          let $_ := trace("OpsDirector: Got error response: " || $uri, $t:trace-status)
          let $_ := trace($resp, $t:trace-debug-io)
          return
            $resp
    } catch ($e) {
      let $_ := trace("OpsDirector: Failed to contact: " || $uri, $t:trace-status)
      let $_ := trace($e, $t:trace-debug-io)
      return
        ("ERROR",
         document {
           object-node {
             "error": string($e/error:code),
             "message": string($e/error:format-string)
           }
         })
    }
};
