xquery version "1.0-ml";

(: Copyright 2015-2018 MarkLogic Corporation.  All Rights Reserved. :)

module namespace d="http://marklogic.com/v1/opsdirector/disconnected";

import module namespace t="http://marklogic.com/v1/opsdirector"
    at "/common/opsdir.xqy";

import module namespace p="http://marklogic.com/v1/opsdirector/polling"
    at "/common/polling.xqy";

import module namespace io="http://marklogic.com/v1/opsdirector/cluster-io"
    at "/common/cluster-io.xqy";

import module namespace sess="http://marklogic.com/v1/opsdirector/session"
    at "/common/session.xqy";

import module namespace config="http://marklogic.com/v1/opsdirector/config"
    at "/common/config.xqy";

import module namespace pki="http://marklogic.com/xdmp/pki"
    at "/MarkLogic/pki.xqy";

declare namespace host="http://marklogic.com/xdmp/status/host";
declare namespace x509="http://marklogic.com/xdmp/x509";

declare default function namespace "http://www.w3.org/2005/xpath-functions";

declare option xdmp:mapping "false";

declare %private variable $TXN-NAME := "opsdir-disconn-poll";
declare %private variable $TIME-LIMIT
  := xs:unsignedLong(config:get-atomic-property("disconn-time-limit", 300));

declare function d:check-hosts(
) as empty-sequence()
{
  if (p:host-to-run())
  then
    (: Are any of the hosts in our cluster new? Do we need to generate
       certificates for them so that managed clusters can talk to them? :)
    xdmp:invoke-function(
      function () {
        let $tmpl-name := "OpsDirector-SSL-Template"
        let $tmpl-id   := pki:get-template-by-name($tmpl-name)/pki:get-template-ids()
        for $host in xdmp:group-hosts() ! xdmp:host-name(.)
        let $cert := /pki:certificate[pki:template-id = $tmpl-id
                                      and pki:host-name = $host
                                      and pki:temporary = false()]
        return
          if (empty($cert))
          then
            let $_ := xdmp:log("Creating signed host certificate for " || $host)
            let $csr-pem :=
              pki:generate-certificate-request(
                pki:get-template-by-name($tmpl-name)/pki:template-id, $host, (), ())
            let $csr-xml := xdmp:x509-request-extract($csr-pem)
            let $cert-xml :=
              <cert xmlns="http://marklogic.com/xdmp/x509">
                <version>2</version>
                <serialNumber>{pki:integer-to-hex(xdmp:random())}</serialNumber>
                <validity>
                  <notBefore>{fn:current-dateTime()}</notBefore>
                  <notAfter>{fn:current-dateTime() + xs:dayTimeDuration("P3650D")}</notAfter>
                </validity>
                {$csr-xml/x509:subject}
                {$csr-xml/x509:publicKey}
                {$csr-xml/x509:v3ext}
              </cert>
            let $cert-pem :=
              xdmp:x509-certificate-generate(
                $cert-xml, (),
                <options xmlns="ssl:options">
                  <credential-id>{xdmp:credential-id($t:CRED-NAME)}</credential-id>
                </options>)
            return
              pki:insert-signed-certificates($cert-pem)
          else
            ()
      },
      <options xmlns="xdmp:eval">
        <database>{xdmp:security-database()}</database>
      </options>)
  else
    ()
};

declare function d:check-clusters(
) as empty-sequence()
{
  if (p:host-to-run())
  then
    let $concurrency := xs:unsignedLong(config:get-atomic-property("disconn-concurrency", 1))
    let $txnlist     := xdmp:transaction($TXN-NAME)
    return
      if (count($txnlist) ge $concurrency)
      then
        if (xdmp:trace-enabled($t:trace-disconnect))
        then
          let $_    := trace("Host " || xdmp:host-name()
                             || " disconn concurrency too high: "
                             || count($txnlist), $t:trace-disconnect)
          let $txns := xdmp:host-status(xdmp:host())/host:transactions
                          /host:transaction[host:transaction-name = $TXN-NAME]
          for $txn in $txns
          let $time := current-dateTime() - $txn/host:start-time
          let $_ := trace("Txn " || $txn/host:transaction-id
                          || " polling disconn for " || $time, $t:trace-disconnect)
          return
            ()
        else
          ()
      else
        d:check-clusters-on-this-host()
  else
    let $_ := trace("Not running check clusters on " || xdmp:host-name(), $t:trace-disconnect)
    return
      ()
};

declare %private function d:check-clusters-on-this-host(
) as empty-sequence()
{
  let $cids := t:managed-resource-ids("clusters")
  let $map  := map:map()
  let $_    := for $cid in $cids
               let $hids  := t:managed-resource-ids($cid, "hosts")
               let $stats := for $hid in $hids
                             let $uri := "/config/host-status/"
                                         ||$cid||"/"||$hid||"/snapshot.json"
                             let $obj := doc($uri)/object-node()
                             return
                               $obj/status
               return
                 if ("ok" = $stats)
                 then map:put($map, $cid, "ok")
                 else map:put($map, $cid, "err")

  let $_    := if (xdmp:trace-enabled($t:trace-disconnect))
               then
                 let $disc := for $cid in $cids
                              where map:get($map, $cid) != "ok"
                              return $cid
                 return
                   (trace("Cluster ids: " || string-join($cids, ", "), $t:trace-disconnect),
                    trace("Clusters that appear to be disconnected: "
                          || string-join($disc, ", "), $t:trace-disconnect),
                    trace("Check clusters on " || xdmp:host-name() || ": "
                          || count($disc) || " of " || count($cids)
                          || " clusters possibly disconnected", $t:trace-disconnect),
                    for $cid in $disc
                    let $cluster  := t:managed-resource($cid, "clusters", $cid)
                    let $hids     := t:managed-resource-ids($cid, "hosts")
                    for $hid in $hids
                    let $host := t:managed-resource($cid, "hosts", $hid)
                    let $uri := "/config/host-status/"
                                ||$cid||"/"||$hid||"/snapshot.json"
                    let $obj := doc($uri)/object-node()
                    return
                      trace("   " || $cluster/opsdirName || "/"
                            || $host/opsdirName || ": "
                            || $obj/status, $t:trace-disconnect))
               else
                 ()
  for $cid in $cids
  return
    if (map:get($map, $cid) = "ok")
    then
      d:check-reconnected($cid)
    else
      d:check-disconnected($cid)
};

declare %private function d:check-disconnected(
  $cid as xs:string
) as empty-sequence()
{
  let $hids := t:managed-resource-ids($cid, "hosts")
  let $disc := for $hid in $hids
               let $uri := "/config/host-status/"
                           ||$cid||"/"||$hid||"/snapshot.json"
               let $obj := doc($uri)/object-node()
               where exists($obj/status) and $obj/status != "ok"
               return
                 $hid
  let $_    := if (xdmp:trace-enabled($t:trace-disconnect))
               then
                 let $cluster := t:managed-resource($cid, "clusters", $cid)
                 return
                   trace("Cluster " || $cluster/opsdirName
                         || (if (count($hids) = count($disc))
                             then " is "
                             else " is not ") || "offline", $t:trace-disconnect)
               else
                 ()
  return
    if (count($hids) = count($disc))
    then
      d:mark-offline($cid)
    else
      ()
};

declare %private function d:check-reconnected(
  $cid as xs:string
) as empty-sequence()
{
  (: Is there an existing alert about the fact that this cluster is offline?
     There should never be more than one, but ... :)
  let $colls := ("http://marklogic.com/opsdirector/v1/alerts/" || $cid,
                 "http://marklogic.com/opsdirector/v1/alerts")
  let $qclus := cts:json-property-value-query("opsdirClusterId", $cid)
  let $qstat := cts:json-property-value-query("status", "open")
  let $qcode := cts:json-property-value-query("code", "OPSDIR-CLUSTER-HOSTS-UNREACHABLE")
  let $alert := cts:search(collection($colls[1]), cts:and-query(($qclus, $qstat, $qcode)))
  where exists($alert)
  return
    xdmp:spawn-function(
      function () {
        let $_ := xdmp:set-transaction-name($TXN-NAME)
        let $_ := d:reconnect-cluster($cid)
        return
          if (xdmp:trace-enabled($t:trace-disconnect))
          then
            let $cluster := t:managed-resource($cid, "clusters", $cid)
            let $txn  := xdmp:host-status(xdmp:host())/host:transactions
                           /host:transaction[host:transaction-id = xdmp:transaction()]
            return
              trace("Host " || xdmp:host-name()
                    || " reconnected "
                    || $cluster/opsdirName
                    || ": " || string(xdmp:elapsed-time()),
                    $t:trace-disconnect)
          else
            ()
      },
      <options xmlns="xdmp:eval">
        <time-limit>{$TIME-LIMIT}</time-limit>
        <update>true</update>
        <commit>auto</commit>
      </options>)
};

declare %private function d:reconnect-cluster(
  $cid as xs:string
) as empty-sequence()
{
  t:close-local-alert($cid, $cid, "OPSDIR-CLUSTER-HOSTS-UNREACHABLE")
};

declare function d:mark-offline(
  $cid as xs:string
) as empty-sequence()
{
  let $cluster := t:managed-resource($cid, "clusters", $cid)
  let $hids    := t:managed-resource-ids($cid, "hosts")
  let $_       := trace("Opening alert for disconnected cluster: "
                         || $cluster/opsdirName, $t:trace-disconnect)
  return
    t:open-local-alert($cid, $hids[1], "OPSDIR-CLUSTER-HOSTS-UNREACHABLE",
                       "unknown", "cluster", $cid, $cluster/opsdirName,
                       "all hosts in cluster are unreachable.")
};
