xquery version "1.0-ml";

(: Copyright 2015-2018 MarkLogic Corporation.  All Rights Reserved. :)

module namespace i="http://marklogic.com/v1/opsdirector/info";

import module namespace t="http://marklogic.com/v1/opsdirector"
    at "/common/opsdir.xqy";

import module namespace rs="http://marklogic.com/v1/opsdirector/resource-scopes"
    at "/common/rscopes.xqy";

import module namespace raw="http://marklogic.com/v1/opsdirector/rawdata"
    at "/common/rawdata.xqy";

import module namespace io="http://marklogic.com/v1/opsdirector/cluster-io"
    at "/common/cluster-io.xqy";

import module namespace p="http://marklogic.com/v1/opsdirector/polling"
    at "/common/polling.xqy";

import module namespace cc="http://marklogic.com/v1/opsdirector/camel-case"
    at "/common/camel-case.xqy";

import module namespace config="http://marklogic.com/v1/opsdirector/config"
    at "/common/config.xqy";

declare default function namespace "http://www.w3.org/2005/xpath-functions";

declare namespace host="http://marklogic.com/xdmp/status/host";

declare option xdmp:mapping "false";

declare %private variable $i:LOCK := "/config/joins.lock";
declare %private variable $i:INFO-POLL := "/config/info-poll.xml";
declare %private variable $i:TXN-NAME  := "opsdir-info-poll";

declare %private function i:get-info(
  $host as object-node()
) as empty-sequence()
{
  (: Force update access to be sequential; apparently there are overlapping
     read/write queries here that sometimes cause deadlock. FIXME: investigate further
     to see if the overlaps can be avoided.
  :)
  xdmp:lock-for-update($i:LOCK),

  let $time-limit := xs:unsignedLong(config:get-atomic-property("info-time-limit", 300))
  let $online     := i:host-online($host)
  let $groups     := if ($online) then i:get-groups($host) else map:map()
  return
    (: We used to do these functions individually, but that can lead to
       inconsistent results if some fail and some succeed; so now do them
       all at once :)
    if ($online and exists($groups))
    then
      let $_ := i:get-tasks($groups, $host)
      let $_ := i:get-host-status($groups, $host)
      let $_ := i:get-server-status($groups, $host)
      let $_ := i:get-local-cluster-status($groups, $host)
      let $_ := i:get-security($host)
      let $_ := i:get-database-counts($host)
      let $_ := i:get-database-status($host)
      let $_ := i:get-forest-counts($host)
      let $_ := i:get-forest-status($host)
      let $_ := i:get-forest-storage($host)
      let $_ := i:get-roles($host)
      let $_ := i:get-privileges($host)
      return
        ()
    else
      let $_ := trace("Apparently offline for info: " || $host/opsdirName, $t:trace-polling)
      return
        ()
};

declare %private function i:host-online(
  $host as object-node()
) as xs:boolean
{
  let $name   := $host/hostName

  let $map    := map:map()
  let $_      := map:put($map, "status", "ok")
  let $_      := map:put($map, "message", "success")

  let $path   := "/manage/v2/hosts/" || $name || "?view=status"
  let $status := io:get-body($host/opsdirClusterId, $name, $path)
  let $_      := if (exists($status/error))
                 then
                   let $_ := trace("Marking host " || $host/opsdirName || " as offline",
                                   $t:trace-disconnect)
                   let $_ := map:put($map, "status", "error")
                   let $_ := map:put($map, "error", $status/error)
                   let $_ := map:put($map, "message", $status/message)
                   return
                     ()
                 else
                   if ($host/opsdirId != $status/host-status/id)
                   then
                     let $_ := trace("Marking host " || $host/opsdirName || " as wrongid",
                                     $t:trace-disconnect)
                     let $_ := map:put($map, "status", "wrongid")
                     let $_ := map:put($map, "error", "Host ID missmatch")
                     let $_ := map:put($map, "message",
                                       $name || " has unexpected host ID; expected "
                                       || $host/opsdirId || ", found "
                                       || $status/host-status/id)
                     return
                       ()
                   else
                     ()

  (: Is this cluster still talking to me? (This handles the AWS case where a
     restart can change the names of hosts in the cluster.) :)
  let $_      := if ($status/opsdirector-session-endpoint)
                 then i:check-session-uri($host, $status/opsdirector-session-endpoint)
                 else ()

  let $status := map:get($map, "status")
  let $uri    := "/config/host-status/" || $host/opsdirClusterId || "/" || $host/opsdirId
                 || "/" || "snapshot.json"
  let $obj    := doc($uri)/object-node()

  let $_      := map:put($map, "opsdirClusterId", string($host/opsdirClusterId))
  let $_      := map:put($map, "opsdirHostId", string($host/opsdirId))
  let $_      := map:put($map, "ts", string(current-dateTime()))

  (: Remove the ts property from the host object so that we don't
     have trouble with sorting by 'ts' when we go looking for
     the least-recently-polled host. :)
  (: Oh for the love of ... this is a hack :)
  let $host-map := map:new(for $key in $host/* except ($host/ts)
                           return map:entry(
                             local-name-from-QName(node-name($key)),
                             $host/*[node-name(.) = node-name($key)]))
  let $_      := map:put($map, "host", xdmp:to-json($host-map)/object-node())

  let $update := xdmp:to-json($map)/object-node()

  let $_      := if (empty($obj))
                 then
                   xdmp:document-insert($uri, $update, $t:DBPERMS, ())
                 else
                   xdmp:node-replace($obj, $update)
  return
    (if ($status eq "ok")
     then
       t:close-local-alert($host/opsdirClusterId, $host/opsdirId, "OPSDIR-GET-ALERTS")
     else
       let $state := if ($status = "error") then "critical" else "unknown"
       let $msg := map:get($map, "message")
       let $msg := if (contains($msg, "-- "))
                   then substring-after($msg, "-- ")
                   else $msg
       return
         t:open-local-alert($host/opsdirClusterId, $host/opsdirId,
                            "OPSDIR-GET-ALERTS", $state,
                            "host", $host/opsdirId, $host/opsdirName,
                            "Failed to contact host: " || $msg),
     $status eq "ok")
};

declare %private function i:check-session-uri(
  $host as object-node(),
  $uri as xs:string
) as empty-sequence()
{
  let $curhost := substring-after($uri, 'https://') => substring-before(':')
  let $odhosts := xdmp:hosts() ! xdmp:host-name(.)
  where not($curhost = $odhosts)
  return
    let $pfx  := substring-before($uri, "//")
    let $sfx  := substring-after($uri, "//")
    let $sess := $pfx || "//" || xdmp:host-name() || ":" || substring-after($sfx, ":")
    return
      let $json := object-node { "opsdirector-session-endpoint": string($sess) }
      let $_    := xdmp:log("Updating session endpoint for " || $host/hostName
                            || " in " || $host/opsdirClusterName)
      let $resp := io:put($host/opsdirClusterId, $host/hostName,
                          "/manage/v2/properties", $json)
      return
        if ((string($resp[1]) = "ERROR") or ($resp[1]/*:code != 201))
        then
          let $_ := trace("Failed to update session endpoint for " || $host/hostName,
                          $t:trace-status)
          return
            ()
        else
          ()
};

declare %private function i:get-groups(
  $host as object-node()
) as map:map?
{
  let $pclus := cts:json-property-value-query("opsdirClusterId", $host/opsdirClusterId)
  let $ptype := cts:json-property-value-query("opsdirResourceType", "servers")
  let $pname := cts:json-property-value-query("opsdirName", "SecureManage")

  let $list := io:get-body($host/opsdirClusterId, $host/hostName, "/manage/v2/groups")
  where ($list/group-default-list)
  return
    let $map  := map:map()
    let $_    := for $group in $list/group-default-list/list-items/list-item
                 let $id     := $group/idref
                 let $name   := $group/nameref
                 let $pgrp   := cts:json-property-value-query("opsdirGroupName", $name)
                 let $secmng := cts:search(collection($t:COLLROOT||"resource"),
                                    cts:and-query(($pclus, $ptype, $pname, $pgrp)))
                 return
                   (map:put($map, $name, $id),
                    if (empty($secmng))
                    then i:create-secure-manage($host, $name)
                    else ())
    return
      $map
};

declare %private function i:get-tasks(
  $group-map as map:map,
  $host as object-node()
) as empty-sequence()
{
  let $name := $host/hostName
  let $list := io:get-body($host/opsdirClusterId, $name, "/manage/v2/tasks")
  where exists($list/tasks-default-list)
  return
    let $_ := i:expire-deleted-tasks($host, $list)
    for $task in $list/tasks-default-list/list-items/list-item
    let $group-id := map:get($group-map, $task/groupnameref)
    let $task-id  := string($task/idref)

    let $obj      := io:get-body($host/opsdirClusterId, $name,
                                 "/manage/v2/tasks/" || $task-id
                                 || "/properties?group-id=" || $group-id)
    let $obj      := cc:to-camel-case($obj)
    where empty($obj/error)
    return
      (: Make the taskHost explicitly null so that we can query for it :)
      let $obj  := if ($obj/taskHost)
                   then $obj
                   else
                     let $map := map:new ((
                                    map:entry("taskHost", null-node { }),
                                    for $property in $obj/node()
                                    let $name := local-name-from-QName(
                                                    node-name($property))
                                    return
                                      map:entry($name, $property)
                                 ))
                     return
                       xdmp:to-json($map)/node()

      let $rsrc := raw:resource($host/opsdirClusterId, "tasks", $task-id)
      return
        if (exists($rsrc) and ($rsrc/taskEnabled = $obj/taskEnabled))
        then
          ()
        else
          i:store($host, "resource", ($group-id, string($task/groupnameref)),
                  $obj, "tasks", $task-id)
};

declare %private function i:expire-deleted-tasks(
  $host as object-node(),
  $list as object-node()
) as empty-sequence()
{
  let $cid     := $host/opsdirClusterId
  let $newlist := $list/tasks-default-list/list-items/list-item/idref
  for $task in cts:search(collection($t:COLLROOT||"resource"),
                 cts:and-query((
                 cts:json-property-value-query("opsdirResourceType", "tasks"),
                 cts:json-property-value-query("opsdirClusterId", $cid))))
               /object-node()
  return
    if (not($task/taskId = $newlist))
    then
      raw:archive-resource($task)
    else
      ()
};

declare %private function i:get-host-status(
  $group-map as map:map,
  $host as object-node()
) as empty-sequence()
{
  let $name   := $host/hostName
  let $status := io:get-body($host/opsdirClusterId, $name,
                             "/manage/v2/hosts/" || $name || "?view=status")
  let $status := cc:to-camel-case($status)
  where exists($status/hostStatus)
  return
    let $name       := string($status/hostStatus/name)
    let $group-name := string($status/hostStatus/relations/relationGroup
                                 [typeref="groups"]/relation/nameref)
    let $group-id   := string(map:get($group-map, $group-name))
    let $rsrc       := raw:resource($host/opsdirClusterId, "host-status",
                                    $host/opsdirId)
    let $statobj    := object-node { "statusProperties":
                                     $status/hostStatus/statusProperties }
    return
      if (exists($rsrc))
      then
        if ($status/hostStatus/id = $rsrc/opsdirId)
        then
          i:store($host, "stats", ($group-id, $group-name),
                  $statobj, "host-status", $host/opsdirId, false(), $name)
        else
          let $_ := trace("Not saving host-status for " || $name || "; host id mismatch",
                          $t:trace-status)
          return
            ()
      else
        i:store($host, "stats", ($group-id, $group-name),
                $statobj, "host-status", $host/opsdirId, false(), $name)
};

declare %private function i:get-server-status(
  $group-map as map:map,
  $host as object-node()
) as empty-sequence()
{
  for $server in raw:resource-ids($host/opsdirClusterId, "servers")
  let $sobj   := raw:resource((), "servers", $server)
  let $param  := "?group-id=" || $sobj/groupName || "&amp;view=status"

  let $status := io:get-body($host/opsdirClusterId, $host/hostName,
                             "/manage/v2/servers/" || $server || $param)
  let $status := cc:to-camel-case($status)
  where exists($status/serverStatus)
  return
    let $name       := string($status/serverStatus/name)
    let $group-name := string($status/serverStatus/relations/relationGroup
                                 [typeref="groups"]/relation/nameref)
    let $group-id   := string(map:get($group-map, $group-name))
    let $rsrc   := raw:resource($host/opsdirClusterId, "server-status", $server)
    return
      let $status := object-node
                        { "statusProperties": $status/serverStatus/statusProperties }
      return
        i:store($host, "stats", ($group-id, $group-name),
                $status, "server-status", $server, false(), $name)
};

declare %private function i:get-local-cluster-status(
  $group-map as map:map,
  $host as object-node()
) as empty-sequence()
{
  let $param  := "?view=status"
  let $status := io:get-body($host/opsdirClusterId, $host/hostName,
                             "/manage/v2" || $param)
  let $status := cc:to-camel-case($status)
  let $name   := string($status/localClusterStatus/name)
  let $id     := string($status/localClusterStatus/id)
  let $rsrc   := raw:resource($host/opsdirClusterId, "local-cluster-status", $id)
  where not($status/error)
  return
    let $status := object-node{ "statusProperties":
                                $status/localClusterStatus/statusRelations/hostsStatus }
    return
      i:store($host, "stats", (),
              $status, "local-cluster-status", $id, false(), $name)
};

(: ============================================================ :)

declare %private function i:create-secure-manage(
  $host as object-node(),
  $group-name as xs:string
) as empty-sequence()
{
  let $server := io:get($host/opsdirClusterId,
                        $host/hostName,
                        "/manage/v2/servers/SecureManage/properties?group-id=" || $host/group)
  return
    if ($server[1]/*:code = 200)
    then
      let $map  := map:new(
                      (map:entry("group-name", $group-name),
                       for $property in $server[2]/node()/node()
                       let $name := local-name-from-QName(node-name($property))
                       where $name != "group-name"
                       return
                         map:entry($name, $property)))
      let $json := xdmp:to-json($map)/node()
      let $_    := xdmp:log("New group detected; creating server SecureManage for " || $group-name
                            || " in " || $host/opsdirClusterName)
      let $resp := io:post($host/opsdirClusterId, $host/hostName,
                           "/manage/v2/servers", $json)
      return
        if ((string($resp[1]) = "ERROR") or ($resp[1]/*:code != 201))
        then
          let $_ := trace("Failed to create SecureManage for " || $group-name
                          || " on " || $host/hostName,
                          $t:trace-status)
          return
            ()
        else
          ()
    else
      let $_ := trace("Failed to get SecureManage properties from " || $host/hostName,
                      $t:trace-status)
      return
        ()
};

declare %private function i:get-database-counts(
  $host as object-node()
) as empty-sequence()
{
  for $db in raw:resource-ids($host/opsdirClusterId, "databases")
  let $counts := io:get-body($host/opsdirClusterId, $host/hostName,
                             "/manage/v2/databases/" || $db || "?view=counts")
  where empty($counts/error)
  return
    let $counts := cc:to-camel-case($counts)
    let $rsrc   := raw:resource($host/opsdirClusterId, "counts", $db)
    return
      if (exists($counts/databaseCounts/countProperties))
      then
        let $counts := object-node { "countProperties":
                                        $counts/databaseCounts/countProperties }
        return
          i:store($host, "stats", (), $counts, "counts", $db, false())
      else
        ()
};

declare %private function i:get-forest-counts(
  $host as object-node()
) as empty-sequence()
{
  for $id in raw:resource-ids($host/opsdirClusterId, "forests")
  let $counts := io:get-body($host/opsdirClusterId, $host/hostName,
                             "/manage/v2/forests/" || $id || "?view=counts")
  where empty($counts/error)
  return
    let $counts := cc:to-camel-case($counts)
    let $rsrc   := raw:resource($host/opsdirClusterId, "counts", $id)
    return
      (: Always update, even if not apparently changed, because we need
         to update the permissions if the database association has changed
         and there's no easy way to find that out from here.
      :)
      if (exists($counts/forestCounts/countProperties))
      then
        let $counts := object-node { "countProperties":
                                        $counts/forestCounts/countProperties }
        return
          i:store($host, "stats", (), $counts, "counts", $id, false())
      else
        ()
};

declare %private function i:get-security(
  $host as object-node()
) as empty-sequence()
{
  let $status := io:get-body($host/opsdirClusterId, $host/hostName,
                             "/manage/v2/security/properties")
  let $status := cc:to-camel-case($status)
  where exists($status/keystore)
  return
    let $rsrc := raw:resource($host/opsdirClusterId, "security", $host/opsdirClusterId)
    return
      i:store($host, "resource", (), $status, "security", $host/opsdirClusterId, false())
};

declare %private function i:get-database-status(
  $host as object-node()
) as empty-sequence()
{
 for $db in raw:resource-ids($host/opsdirClusterId, "databases")
 let $status := io:get-body($host/opsdirClusterId, $host/hostName,
                            "/manage/v2/databases/" || $db || "?view=status")
 let $status := cc:to-camel-case($status)
 where exists($status/databaseStatus)
 return
   let $name := string($status/databaseStatus/name)
   let $rsrc := raw:resource($host/opsdirClusterId, "database-status", $db)
   return
     let $status := object-node
                       { "statusProperties": $status/databaseStatus/statusProperties }
     return
       i:store($host, "stats", (), $status, "database-status", $db, false(),$name)
};

declare %private function i:get-forest-status(
  $host as object-node()
) as empty-sequence()
{
  for $forest in raw:resource-ids($host/opsdirClusterId, "forests")
  let $status := io:get-body($host/opsdirClusterId, $host/hostName,
                             "/manage/v2/forests/" || $forest || "?view=status")
  let $status := cc:to-camel-case($status)
  where exists($status/forestStatus)
  return
    let $name   := string($status/forestStatus/name)
    let $rsrc   := raw:resource($host/opsdirClusterId, "forest-status", $forest)
    return
      let $status := object-node
                        { "statusProperties": $status/forestStatus/statusProperties }
      return
        i:store($host, "stats", (), $status, "forest-status", $forest, false(), $name)
};

declare %private function i:get-forest-storage(
  $host as object-node()
) as empty-sequence()
{
  for $id in raw:resource-ids($host/opsdirClusterId, "forests")
  let $status := io:get-body($host/opsdirClusterId, $host/hostName,
                             "/manage/v2/forests/" || $id || "?view=storage")
  let $status := cc:to-camel-case($status)
  where exists($status/forestStorageList)
  return
    let $rsrc := raw:resource($host/opsdirClusterId, "forest-storage",
                                     $host/opsdirId)
    let $status := object-node
                     { "storageListItems": $status/forestStorageList/storageListItems }
    return
      i:store($host, "stats", (), $status, "forest-storage", $id, false())
};

declare %private function i:get-roles(
  $host as object-node()
) as empty-sequence()
{
  let $_    := trace("Read role list from " || $host/hostName, $t:trace-debug)
  let $list := io:get-body($host/opsdirClusterId, $host/hostName, "/manage/v2/roles")
  where exists($list/role-default-list)
  return
    let $obj  := object-node {
                   "roles": array-node {
                     for $role in $list/role-default-list/list-items/list-item
                     return
                       object-node {
                         "id": $role/idref,
                         "name": $role/nameref,
                         "description": $role/description
                       }
                   }
                 }
    let $rsrc := raw:resource($host/opsdirClusterId, "roles", $host/opsdirClusterId)
    return
      if (deep-equal($obj/roles, $rsrc/roles))
      then ()
      else i:store($host, "role-list", (), $obj, "roles", $host/opsdirClusterId)
};

declare %private function i:get-privileges(
  $host as object-node()
) as empty-sequence()
{
  let $list := io:get-body($host/opsdirClusterId, $host/hostName, "/manage/v2/privileges")
  where ($list/privilege-default-list)
  return
    let $obj  := object-node {
                   "privileges": array-node {
                     for $priv in $list/privilege-default-list/list-items/list-item
                     where $priv/kind = "execute"
                     return
                       object-node {
                         "id": $priv/idref,
                         "name": $priv/nameref,
                         "actin": $priv/action,
                          "kind": $priv/kind
                       }
                   }
                 }
    let $rsrc := raw:resource($host/opsdirClusterId, "privileges",
                                    $host/opsdirClusterId)
    return
      if (deep-equal($obj/privileges, $rsrc/privileges))
      then ()
      else i:store($host, "privilege-list", (), $obj, "privileges", $host/opsdirClusterId)
};

(: ============================================================ :)

declare %private function i:store(
  $host as object-node(),
  $kind as xs:string,
  $group-info as xs:string*,
  $obj as object-node(),
  $type as xs:string,
  $id as xs:string?
) as empty-sequence()
{
  i:store($host, $kind, $group-info, $obj, $type, $id, true(), ())
};

declare %private function i:store(
  $host as object-node(),
  $kind as xs:string,
  $group-info as xs:string*,
  $obj as object-node(),
  $type as xs:string,
  $id as xs:string?,
  $timestamped as xs:boolean
) as empty-sequence()
{
  i:store($host, $kind, $group-info, $obj, $type, $id, $timestamped, ())
};

declare %private function i:store(
  $host as object-node(),
  $kind as xs:string,
  $group-info as xs:string*,
  $obj as object-node(),
  $type as xs:string,
  $id as xs:string?,
  $timestamped as xs:boolean,
  $resource-name as xs:string?
) as empty-sequence()
{
  let $uri  := "/" || $kind || "/" || $type || "/" || $host/opsdirClusterId || "/"
               || (if (exists($id)) then $id || "/" else "")
               || (if ($timestamped) then t:timestamp() else "snapshot")
               || ".json"
  let $rsrc := raw:resource($host/opsdirClusterId, $type, $id)
  let $rel  := if ($type = "forest-status" or $type = "forest-storage")
               then
                 raw:resource($host/opsdirClusterId, "forests", $id)
               else
                 if ($type = "counts")
                 then
                   (: frack, no easy way to tell if its a forest count :)
                   if ($id and $id = raw:resource-ids((), "forests"))
                   then
                     raw:resource($host/opsdirClusterId, "forests", $id)
                   else
                     $rsrc
                 else
                   $rsrc

  (: this is a bit odd; stats can be in their own collection, but
     the others all need to be resources. maybe we should do better? :)
  let $coll := if ($kind = "stats")
               then $t:COLLROOT || $kind
               else $t:COLLROOT || "resource"

  let $map  := map:new(
                 (map:entry("opsdirKind", $kind),
                  map:entry("opsdirClusterId", $host/opsdirClusterId),
                  map:entry("opsdirClusterName", $host/opsdirClusterName),
                  map:entry("opsdirHostId", $host/opsdirId),
                  map:entry("opsdirHostName", $host/hostName),
                  map:entry("opsdirResourceName", $resource-name),
                  if (empty($group-info))
                  then ()
                  else (map:entry("opsdirGroupId", $group-info[1]),
                        map:entry("opsdirGroupName", $group-info[2]))))
  let $json := t:add-opsdir-properties($obj, $rsrc, $type, $id, $map)

  (: Make sure we preserve permissions...consider resource group membership :)
  let $perms := if (exists($rel))
                then xdmp:document-get-permissions(xdmp:node-uri($rel))
                else if (doc-available($uri))
                     then xdmp:document-get-permissions($uri)
                     else $t:DBPERMS

  (: The host status files contain the license information, so make sure
     that opsdir-license-admin can read them :)
  let $perms := if ($type = "host-status")
                then ($perms, xdmp:permission("opsdir-license-admin", "read"))
                else $perms

  (: The task files have to have the permissions of the corresponding host :)
  let $perms := if ($type = "tasks")
                then
                  let $cid      := $json/opsdirClusterId
                  let $gname    := $json/opsdirGroupName
                  let $qtype    := cts:json-property-value-query("opsdirResourceType", "hosts")
                  let $qclus    := cts:json-property-value-query("opsdirClusterId", $cid)
                  let $qgrp     := cts:json-property-value-query("opsdirGroupName", $gname)
                  let $hosts    := cts:search(collection($t:COLLROOT||"resource"),
                                      cts:and-query(($qtype, $qclus, $qgrp)))/object-node()
                  return
                    $t:DBPERMS (: FIXME: THIS IS WRONG :)
                else
                  $perms

  let $_    := trace("Store: " || $uri, $t:trace-info-uri)
  let $_    := xdmp:document-insert($uri, $json, $perms, $coll)

  return
    if (empty($rsrc))
    then
      ()
    else
      if ($timestamped)
      then
        raw:archive-resource($rsrc)
      else
        ()
};

declare function i:get-logs(
  $cluster-id as xs:string,
  $host-id as xs:string,
  $filename as xs:string,
  $start as xs:string?,
  $end as xs:string?,
  $regex as xs:string?,
  $regex-flags as xs:string?
)
{
  xdmp:security-assert("http://marklogic.com/xdmp/privileges/opsdir-user",
                       "execute"),

  let $rsrc := raw:resource($cluster-id, "hosts", $host-id)
  let $hostname := $rsrc/hostName
  return
    if (exists($hostname))
    then
      let $opts := ("filename="||$filename,
                    (if (exists($start)) then "start="||$start else ()),
                    (if (exists($end)) then "end="||$end else ()),
                    (if (exists($regex)) then "regex="||$regex else ()),
                    (if (exists($regex-flags)) then "regex-flags="||$regex-flags else ()),
                    "format=json")
      let $path := "/manage/v2/logs?" || string-join($opts, "&amp;")
      let $body := io:get-body($cluster-id, $hostname, $path)
      let $now  := string(current-dateTime())
      where exists($body/logfile)
      return
        array-node {
          (for $node in $body/logfile/log
           let $data := $host-id || "/" || $now || "/" || $node/message
           let $hash := xdmp:hash64($data)
           return
             object-node {
                "opsdirClusterId": $cluster-id,
                "opsdirHostId": $host-id,
                "hostTs": string($node/timestamp),
                "ts": $now,
                "level": string($node/level),
                "uid": string($hash),
                "message": string($node/message)
             })
        }
    else
      let $_ := trace("No hostName get-logs for " || $cluster-id || "/" || $host-id,
                      $t:trace-status)
      return
        ()
};

declare function i:get-logs-text(
  $cluster-id as xs:string,
  $host-id as xs:string,
  $filename as xs:string,
  $start as xs:string?,
  $end as xs:string?,
  $regex as xs:string?,
  $regex-flags as xs:string?
)
{
  xdmp:security-assert("http://marklogic.com/xdmp/privileges/opsdir-user",
                       "execute"),

  let $rsrc := raw:resource($cluster-id, "hosts", $host-id)
  let $hostname := $rsrc/hostName
  return
    if (exists($hostname))
    then
      let $start := if ($start castable as xs:dayTimeDuration)
                    then current-dateTime() + xs:dayTimeDuration($start)
                    else $start
      let $end   := if ($end castable as xs:dayTimeDuration)
                    then current-dateTime() + xs:dayTimeDuration($end)
                    else $end
      let $opts := ("filename="||$filename,
                    (if (exists($start)) then "start="||$start else ()),
                    (if (exists($end)) then "end="||$end else ()),
                    (if (exists($regex)) then "regex="||$regex else ()),
                    (if (exists($regex-flags)) then "regex-flags="||$regex-flags else ()))
      let $path := "/manage/v2/logs?" || string-join($opts, "&amp;")
      let $body := io:get-body($cluster-id, $hostname, $path, "text/plain")
      return
        if ($body/error)
        then
          error((), "OD-LOGERR", $body/message)
        else
          $body
    else
      error((), "OD-NOTMNGD", ("The host '" || $host-id
                               || "' in cluster '" || $cluster-id || "' "
                               || "is not managed by OpsDirector at this time."))
};

declare function i:poll-clusters(
)
{
  p:poll-clusters("info", i:poll-cluster#1)
};

declare function i:poll-cluster(
  $host as object-node()
) as empty-sequence()
{
  i:get-info($host)
};
