xquery version "1.0-ml";

(: Copyright 2015-2018 MarkLogic Corporation.  All Rights Reserved. :)

(: This library contains convenience functions for things that are not
   easily achieved in JavaScript but that are bleeding dead simple
   in XQuery. [expletive deleted.]

   N.B. Some are amp'.
:)

module namespace jsl="http://marklogic.com/v1/opsdirector/jslib";

import module namespace admin="http://marklogic.com/xdmp/admin"
    at "/MarkLogic/admin.xqy";

import module namespace manage="http://marklogic.com/manage"
    at "/MarkLogic/manage/manage.xqy";

import module namespace config="http://marklogic.com/v1/opsdirector/config"
    at "/common/config.xqy";

import module namespace upgrade="http://marklogic.com/v1/opsdirector/upgrade"
    at "/common/upgrade.xqy";

declare default function namespace "http://www.w3.org/2005/xpath-functions";

declare namespace host="http://marklogic.com/xdmp/status/host";
declare namespace meter="http://marklogic.com/manage/meters";

declare option xdmp:mapping "false";

declare function jsl:dt-diff(
  $now as xs:string,
  $then as xs:string
) as xs:string?
{
  if ($now castable as xs:dateTime
      and $then castable as xs:dateTime)
  then
    let $duration := xs:dateTime($now) - xs:dateTime($then)
    let $days     := days-from-duration($duration)
    let $hours    := hours-from-duration($duration)
    let $minutes  := if (minutes-from-duration($duration) lt 10)
                     then "0" || minutes-from-duration($duration)
                     else minutes-from-duration($duration)
    let $label    := if ($days eq 1) then " day " else " days "
    return
      $days || $label || $hours || ":" || $minutes
  else
    ()
};

declare function jsl:gzip-reply(
  $results as object-node()
)
{
  xdmp:security-assert("http://marklogic.com/xdmp/privileges/opsdir-user",
                       "execute"),

  let $encodings := tokenize(xdmp:get-request-header("accept-encoding"), ",\s*")
  return
    if ("gzip" = $encodings)
    then
      (xdmp:set-response-content-type("application/json"),
       xdmp:add-response-header("Content-encoding", "gzip"),
       xdmp:gzip($results))
    else
      $results
};

declare function jsl:gzip-array-reply(
  $results as array-node()
)
{
  xdmp:security-assert("http://marklogic.com/xdmp/privileges/opsdir-user",
                       "execute"),

  let $encodings := tokenize(xdmp:get-request-header("accept-encoding"), ",\s*")
  return
    if ("gzip" = $encodings)
    then
      (xdmp:set-response-content-type("application/json"),
       xdmp:add-response-header("Content-encoding", "gzip"),
       xdmp:gzip($results))
    else
      $results
};

declare function jsl:server-authentication(
) as xs:string
{
  let $config := admin:get-configuration()
  return
    admin:appserver-get-authentication($config, xdmp:server())
};

declare function jsl:roles-and-subroles(
  $role-names as xs:string*
) as xs:string*
{
  xdmp:security-assert("http://marklogic.com/xdmp/privileges/opsdir-user",
                       "execute"),

  let $ids := for $name in $role-names
              let $asid := if ($name castable as xs:unsignedLong)
                           then
                             let $id := $name cast as xs:unsignedLong
                             return
                               try {
                                 (: is it really a role id? :)
                                 let $_ := xdmp:role-name($id)
                                 return
                                   $id
                               } catch ($e) {
                                 ()
                               }
                           else
                             ()
              let $ids  := if ($asid)
                           then jsl:roles-and-subroles-by-id($asid)
                           else ()
              return
                if ($ids)
                then $ids
                else (xdmp:role($name), xdmp:role-roles($name))
  return
    distinct-values($ids) ! string(.)
};

declare %private function jsl:roles-and-subroles-by-id(
  $roleid as xs:unsignedLong
) as xs:unsignedLong*
{
  let $name := xdmp:role-name($roleid)
  return
    ($roleid, xdmp:role-roles($name))
};

declare function jsl:client-session-status(
) as object-node()
{
  let $config := admin:get-configuration()
  let $hostid := string(xdmp:host())
  let $map    := map:map()

  (: If the user isn't an admin, then this will fail. Compenstate as best we can. :)
  let $_      := try {
                   for $hid in admin:get-host-ids($config)
                   let $status := xdmp:host-status($hid)
                   let $info := object-node {
                                  "version": string($status/host:version),
                                  "effectiveVersion": string($status/host:effective-version),
                                  "softwareVersion": string($status/host:software-version),
                                  "hostName": string($status/host:host-name)
                                }
                   return
                     map:put($map, string($hid), $info)
                 } catch ($e) {
                   if ($e/error:code = "SEC-PRIV")
                   then ()
                   else xdmp:rethrow()
                 }
  let $allVersions    := map:keys($map) ! map:get($map, .)/softwareVersion/string()
  let $clusterVersion := max($allVersions)

  let $hostVersion    := string(xdmp:effective-version())
  let $upgradeInProg  := xdmp:cluster-effective-version() lt 9000500
                         or (count(distinct-values($allVersions)) gt 1)

  let $updateInProg   := upgrade:in-progress()
  let $hostsToUpgrade := for $hid in map:keys($map)
                         where map:get($map,$hid)/softwareVersion != $clusterVersion
                         return
                           map:get($map, $hid)/hostName
  let $statusMap      := map:new((
                           map:entry("version",
                                     (config:get-atomic-property("version"), "2.0.0")[1]),
                           map:entry("auth", jsl:server-authentication()),
                           map:entry("opsDirVersion",
                                     (config:get-atomic-property("version"), "1.0-1")[1]),
                           map:entry("hostVersion", $hostVersion),
                           map:entry("clusterVersion", $clusterVersion),
                           map:entry("upgradeInProgress", $upgradeInProg),
                           map:entry("updateInProgress", $updateInProg),
                           if ($upgradeInProg)
                           then map:entry("hostsToUpgrade", array-node { $hostsToUpgrade })
                           else ()
                         ))
  let $_              := if (xdmp:trace-enabled("opsdir-debug-upgrade"))
                         then map:put($statusMap, "hostInfo", $map)
                         else ()
  return
    if (doc-available("/debug/client-session-status.json"))
    then doc("/debug/client-session-status.json")/object-node()
    else xdmp:to-json($statusMap)/object-node()
};

declare function jsl:resource-metrics(
    $resource-type as xs:string,
    $resources,
    $period as xs:string?,
    $start as xs:dateTime?,
    $end as xs:dateTime?,
    $summary as xs:boolean?,
    $detail as xs:boolean?,
    $metrics as xs:string*,
    $format as xs:string*,
    $meters-database as xs:string?,
    $meters-collection as xs:string?,
    $cluster-id,
    $host-id
)
{
  xdmp:security-assert("http://marklogic.com/xdmp/privileges/opsdir-user",
                       "execute"),

  manage:resource-metrics($resource-type,
    $resources,
    $period,
    $start,
    $end,
    $summary,
    $detail,
    $metrics,
    $format,
    $meters-database,
    $meters-collection,
    $cluster-id,
    $host-id)
};

declare function jsl:resource-metrics-periods(
    $start as xs:dateTime?,
    $end as xs:dateTime?
)
{
  xdmp:security-assert("http://marklogic.com/xdmp/privileges/opsdir-user",
                       "execute"),

let $cluster-ids := cts:element-values(xs:QName("meter:cluster-id"),(),(),
    cts:and-query((
        cts:collection-query("http://marklogic.com/opsdirector/v1/meters"),
        cts:element-range-query(xs:QName("meter:end-time"), ">=",$start),
        cts:element-range-query(xs:QName("meter:start-time"), "<", $end))))

return array-node{
   for $cluster-id in $cluster-ids
   let $periods := cts:element-values(xs:QName("meter:period"),(),(),
       cts:and-query((
           cts:collection-query("http://marklogic.com/opsdirector/v1/meters"),
           cts:element-range-query(xs:QName("meter:cluster-id"),"=",$cluster-id),
           cts:element-range-query(xs:QName("meter:end-time"), ">=",$start),
           cts:element-range-query(xs:QName("meter:start-time"), "<", $end))))
   return object-node{
       "cluster-id":$cluster-id,
       "available-periods": array-node{ $periods}}}
};

declare function jsl:resource-time-series(
    $resource as xs:string,
    $start as xs:dateTime,
    $end as xs:dateTime,
    $period as xs:string,
    $resource-type as xs:QName
)
{
    cts:time-series($resource,
        ($start, $end),
        ("document","concurrent"),
        cts:and-query((
          cts:collection-query("meters"),
          cts:element-range-query(xs:QName("meter:period"), "=",$period),
          cts:and-query((
              cts:element-range-query(xs:QName("meter:end-time"),">=", $start),
              cts:element-range-query(xs:QName("meter:end-time"),"<=", $end))),
              cts:element-query(xs:QName("meter:server-status"), cts:and-query(())))))
};
