xquery version "1.0-ml";

(: Copyright 2015-2018 MarkLogic Corporation.  All Rights Reserved. :)

module namespace n="http://marklogic.com/v1/opsdirector/notifications";

import module namespace t="http://marklogic.com/v1/opsdirector"
    at "/common/opsdir.xqy";

import module namespace rs="http://marklogic.com/v1/opsdirector/resource-scopes"
    at "/common/rscopes.xqy";

import module namespace config="http://marklogic.com/v1/opsdirector/config"
    at "/common/config.xqy";

declare option xdmp:mapping "false";

declare variable $n:NOTIFICATION_COLLECTION :=
   "http://marklogic.com/opsdirector/v1/notifications";

declare function n:email-notifications(
  $alert as object-node()
) as empty-sequence()
{
  let $notifications := n:notifications($alert)
  return
    if (empty($notifications))
    then
      let $_ := trace("None for "
                      || $alert/code || " on " || $alert/resourceName,
                      $t:trace-email-alerts)
      return
        ()
    else
      let $template := n:email-template($alert/state)
      for $notice in $notifications
      return
        n:email-notice($template, $notice, $alert)
};

declare function n:email-notice(
  $template as node(),
  $notice as object-node(),
  $alert as object-node()
) as empty-sequence()
{
  let $payload := n:formatted-email-notice($template, $notice, $alert)
  let $auth    := doc("/config/email-config.json")/email/authentication
  let $user    := $auth/username/string()
  let $pass    := $auth/password/string()
  let $options := <options xmlns="xdmp:email">
                    <verify-cert>false</verify-cert>
                    { if (exists($user) and exists($pass))
                      then
                        <authentication>
                          <username>{$user}</username>
                          <password>{$pass}</password>
                        </authentication>
                      else
                        ()
                    }
                  </options>
  let $xmlpayload := n:json-email-to-xml($payload)
  let $_ := trace(("Email payload:", $xmlpayload), $t:trace-email-alerts)
  return
    xdmp:email($xmlpayload, $options)
};

(: https://bugtrack.marklogic.com/50644 :)
declare function n:json-email-to-xml(
  $json as object-node()
) as element()
{
  <em:Message xmlns:em="URN:ietf:params:email-xml:"
              xmlns:rf="URN:ietf:params:rfc822:">
    { for $property in $json/node()
      let $name := local-name-from-QName(node-name($property))
      return
        switch ($name)
        case "reply-to"
        return
          <rf:reply-to>
            <em:Address>
              <em:name>{$property/name}</em:name>
              <em:adrs>{$property/address}</em:adrs>
            </em:Address>
          </rf:reply-to>
        case "from"
        return
          <rf:from>
            <em:Address>
              <em:name>{$property/name}</em:name>
              <em:adrs>{$property/address}</em:adrs>
            </em:Address>
          </rf:from>
        case "to"
        return
          <rf:to>
            { $property/addresses !
                <em:Address>
                  { if (./name) then <em:name>{./name}</em:name> else () }
                  <em:adrs>{./address}</em:adrs>
                </em:Address>
            }
          </rf:to>
        case "content"
        return
          <em:content>
            { if (contains($json/content-type, 'multipart/'))
              then
                let $newline := "&#13;&#10;"
                let $boundary := $json/content/boundary
                let $att := $json/content/parts[2]
                return
                  concat(
                    '--', $boundary, $newline,
                    $newline,
                    $json/content/parts[1],
                    $newline,
                    "--",$boundary,$newline,
                    "Content-Type:", $att/content-type, $newline,
                    "Content-Disposition: attachment; filename=",
                       $att/filename, $newline,
                    "Content-Transfer-Encoding: base64", $newline,
                    $newline,
                    $att/attachment, $newline,
                    "--",$boundary,"--", $newline)
              else
                replace($property, "\\n", "&#10;")
            }
          </em:content>
        default
        return
          element { fn:QName("URN:ietf:params:rfc822:", $name) }
                  { string($property) }
    }
  </em:Message>
};

declare function n:formatted-email-notice(
  $template as node(),
  $notice as object-node(),
  $alert as object-node()
) as object-node()
{
  let $payload := n:expand-template($template, $notice, $alert)
  let $payload := if ($notice/includeJson = "true")
                  then
                    n:email-attachment($payload, $alert)
                  else
                    $payload
  let $sendto  := tokenize($notice/email, "\s*,\s*")

  let $_       := trace((if (empty($sendto))
                         then "No recipients for alert: "
                         else "Send email for alert: ")
                        || $alert/code || " on " || $alert/resourceName,
                        $t:trace-email-alerts)

  let $map      := map:new(
                      for $property in $payload/node()
                      let $name := local-name-from-QName(node-name($property))
                      return
                        map:entry($name, $property)
                      )

  let $z    := replace(format-dateTime(current-dateTime(), "[Z]"), ":", "")
  let $date := format-dateTime(current-dateTime(),
                  "[FNn,*-3], [D01] [MNn,*-3] [Y0001] [H01]:[m01]:[s01] ") || $z

  let $_      := map:put($map, "date", $date)

  let $toaddr := map:new(
                   map:entry("addresses",
                             $sendto ! map:new(map:entry("address", .))))

  let $_      := map:put($map, "to", $toaddr)
  let $_      := map:put($map, "message-id",
                         "<opsdir-"
                         || xdmp:hmac-sha256("opsdir", string(xdmp:random()), "hex")
                         || "@marklogic.com>")
  return
    xdmp:to-json($map)/object-node()
};

declare %private function n:json-to-map(
  $template as node()
)
{
  if ($template/self::text())
  then
    string($template)
  else if ($template/self::object-node())
       then
         let $map := map:new(
           for $property in $template/node()
           let $name := local-name-from-QName(node-name($property))
           return
             map:entry($name,
                       n:json-to-map($property))
           )
         return
           $map
       else if ($template/self::array-node())
            then
              $template/node() ! n:json-to-map(.)
            else
              $template
};

declare function n:email-attachment(
  $payload as object-node(),
  $attachment as object-node()
) as object-node()
{
  let $boundary := "B" || xdmp:hmac-sha256("random", string(xdmp:random()), "hex")
  let $ctype    := "multipart/mixed; boundary=" || $boundary
  let $attbody  := xs:base64Binary(xdmp:base64-encode(xdmp:quote($attachment)))
  let $attfn    := substring(string(current-dateTime()), 1, 19)
                     ! replace(., ":", "")
  let $attfn    := "opsdir-notice-" || $attfn || ".json"
  let $att      := object-node {
                     "content-type": "application/json",
                     "content-disposition": "attachment",
                     "filename": $attfn,
                     "content-transfer-encoding":"base64",
                     "attachment": $attbody
                    }
  let $map      := map:new(
                      for $property in $payload/node()
                      let $name := local-name-from-QName(node-name($property))
                      where $name != "content" and $name != "content-type"
                      return
                        map:entry($name, $property)
                      )
  let $content  := $payload/content
  let $_        := map:put($map, "content-type", $ctype)
  let $_        := map:put($map, "content",
                           map:new((map:entry("boundary", $boundary),
                                    map:entry("parts", ($content, $att)))))
  return
    xdmp:to-json($map)/object-node()
};

declare function n:notifications(
  $alert as object-node()
) as object-node()*
{
  let $email-notifications
    := cts:search(collection($NOTIFICATION_COLLECTION),
                  cts:and-query((
                    cts:json-property-value-query("type", "email"),
                    cts:json-property-value-query("enabled", "true"))))
         /object-node()
  for $notice in $email-notifications
  let $match   := for $rgroup in $notice/scopes ! rs:resource-group(.)
                  where $rgroup/resource = $alert/resourceId
                  return
                    $rgroup
  where exists($match)
  return
    $notice
};

declare function n:email-template(
  $state as xs:string
) as object-node()
{
  (: Longer term, these should be configurable in the database... :)
  let $lines := ("Notice name: %%T.name%%",
                 "Alert Type: %%T.threshold%%",
                 "Timestamp: %%A.ts%%",
                 "Resource group: %%T.scopes%%",
                 "Host: %%A.host%%",
                 "Details: %%A.message%%")

  let $e-name := config:get-atomic-property("email-return-name",
                   "OpsDirector Notifications")
  let $e-addr := config:get-atomic-property("email-return-address",
                   "do-not-reply@marklogic.com")

  return
    object-node {
      "from": object-node {
          "name": $e-name,
          "address": $e-addr
      },
      "reply-to": object-node {
          "name": $e-name,
          "address": $e-addr
      },
      "subject": "%%T.name%% - %%T.threshold%% - %%A.resourceName%%",
      "content-type": "text/plain",
      "content": string-join($lines, "&#10;")
    }
};

(: N.B. This function returns a *map* because that's what xdmp:email expects :)
declare function n:expand-template(
  $template as node(),
  $notice as object-node(),
  $alert as object-node()
)
{
  if ($template/self::text())
  then
    string(n:expand-string($template, $notice, $alert))
  else if ($template/self::object-node())
       then
         let $map := map:new(
           for $property in $template/node()
           let $name := local-name-from-QName(node-name($property))
           return
             map:entry($name,
                       n:expand-template($property, $notice, $alert))
           )
         return
           xdmp:to-json($map)/object-node()
       else if ($template/self::array-node())
            then
              $template/node() ! n:expand-template(., $notice, $alert)
            else
              $template
};

declare %private function n:expand-string(
  $string as xs:string,
  $notice as object-node(),
  $alert as object-node()
) as xs:string
{
  if (contains($string, "%%"))
  then
    let $prefix := substring-before($string, "%%")
    let $rest   := substring-after($string, "%%")
    let $match  := n:substitute(substring-before($rest, "%%"), $notice, $alert)
    let $rest   := substring-after($rest, "%%")
    return
      $prefix || $match || n:expand-string($rest, $notice, $alert)
  else
    $string
};

declare %private function n:substitute(
  $match as xs:string,
  $notice as object-node(),
  $alert as object-node()
) as xs:string {
  let $name  := substring-after($match, ".")
  let $value := if (starts-with($match, "T."))
                then
                  string($notice/*[local-name(.) = $name])
                else if (starts-with($match, "A."))
                then
                  string($alert/*[local-name(.) = $name])
                else
                  "%%" || $match || "%%"
  return
    (: Handle a few special cases :)
    if ($match = "T.scopes")
    then
      string-join($notice/scopes ! rs:resource-group(.)/name, ", ")
    else if ($match = "A.host")
         then
           if ($alert/resourceType = "host")
           then
             string($alert/resourceName)
           else
             if ($alert/opsdirClusterId and $alert/opsdirHostId)
             then
               string(t:managed-resource($alert/opsdirClusterId,
                                         (), $alert/opsdirHostId)/opsdirName)
             else
               "UNKNOWN"
         else
           $value
};
