xquery version "1.0-ml";

(: Copyright 2018 MarkLogic Corporation.  All Rights Reserved. :)

module namespace rbac="http://marklogic.com/v1/opsdirector/rbac";

import module namespace sec="http://marklogic.com/xdmp/security"
    at "/MarkLogic/security.xqy";

import module namespace t="http://marklogic.com/v1/opsdirector"
    at "/common/opsdir.xqy";

import module namespace ts="http://marklogic.com/v1/opsdirector/security"
    at "/common/opsec.xqy";

import module namespace json="http://marklogic.com/xdmp/json"
    at "/MarkLogic/json/json.xqy";

import module namespace config="http://marklogic.com/v1/opsdirector/config"
    at "/common/config.xqy";

declare namespace jb="http://marklogic.com/xdmp/json/basic";

declare default function namespace "http://www.w3.org/2005/xpath-functions";

declare option xdmp:mapping "false";

declare function rbac:create-role(
  $name as xs:string,
  $subroles as xs:string*,
  $description as xs:string
) as object-node()
{
  xdmp:security-assert("http://marklogic.com/xdmp/privileges/opsdir-admin", "execute"),

  try {
    let $role := ts:create-role($name, $subroles, $description)
    return
      object-node {
        "name": $name,
        "subroles": array-node { $subroles },
        "description": $description
      }
  } catch ($e) {
    if ($e/error:code = "SEC-BADROLENAME")
    then
      error((), "OD-BADROLENAME", "Invalid role name: " || $name)
    else
      xdmp:rethrow()
  }
};

declare function rbac:list-roles(
  $regex as xs:string?
) as object-node()*
{
  (: Underlying function provides regex search, but we don't really expose
     regex search in Ops Director, just plain text search. So we escape
     an regex chars.
  :)
  (: Escape regex escape characters ... :)
  let $regex := replace($regex, '(\.|\[|\]|\\|\||\-|\^|\$|\?|\*|\+|\{|\}|\(|\))','\\$1')
  for $name in ts:get-role-names((), $regex)
  return
    if (rbac:exclude-role($name))
    then
      ()
    else
      rbac:get-role($name)
};

declare %private function rbac:exclude-role(
  $name as xs:string
) as xs:boolean
{
  let $exclude := config:get-atomic-property("exclude-roles")
  let $exclude := if (exists($exclude))
                  then tokenize(normalize-space($exclude), "\s+")
                  else ()
  return
    ends-with($name, '-internal')
    or starts-with($name, 'opsdir-rg-internal')
    or ($name = $exclude)
};

declare function rbac:get-role(
  $name as xs:string
) as object-node()?
{
  let $role := if ($name castable as xs:unsignedLong)
               then
                 try {
                   let $name := xdmp:role-name($name cast as xs:unsignedLong)
                   return
                     $name
                 } catch ($e) {
                   $name
                 }
               else
                 $name
  return
    try {
      let $id := xdmp:role($role)
      let $subroles := for $role in ts:get-role-subroles($role)
                       where not(rbac:exclude-role($role))
                       return $role
      return
        object-node {
          "name": $role,
          "description": ts:get-role-description($role),
          "subroles": array-node { $subroles }
        }
    } catch ($e) {
      xdmp:log(("Failed to get role: " || $name || " (" || $role || ")", $e))
    }
};

declare function rbac:update-role(
  $role as object-node()
) as empty-sequence()
{
  xdmp:security-assert("http://marklogic.com/xdmp/privileges/opsdir-admin", "execute"),

  let $_ := for $key in $role/* ! local-name-from-QName(node-name(.))
            where $key ne "name" and $key ne "description" and $key ne "subroles"
            return
              error((), "OD-BADROLE", ("Invalid role object: " || $key))
  let $_ := if (empty($role/name))
            then
              error((), "OD-BADROLE", ("Invalid role object"))
            else
              ()
  let $_ := if (not(ts:role-exists($role/name)))
            then
              error((), "OD-NOROLE", ("Role does not exist"))
            else
              ()

  (: Oh how I have come to loathe JSON in the server. If subroles is
     an empty array, then $role/subroles is false. Because reasons. :)
  let $json := xdmp:from-json($role)
  return
    (if (map:contains($json, "description"))
     then
       ts:set-role-description(string($role/name), string($role/description))
     else
       (),
     if (map:contains($json, "subroles"))
     then
       ts:set-role-subroles(string($role/name), $role/subroles ! string(.))
     else
       ())
};

declare function rbac:delete-role(
  $name as xs:string
) as empty-sequence()
{
  xdmp:security-assert("http://marklogic.com/xdmp/privileges/opsdir-admin", "execute"),

  let $role := if ($name castable as xs:unsignedLong)
               then
                 try {
                   let $name := xdmp:role-name($name cast as xs:unsignedLong)
                   return
                     $name
                 } catch ($e) {
                   $name
                 }
               else
                 $name
  return
    ts:remove-role($role)
};
