xquery version "1.0-ml";

(: Copyright 2014-2018 MarkLogic Corporation.  All Rights Reserved. :)

module namespace rg="http://marklogic.com/v1/opsdirector/resource-groups";

import module namespace sec="http://marklogic.com/xdmp/security"
    at "/MarkLogic/security.xqy";

import module namespace t="http://marklogic.com/v1/opsdirector"
    at "/common/opsdir.xqy";

import module namespace rs="http://marklogic.com/v1/opsdirector/resource-scopes"
    at "/common/rscopes.xqy";

import module namespace ts="http://marklogic.com/v1/opsdirector/security"
    at "/common/opsec.xqy";

import module namespace json="http://marklogic.com/xdmp/json"
    at "/MarkLogic/json/json.xqy";

declare namespace jb="http://marklogic.com/xdmp/json/basic";

declare default function namespace "http://www.w3.org/2005/xpath-functions";

declare option xdmp:mapping "false";

declare variable $LOCK := "/config/resource-group.lock";

declare function rg:create(
  $name as xs:string,
  $description as xs:string,
  $type as xs:string
) as object-node()
{
  xdmp:lock-for-update($LOCK),
  let $rg  := rs:resource-group-create($name, $description, $type)
  let $_   := trace(("Created resource group: ", $rg/opsdirId), $t:trace-debug-rgrp)
  return
    object-node {
       "id": string($rg/opsdirId),
       "opsdirKind": string($rg/opsdirKind),
       "name": string($rg/name),
       "description": string($rg/description),
       "type": string($rg/opsdirResourceType),
       "role": array-node { () },
       "size": count($rg/resource),
       "resource": array-node { $rg/resource ! string(.) }
    }
};

declare %private function rg:get-values(
  $list, (: untyped :)
  $errmsg as xs:string
) as xs:string*
{
  let $values := for $obj in $list
                 return
                   typeswitch ($obj)
                   case json:array return json:array-values($obj)
                   case xs:string  return $obj
                   case xs:unsignedLong return string($obj)
                   default return error((), "OD-BADLIST", $errmsg)
  for $value in $values
  return
    if (exists(xdmp:to-json($value)/node()/self::null-node()))
    then
      error((), "OD-NONULLS", $errmsg)
    else
      string($value)
};

declare function rg:list(
) as object-node()
{
  rg:list(())
};

declare function rg:list(
  $resource-type as xs:string?
) as object-node()
{
  let $groups := rs:resource-groups($resource-type)
  return
    object-node {
      "resourceGroups": array-node {
        for $group in $groups
        let $roles  := for $perm in $group/groupPermissions
                       where $perm/role != $group/relatedRole
                       return
                         try {
                           xdmp:role-name(xs:unsignedLong($perm/role))
                         } catch ($e) {
                           (: What if the role has been deleted? :)
                           if ($e/error:code = "SEC-ROLEDNE")
                           then ()
                           else xdmp:rethrow()
                         }
        return
          object-node {
            "id": string($group/opsdirId),
            "opsdirKind": string($group/opsdirKind),
            "name": string($group/name),
            "description": string($group/description),
            "type": string($group/opsdirResourceType),
            "role": array-node { $roles },
            "size": $group/size,
            "resource": array-node { $group/resource ! string(.) }
         }
      }
    }
};

declare function rg:get(
  $id as xs:string
) as object-node()?
{
  let $rg := rs:resource-group($id)
  where exists($rg)
  return
    let $roles := for $perm in $rg/groupPermissions
                  where $perm/role != $rg/relatedRole
                  return
                    try {
                      xdmp:role-name(xs:unsignedLong($perm/role))
                    } catch ($e) {
                      (: What if the role has been deleted? :)
                      if ($e/error:code = "SEC-ROLEDNE")
                      then ()
                      else xdmp:rethrow()
                    }
    return
      object-node {
        "id": string($rg/opsdirId),
        "opsdirKind": string($rg/opsdirKind),
        "name": string($rg/name),
        "description": string($rg/description),
        "type": string($rg/opsdirResourceType),
        "role": array-node { $roles },
        "size": $rg/size,
        "resource": array-node { $rg/resource ! string(.) }
      }
};

declare function rg:delete(
  $id as xs:string
) as empty-sequence()
{
  xdmp:lock-for-update($LOCK),
  let $_ := trace("Deleted resource group: " || $id, $t:trace-debug-rgrp)
  return
    rs:resource-group-delete($id)
};

declare function rg:update(
  $id as xs:string,
  $newgroup as object-node()
)
{
    rg:update($id,$newgroup,())
};

declare function rg:update(
  $id as xs:string,
  $newgroup as object-node(),
  $role-operation as xs:string?
) as empty-sequence()
{
  xdmp:lock-for-update($LOCK),
  let $rg := rs:resource-group($id)
  return
    if (empty($rg))
    then
      error((), "OD-BADID", "No resource group with id: " || $id)
    else
      let $type    := $newgroup/type/data()
      let $json    := json:transform-from-json($newgroup)

      (: Use the XML format so that we can tell the $#?!@ing difference
         between an absent 'resource' key and a 'resource' key with a
         value that is an empty array. :)

      let $rg      := if (exists($json/jb:name))
                      then rs:resource-group-set-name($rg, $json/jb:name/data())
                      else $rg

      let $rg      := if (exists($json/jb:description))
                      then rs:resource-group-set-description($rg, $json/jb:description/data())
                      else $rg

      let $rsrcs   := let $jrsrc := $json/jb:resource
                      where exists($jrsrc)
                      return
                        switch ($jrsrc/@type)
                        case "null"   return ()
                        case "string" return string($jrsrc)
                        case "array"  return $jrsrc/jb:item ! string(.)
                        default return
                          error((), "OD-BADRSRC", "Resource must be an array of resource ids")

      let $rsrcs   := try {
                        $rsrcs ! xs:unsignedLong(.)
                      } catch ($e) {
                        if ($e/error:code = "XDMP-CAST")
                        then
                          error((), "OD-BADID", "Resource ID is not an unsignedLong")
                        else
                          xdmp:rethrow()
                      }

      let $rg      := if ($json/jb:resource)
                      then rs:resource-group-set-resources($rg, $rsrcs ! xs:string(.))
                      else $rg

      let $roles   := let $jrole := $json/jb:role
                      where exists($jrole)
                      return
                        switch ($jrole/@type)
                        case "null"   return ()
                        case "string" return string($jrole)
                        case "array"  return $jrole/jb:item ! string(.)
                        default return
                          error((), "OD-BADROLE", "Role must be an array of role names")

      (: Bugwards compatibility...if $roles are specified and the
         roleOperation is not, treat it as a "set" operation. :)
      let $role-operation := if (empty($role-operation) and exists($json/jb:role))
                             then "set"
                             else $role-operation

      let $existing-roles := for $gp in $rg/groupPermissions
             return if(string($gp/capability) eq "read")
                    then string(ts:get-role-names($gp/role cast as xs:unsignedLong))
                    else ()

      let $roles   := switch($role-operation)
                      case "add" return distinct-values(($roles,$existing-roles))
                      case "set" return distinct-values($roles)
                      case "remove" return
                        for $role in $existing-roles
                        return if($role = $roles)
                            then ()
                            else $role
                      default return
                        (: Remove any roles that no longer exist. :)
                        for $role in $existing-roles
                        let $exists := try { xdmp:role($role) } catch ($e) { () }
                        where exists($exists)
                        return
                          $role

      (: Always update the group permissions in case the complement of roles
         has changed outside of the purvue of Ops Director. :)
      let $rg      := rs:resource-group-set-group-permissions($rg, $roles)

      let $_       := if ($rg/opsdirId = $id or empty($id))
                      then ()
                      else error((), "OD-NOTALLOWED", "Cannot change resource group id")
      let $_       := if ($rg/opsdirResourceType = $type or empty($type))
                      then ()
                      else error((), "OD-NOTALLOWED", "Cannot change resource group type")
      return
        (rs:resource-scopes-save($rg),
         let $_ := trace("Updated resource group: " || $id, $t:trace-debug-rgrp)
         return ())
};
