xquery version "1.0-ml";

(: Copyright 2015-2018 MarkLogic Corporation.  All Rights Reserved. :)

module namespace rm="http://marklogic.com/v1/opsdirector/resource-scopes-map";

import module namespace t="http://marklogic.com/v1/opsdirector"
    at "/common/opsdir.xqy";

declare default function namespace "http://www.w3.org/2005/xpath-functions";

declare option xdmp:mapping "false";

(: A resource-scope-map is a data structure for keeping track of the
   relationships between resources and resource scopes. It's a map.

   map["scopes"]    -> map[scopeId]    -> map[resourceId, resource]
   map["resources"] -> map[resourceId] -> map[scopeId, scope]
   map["deleted"]   -> map[resourceId, (scopeIds)]
   map["queue"]     -> map[scopeId, scope]
   map["pscopes"]   -> map[scopeId, scope]
   map["prcrcs"]    -> map[scopeId, scope]
:)

declare function rm:new(
  $initial-scopes as object-node()*
) as map:map
{
  map:new((
    map:entry("scopes", map:map()),
    map:entry("resources", map:map()),
    map:entry("deleted", map:map()),
    map:entry("queue", map:new($initial-scopes ! map:entry(./opsdirId, .))),
    map:entry("pscopes", map:map()),
    map:entry("prsrcs", map:map())))
};

declare function rm:finished(
  $map as map:map
) as xs:boolean
{
  if (empty(map:keys(map:get($map, "queue"))))
  then
    let $rscopes := map:get($map, "scopes")
    let $rsrcs   := map:get($map, "resources")
    let $_       := trace(count(map:keys($rscopes)) || " scopes; "
                        || count(map:keys($rsrcs)) || " resources", $t:trace-debug-scopes)
    let $_       := rm:dump-map($map)
    return
      true()
  else
    false()
};

declare function rm:queue(
  $map as map:map
) as object-node()*
{
  (: N.B. This is a destructive call! :)
  let $queueMap   := map:get($map, "queue")
  let $items      := map:keys($queueMap) ! map:get($queueMap, .)
  let $procScopes := map:get($map, "pscopes")
  let $_          := $items ! map:put($procScopes, ./opsdirId, .)
  let $_          := map:put($map, "queue", map:map())
  return
    $items
};

declare function rm:unassociate(
  $map as map:map,
  $rsrc as object-node(),
  $scope as object-node()
) as empty-sequence()
{
  let $delMap := map:get($map, "deleted")
  let $_      := if (map:contains($delMap, $rsrc/opsdirId))
                 then ()
                 else map:put($delMap, $rsrc/opsdirId, ())
  let $cscope := map:get($delMap, $rsrc/opsdirId)
  let $procRsrcs := map:get($map, "prsrcs")
  return
    (if (map:contains($procRsrcs, $rsrc/opsdirId))
     then
       ()
     else
       map:put($procRsrcs, $rsrc/opsdirId, $rsrc),
     map:put($delMap, $rsrc/opsdirId, distinct-values(($cscope, $scope/opsdirId))))
};

declare function rm:associate(
  $map as map:map,
  $rsrc as object-node(),
  $scope as object-node()
) as empty-sequence()
{
  let $scopeMap := map:get($map, "scopes")
  let $_        := if (map:contains($scopeMap, $scope/opsdirId))
                   then ()
                   else
                     let $_ := trace(rm:to-string("Scope:    ", $scope),
                                     $t:trace-debug-scopes)
                     return
                       map:put($scopeMap, $scope/opsdirId, map:map())
  let $scopeMap := map:get($scopeMap, $scope/opsdirId)

  let $rsrcMap  := map:get($map, "resources")
  let $_        := if (map:contains($rsrcMap, $rsrc/opsdirId))
                   then ()
                   else
                      let $_ := trace(rm:to-string("Resource: ", $rsrc),
                                      $t:trace-debug-scopes)
                     return
                       map:put($rsrcMap, $rsrc/opsdirId, map:map())
  let $rsrcMap  := map:get($rsrcMap, $rsrc/opsdirId)

  let $queueMap   := map:get($map, "queue")
  let $procScopes := map:get($map, "pscopes")
  let $procRsrcs  := map:get($map, "prsrcs")
  return
    (if (map:contains($queueMap, $scope/opsdirId)
         or map:contains($procScopes, $scope/opsdirId))
     then
       ()
     else
       map:put($queueMap, $scope/opsdirId, $scope),

     if (map:contains($procRsrcs, $rsrc/opsdirId))
     then
       ()
     else
       map:put($procRsrcs, $rsrc/opsdirId, $rsrc),

     if (map:contains($scopeMap, $rsrc/opsdirId))
     then
       ()
     else
       map:put($scopeMap, $rsrc/opsdirId, $rsrc),

     if (map:contains($rsrcMap, $scope/opsdirId))
     then
       ()
     else
       map:put($rsrcMap, $scope/opsdirId, $scope))
};

declare function rm:resources(
  $map as map:map
) as object-node()*
{
  let $procRsrcs := map:get($map, "prsrcs")
  return
    map:keys($procRsrcs) ! map:get($procRsrcs, .)
};

declare function rm:get-scopes(
  $map as map:map,
  $rsrc as object-node()
) as object-node()*
{
  let $rsrcs     := map:get($map, "resources")
  let $scopeMap  := map:get($rsrcs, $rsrc/opsdirId)
  let $delMap    := map:get($map, "deleted")
  let $delscopes := map:get($delMap, $rsrc/opsdirId)
  where exists($scopeMap)
  return
    for $id in map:keys($scopeMap)
    where not($id = $delscopes)
    return
      map:get($scopeMap, $id)
};

declare %private function rm:dump-map(
  $map as map:map
) as empty-sequence()
{
  let $rsrcs := map:get($map, "resources")
  for $rid in map:keys($rsrcs)
  let $rsrc := map:get(map:get($map, "prsrcs"), $rid)
  let $scopes := rm:get-scopes($map, $rsrc)
  return
    let $_ := trace(rm:to-string("Resource: ", $rsrc), $t:trace-debug-scopes)
    for $scope in $scopes
    let $_ := trace(rm:to-string("   Scope: ", $scope), $t:trace-debug-scopes)
    return
      ()
};

declare %private function rm:to-string(
  $prefix as xs:string,
  $obj as object-node()
) as xs:string
{
  let $parts := ($obj/opsdirKind/string(),
                 $obj/opsdirResourceType/string(),
                 $obj/opsdirName/string(),
                 $obj/resourceName/string())
  return
    $prefix || $obj/opsdirId || " " || string-join($parts, "/")
};
