xquery version "1.0-ml";

(: Copyright 2015-2018 MarkLogic Corporation.  All Rights Reserved. :)

module namespace rs="http://marklogic.com/v1/opsdirector/resource-scopes";

import module namespace rm="http://marklogic.com/v1/opsdirector/resource-scopes-map"
    at "/common/rmap.xqy";

import module namespace t="http://marklogic.com/v1/opsdirector"
    at "/common/opsdir.xqy";

import module namespace ts="http://marklogic.com/v1/opsdirector/security"
    at "/common/opsec.xqy";

declare default function namespace "http://www.w3.org/2005/xpath-functions";

declare option xdmp:mapping "false";

(: ====================================================================== :)

declare %private variable $QOPTS
  := ("collation=http://marklogic.com/collation/");

(: ====================================================================== :)

declare function rs:resource-group-create(
  $name as xs:string,
  $description as xs:string,
  $resource-type as xs:string
) as object-node()
{
  (: N.B. There's no option to create a resource group with an initial
     set of resources because (1) the app doesn't need that today and
     (2) the existing resource documents wouldn't get their permissions
     updated. Fixing the latter doesn't seem worth the effort right now,
     given the former :)

  if (normalize-space($name) = "")
  then error((), "OD-BADNAME", ("Name must not be empty"))
  else (),

  let $resources := t:assert-resource-type((), $resource-type)

  let $role   := rs:create-resource-group-role($name)
  let $roleid := ts:create-role($role)
  let $_      := xdmp:sleep(1000) (: Give the first transactions a chance to commit :)
  let $_      := ts:add-roles-to-user($role)
  let $_      := xdmp:sleep(1000) (: Give the second transaction a chance to commit :)

  let $id     := string(xdmp:random())
  let $uri    := "/resource-group/" || $id || "/" || t:timestamp() || ".json"

  let $group  := object-node {
                   "opsdirKind": $t:RESOURCE-GROUP,
                   "opsdirResourceType": $resource-type,
                   "opsdirId": $id,
                   "opsdirDtstart": string(current-dateTime()),
                   "opsdirDtend": "9999-12-31T23:59:59Z",
                   "ts": string(current-dateTime()),
                   "name": $name,
                   "groupPermissions": array-node {
                      (object-node {
                        "role": string($roleid),
                        "capability": "update"
                       })
                   },
                   "description": $description,
                   "size": 0,
                   "resource": array-node { () }
                 }
  return
    (
     (: This next bit has to be amped and has to be in a diff txn :)
     ts:insert-resource-group($uri, $role, $group),
     $group
    )
};

declare function rs:resource-group-delete(
  $id as xs:string
) as empty-sequence()
{
  let $rg := rs:resource-group($id)
  where exists($rg)
  return
    let $roleid  := $rg/relatedRole
    let $role    := ts:get-role-names($roleid)

    (: Remove all roles and resources; then save it to update the permissions :)
    let $newrg := rs:resource-group-set-group-permissions($rg, ())
    let $newrg := rs:resource-group-set-resources($newrg, ())
    let $newrg := t:replace-property($newrg, "opsdirDtend", string(current-dateTime()))
    let $newrg := t:replace-property($newrg, "DELETE", "true")
    return
      (rs:resource-scopes-save($newrg),
       ts:remove-role($role))
};

declare function rs:resource-group-ids(
) as xs:string*
{
  rs:resource-group-ids(())
};

declare function rs:resource-group-ids(
  $resource-type as xs:string?
) as xs:string*
{
  rs:resource-groups($resource-type)/opsdirId
};

declare function rs:resource-groups(
) as object-node()*
{
  rs:resource-groups(())
};

declare function rs:resource-groups(
  $resource-type as xs:string?
) as object-node()*
{
  if (empty($resource-type) or ($resource-type = $t:RESOURCE-TYPES))
  then ()
  else error((), "OD-BADRTYPE", ("Bad resource type: " || $resource-type)),

  let $type-query := if (empty($resource-type))
                     then
                       ()
                     else
                       cts:element-range-query(
                         xs:QName("opsdirResourceType"), "=", $resource-type, $QOPTS)
  let $dt-query   := cts:element-range-query(
                       xs:QName("opsdirDtend"), ">", current-dateTime())

  return
    cts:search(collection($t:COLLROOT || "resource-group"),
               cts:and-query(($type-query, $dt-query)),
               "unfiltered")/object-node()
};

declare function rs:resource-group(
  $id as xs:string
) as object-node()?
{
  let $id-query  := cts:element-range-query(xs:QName("opsdirId"), "=", $id, $QOPTS)
  let $dt-query  := cts:element-range-query(
                        xs:QName("opsdirDtend"), ">", current-dateTime())
  return
    cts:search(collection($t:COLLROOT || "resource-group"),
               cts:and-query(($id-query, $dt-query)), ("unfiltered"))/object-node()
};

declare function rs:resource-group-set-name(
  $group as object-node(),
  $name as xs:string
) as object-node()
{
  rs:replace-property(t:assert-scope($group, $t:RESOURCE-GROUP), "name", $name)
};

declare function rs:resource-group-set-description(
  $group as object-node(),
  $description as xs:string
) as object-node()
{
  let $group := t:assert-scope($group, $t:RESOURCE-GROUP)
  return
    rs:replace-property($group, "description", $description)
};

declare function rs:resource-group-add-group-permissions(
  $group as object-node(),
  $roles as xs:string*
) as object-node()
{
  let $group  := rs:assert-scope($group, $t:RESOURCE-GROUP)
  let $roles  := rs:assert-roles($roles)
  let $update := (for $role in distinct-values($roles)
                  return
                    object-node { "role": $role, "capability": "read" },
                  for $perm in $group/groupPermissions
                  where not($perm/role = $roles)
                  return
                    $perm)
  return
    rs:replace-property($group, "groupPermissions", array-node { $update })
};

declare function rs:resource-group-remove-group-permissions(
  $group as object-node(),
  $roles as xs:string*
) as object-node()
{
  let $group  := rs:assert-scope($group, $t:RESOURCE-GROUP)
  let $roles  := rs:assert-roles($roles)
  let $update := (for $perm in $group/groupPermissions
                  where not($perm/role = $roles)
                  return
                    $perm)
  return
    rs:replace-property($group, "groupPermissions", array-node { $update })
};

declare function rs:resource-group-set-group-permissions(
  $group as object-node(),
  $roles as xs:string*
) as object-node()
{
  let $group  := rs:assert-scope($group, $t:RESOURCE-GROUP)
  let $roles  := rs:assert-roles($roles)
  let $keep   := $group/groupPermissions[capability = "update"]
  let $update := (for $role in distinct-values($roles)
                  return
                    object-node { "role": $role, "capability": "read" })
  return
    rs:replace-property($group, "groupPermissions", array-node { ($keep, $update) })
};

declare function rs:resource-group-add-resources(
  $group as object-node(),
  $ids as xs:string*
) as object-node()
{
  let $type    := $group/opsdirResourceType
  let $_       := for $id in $ids
                  let $rsrc := t:managed-resource((), $type, $id)
                  where empty($rsrc)
                  return
                    error((), "OD-BADRSRC", ("Bad resource; not a " || $type))
  let $group   := rs:assert-scope($group, $t:RESOURCE-GROUP)
  let $current := $group/resource
  let $update  := distinct-values(($current, $ids))
  let $obj     := rs:replace-property($group, "resource", array-node { $update })
  return
    rs:replace-property($obj, "size", count($update))
};

declare function rs:resource-group-remove-resources(
  $group as object-node(),
  $rsrcs as xs:string*
) as object-node()
{
  let $group  := t:assert-scope($group, $t:RESOURCE-GROUP)
  let $update := for $rsrc in $group/resource
                 where not($rsrc = $rsrcs)
                 return
                   $rsrc
  let $obj     := rs:replace-property($group, "resource", array-node { $update })
  return
    rs:replace-property($obj, "size", count($update))
};

declare function rs:resource-group-set-resources(
  $group as object-node(),
  $rsrcs as xs:string*
) as object-node()
{
  let $type    := $group/opsdirResourceType
  let $group  := t:assert-scope($group, $t:RESOURCE-GROUP)
  let $update := distinct-values($rsrcs)
  let $_       := for $id in $update
                  let $rsrc := t:managed-resource((), $type, $id)
                  where empty($rsrc)
                  return
                    error((), "OD-BADRSRC", ("Bad resource; not a " || $type))
  let $obj    := rs:replace-property($group, "resource", array-node { $update })
  return
    rs:replace-property($obj, "size", count($update))
};

(: ====================================================================== :)

declare function rs:resource-scopes-save(
  $scopes as object-node()*
) as empty-sequence()
{
  let $_   := trace("Resource scopes save " || count($scopes) || " scope(s)",
                    $t:trace-debug-scopes)
  let $_   := for $scope at $index in $scopes
              return
                trace(rs:to-string("[" || $index || "] ", $scope), $t:trace-debug-scopes)

  let $map := rm:new($scopes)

  (: What about resources that have been removed from a scope? :)
  (: I think this can only happen with resource groups... :)
  let $old := rm:new(for $scope in $scopes
                     where $scope/opsdirKind = "resource-group"
                     return
                       rs:resource-group($scope/opsdirId))

  let $_       := trace("============================================================",
                         $t:trace-debug-scopes)

  let $map     := rs:resource-scopes-closure($map)
  let $rsrcs   := rm:resources($map)

  let $old     := rs:resource-scopes-closure($old)
  let $orsrcs  := rm:resources($old)

  let $rsrcs   := $rsrcs | $orsrcs

  let $_       := trace("/===========================================================",
                         $t:trace-debug-scopes)

  (: For each resource, update the permissions, if necessary :)
  let $_  := for $rsrc in $rsrcs
             let $uri     := xdmp:node-uri($rsrc)
             let $rscopes := rm:get-scopes($map, $rsrc)
             let $newperm := rs:effective-resource-permissions($rsrc, $rscopes)
             let $curperm := xdmp:document-get-permissions($uri)
             return
               if (rs:same-permissions($curperm, $newperm))
               then
                 let $_ := trace(("Same permissions: " || $uri), $t:trace-debug)
                 let $_ := trace(("Current:", rs:show-perms($curperm)), $t:trace-debug-perm)
                 return ()
               else
                 let $_ := trace(("Updating permissions: " || $uri), $t:trace-debug)
                 let $_ := trace(("Current:", rs:show-perms($curperm)), $t:trace-debug-perm)
                 let $_ := trace(("New:", rs:show-perms($newperm)), $t:trace-debug-perm)
                 return
                   xdmp:document-set-permissions($uri, $newperm)

  (: For each resource and resource-group, save it if it's been changed :)
  return
    for $scope in $scopes
    where empty(xdmp:node-uri($scope))
    return
      let $_ := trace(("Saving resource scope:", $scope), $t:trace-debug-scopes)
      return
        ts:resource-scope-save($scope)
};

declare function rs:resource-scopes-closure(
  $map as map:map
) as map:map
{
  (: For a given scope, we need to find all of the resources covered
     by that scope. For each resource, we need to find all of the
     resource scopes that contain that resource. The closure is the
     set that contains all of the scopes for all of the resources
     in all of the scopes. We use a map to keep track so that
     we don't have to re-process resources or scopes we've already
     processed.

     The input map is expected to have three keys:
     * scopes: an ID/scope map of the scopes processed
     * resources: an ID/resource map of the resources processed
     * queue: a list of scopes yet to be processed.

     The function recurses until the queue is empty.
  :)

  let $queue := rm:queue($map)
  let $_     := for $scope in $queue
                let $_ := trace(rs:to-string("Closure:  ", $scope), $t:trace-debug-scopes)
                for $rsrc in rs:resource-scope-resources($scope)
                return
                  (rm:associate($map, $rsrc, $scope),
                   for $scope in rs:applicable-scopes($rsrc)
                   return
                     rm:associate($map, $rsrc, $scope))
  return
    if (rm:finished($map))
    then
      $map
    else
      rs:resource-scopes-closure($map)
};

declare function rs:effective-resource-permissions(
  $resource as object-node()
) as element(sec:permission)+
{
(:
  let $_ := trace("Effective resource permissions for "
                   || $resource/opsdirResourceType || " " || $resource/opsdirName,
                  $t:trace-debug-scopes)
:)
  let $scopes := rs:applicable-scopes($resource)
(:
  let $_ := trace(("Scopes:", $scopes), $t:trace-debug-scopes)
:)
  return
    rs:effective-resource-permissions($resource, $scopes)
};

declare function rs:effective-resource-permissions(
  $resource as object-node(),
  $scopes as object-node()*
) as element(sec:permission)+
{
  let $cluster := $resource/opsdirClusterId
  let $type    := $resource/opsdirResourceType
  let $id      := $resource/opsdirId
  let $scopes  := $scopes ! rs:assert-scope(.)

  let $roles   := ($scopes/permissions/role ! xs:unsignedLong(.),
                   $scopes/groupPermissions/role ! xs:unsignedLong(.))

  let $roles := distinct-values($roles)

  return
    ($t:DBPERMS,
     for $roleid in $roles
     let $role := ts:get-role-names($roleid)
     let $cap  := "read"

     where exists($role) and not(starts-with($role, "opsdir-rg-internal-"))
     return
       let $_   := trace("Resource permission: "
                         || $role || " can " || string-join($cap,",") || " "
                         || $type || " " || $id || " on cluster "
                         || $cluster, $t:trace-debug-perm)
       return
         xdmp:permission($role, $cap))
};

declare %private function rs:same-permissions(
  $curperm as element(sec:permission)*,
  $newperm as element(sec:permission)*
) as xs:boolean
{
  let $a := for $perm in $curperm
            return
              if ($newperm[sec:capability = $perm/sec:capability
                           and sec:role-id = $perm/sec:role-id])
              then ()
              else ("notsame")
  let $b := for $perm in $newperm
            return
              if ($curperm[sec:capability = $perm/sec:capability
                           and sec:role-id = $perm/sec:role-id])
              then ()
              else ("notsame")
  return
    empty($a) and empty($b)
};

declare %private function rs:show-perms(
  $perms as element(sec:permission)*
) as xs:string*
{
  $perms ! rs:show-perm(.)
};

declare %private function rs:show-perm(
  $perm as element(sec:permission)
) as xs:string
{
  let $role := ts:get-role-names($perm/sec:role-id)
  return
    "Permission: " || $role || " " || $perm/sec:capability
};

declare %private function rs:to-string(
  $prefix as xs:string,
  $obj as object-node()
) as xs:string
{
  let $parts := ($obj/opsdirKind/string(),
                 $obj/opsdirResourceType/string(),
                 $obj/opsdirName/string())
  return
    $prefix || $obj/opsdirId || " " || string-join($parts, "/")
};

(: ====================================================================== :)

declare function rs:resource-scope(
  $cluster as xs:string?,
  $resource-type as xs:string,
  $resource-id as xs:string
) as object-node()
{
  rs:resource-scope($cluster, $resource-type, $resource-id, true())
};

declare %private function rs:resource-scope(
  $cluster as xs:string?,
  $resource-type as xs:string,
  $resource-id as xs:string,
  $create as xs:boolean
) as object-node()?
{
  if (empty($cluster) or ($cluster = t:managed-resource-ids("clusters")))
  then ()
  else error((), "OD-NOCLUSTER", ("No such cluster: " || $cluster)),

  let $obj  := t:managed-resource($cluster, $resource-type, $resource-id)
  let $_    := if (empty($obj))
               then error((), "OD-NORESOURCE", ("Invalid scope, no such resource ("
                              || $resource-type || "/" || $resource-id || ")"))
               else ()

  let $cid  := string($obj/opsdirClusterId)
  let $rid  := string($obj/opsdirId)

  let $dt-query  := cts:element-range-query(
                        xs:QName("opsdirDtend"), ">", current-dateTime())
  let $c-query   := cts:json-property-range-query("opsdirClusterId", "=", $cid, $QOPTS)
  let $id-query  := cts:json-property-range-query("opsdirId", "=", $rid, $QOPTS)
  let $k-query   := cts:json-property-value-query("opsdirKind", "resource")

  let $doc       := cts:search(collection($t:COLLROOT || "resource-scope"),
                               cts:and-query(($dt-query, $c-query, $id-query, $k-query)),
                               "unfiltered")

  let $_    := if (count($doc) gt 1)
               then error((), "OD-INTERR" ("Internal error: multiple resources for "
                              || "resource-scope(" || $cluster || ", "
                              || $resource-type || ", " || $resource-id || ")"))
               else ()
  return
    if (empty($doc) and $create)
    then
      object-node {
        "opsdirKind": $t:RESOURCE-SCOPE,
        "opsdirClusterId": $cid,
        "opsdirResourceType": $resource-type,
        "opsdirId": $rid,
        "opsdirDtstart": string(current-dateTime()),
        "ts": string(current-dateTime()),
        "opsdirDtend": "9999-12-31T23:59:59Z"
      }
    else
      $doc/object-node()
};

declare function rs:default-resource-scope(
  $cluster as xs:string?,
  $resource-type as xs:string?
) as object-node()
{
  rs:default-resource-scope($cluster, $resource-type, true())
};

declare %private function rs:default-resource-scope(
  $cluster as xs:string?,
  $resource-type as xs:string?,
  $create as xs:boolean
) as object-node()?
{
  if (empty($resource-type) or ($resource-type = ("all", $t:RESOURCE-TYPES)))
  then ()
  else error((), "OD-BADRTYPE", ("Bad resource type: " || $resource-type)),

  if (empty($cluster) or ($cluster = ("all", t:managed-resource-ids("clusters"))))
  then ()
  else error((), "OD-NOCLUSTER", ("No such cluster: " || $cluster)),

  let $cid   := if (empty($cluster)) then "all" else $cluster
  let $rtype := if (empty($resource-type)) then "all" else $resource-type

  let $dt-query  := cts:element-range-query(
                        xs:QName("opsdirDtend"), ">", current-dateTime())
  let $c-query   := cts:element-range-query(
                        xs:QName("opsdirClusterId"), "=", $cid, $QOPTS)
  let $t-query   := cts:element-range-query(
                        xs:QName("opsdirResourceType"), "=", $rtype, $QOPTS)
  let $id-query  := cts:element-range-query(xs:QName("opsdirId"), "=", "all", $QOPTS)
  let $doc       := cts:search(collection($t:COLLROOT || "resource-scope"),
                               cts:and-query(($dt-query, $c-query, $t-query, $id-query)),
                               "unfiltered")

  let $_    := if (count($doc) gt 1)
               then error((), "OD-INTERR", ("Internal error: multiple resources for "
                              || "default-resource-scope(" || $cluster || ", "
                              || $resource-type || ")"))
               else ()
  return
    if (empty($doc) and $create)
    then
      object-node {
        "opsdirKind": $t:DEFAULT-RESOURCE-SCOPE,
        "opsdirClusterId": $cid,
        "opsdirResourceType": $resource-type,
        "opsdirId": "all",
        "opsdirDtstart": string(current-dateTime()),
        "ts": string(current-dateTime()),
        "opsdirDtend": "9999-12-31T23:59:59Z"
      }
    else
      $doc/object-node()
};

declare function rs:resource-scope-resources(
  $scope as object-node()
) as object-node()*
{
  let $map       := map:map()
  let $resources := switch ($scope/opsdirKind)
                    case $t:DEFAULT-RESOURCE-SCOPE
                    return
                      rs:default-resource-resources($scope)
                    case $t:RESOURCE-SCOPE
                    return
                      t:managed-resource($scope/opsdirClusterId,
                                         $scope/opsdirResourceType,
                                         $scope/opsdirId)
                    case $t:RESOURCE-GROUP
                    return
                      $scope/resource ! t:managed-resource((), $scope/opsdirResourceType, .)
                    default
                    return
                      error((), "OD-BADSCOPE", ("Unexpected scope: " || local-name($scope)))
  let $_         := $resources ! map:put($map, xdmp:node-uri(.), .)
  let $_         := $resources ! rs:inherited-resources(.) ! map:put($map, xdmp:node-uri(.), .)
  return
    map:keys($map) ! map:get($map, .)
};

declare %private function rs:default-resource-resources(
  $scope as object-node()
) as object-node()*
{
  let $cluster  := if ($scope/opsdirClusterId = "all")
                   then ()
                   else $scope/opsdirClusterId
  let $type     := if ($scope/opsdirResourceType = "all")
                   then ()
                   else $scope/opsdirResourceType

  let $c-query   := if ($cluster)
                    then
                      cts:element-range-query(
                          xs:QName("opsdirClusterId"), "=", $cluster, $QOPTS)
                    else
                      ()

  let $t-query   := if ($type)
                    then
                      cts:element-range-query(
                          xs:QName("opsdirResourceType"), "=", $type, $QOPTS)
                    else
                      ()
  return
     cts:search(collection($t:JOINS-COLLECTION), cts:and-query(($c-query, $t-query)),
                "unfiltered")/object-node()
};

declare function rs:inherited-resources(
  $rsrc as object-node()
) as object-node()*
{
  let $inherited :=
    switch ($rsrc/opsdirResourceType)
    case "clusters"
    return
      let $c-query := cts:json-property-value-query("opsdirClusterId", $rsrc/opsdirClusterId)
      return
         cts:search(collection($t:JOINS-COLLECTION), $c-query, "unfiltered")/object-node()
    case "hosts"
    return
      let $cquery  := cts:json-property-value-query("opsdirClusterId", $rsrc/opsdirClusterId)
      let $hquery  := cts:json-property-value-query("hostIds", $rsrc/opsdirId)
      let $k1query := cts:json-property-value-query("opsdirResourceType", "tasks")
      let $k2query := cts:json-property-value-query("opsdirResourceType", "forests")
      let $k3query := cts:json-property-value-query("opsdirResourceType", "databases")
      return
         cts:search(collection($t:JOINS-COLLECTION),
                    cts:and-query(($cquery, $hquery,
                                   cts:or-query(($k1query, $k2query, $k3query)))),
                    "unfiltered")/object-node()
    case "databases"
    return
      let $cquery := cts:json-property-value-query("opsdirClusterId", $rsrc/opsdirClusterId)
      let $dquery := cts:json-property-value-query("databaseIds", $rsrc/opsdirId)
      let $kquery := cts:json-property-value-query("opsdirResourceType", "forests")
      return
         cts:search(collection($t:JOINS-COLLECTION),
                    cts:and-query(($cquery, $dquery, $kquery)), "unfiltered")/object-node()
    case "servers"
    return
      let $dbs   := ($rsrc/contentDatabase, $rsrc/modulesDatabase,
                     $rsrc/lastLoginDatabase)
      let $rsrcs := for $db in $dbs
                    let $id := t:managed-resource-id($rsrc/opsdirClusterId,
                                                     "databases", $db)
                    where exists($id)
                    return
                      t:managed-resource($rsrc/opsdirClusterId, "databases", $id)
      return
        ($rsrcs, $rsrcs ! rs:inherited-resources(.))
    default
    return
      ()

  let $ids := distinct-values(($rsrc/opsdirId, $inherited/opsdirId))

  let $inherited_alerts :=
    cts:search(collection($t:ALERTS-COLLECTION),
               cts:and-query(
                 (cts:json-property-value-query("resourceId", $ids),
                  cts:json-property-value-query("status", "open"))))
               /object-node()

  return
    ($inherited, $inherited_alerts)
};

declare %private function rs:inherited-groups(
  $rsrc as object-node()
) as object-node()*
{
  (: We start out with a resource object that might be an underlying /resource/
     and not a join, so we get the join document. We can compute all of the
     cluster-level resource groups directly and once. The inherited groups we
     find recursively.
   :)
  let $cid       := $rsrc/opsdirClusterId
  let $joinrsrc  := t:managed-resource($cid, $rsrc/opsdirResourceType, $rsrc/opsdirId)
  let $cluster   := rs:resource-groups("clusters")[resource = $cid]
  let $type      := $rsrc/opsdirResourceType
  let $name      := $rsrc/opsdirName
  let $inherited := if (empty($joinrsrc))
                    then
                      xdmp:log("Failed to find join resource for " || $type || " " || $name)
                    else
                      rs:find-inherited-groups($joinrsrc)
  return
    ($cluster | $inherited)
};

declare %private function rs:find-inherited-groups(
  $rsrc as object-node()
) as object-node()*
{
  let $id        := $rsrc/opsdirId
  let $cid       := $rsrc/opsdirClusterId
  let $simple    := rs:resource-groups($rsrc/opsdirResourceType)[resource = $id]
  let $inherited := switch ($rsrc/opsdirResourceType)
                    case "forests"
                    return
                      (for $id in $rsrc/databaseIds
                       let $db := t:managed-resource($cid, "databases", $id)
                       where exists($db)
                       return
                         rs:inherited-groups($db),
                       for $id in $rsrc/hostIds
                       let $host := t:managed-resource($cid, "hosts", $id)
                       where exists($host)
                       return
                         rs:inherited-groups($host))

                    case "databases"
                    return
                      (for $id in $rsrc/serverIds
                       let $server := t:managed-resource($cid, "servers", $id)
                       where exists($server)
                       return
                         rs:inherited-groups($server),
                       for $id in $rsrc/hostIds
                       let $host := t:managed-resource($cid, "hosts", $id)
                       where exists($host)
                       return
                         rs:inherited-groups($host))

                    default
                    return
                      ()
  return
    ($simple | $inherited)
};

(: ====================================================================== :)

declare function rs:applicable-scopes(
  $resource as object-node()
) as object-node()*
{
  (: Hack; the applicable scopes for an alert are the scopes
     applicable to the resource that this alert pertains to. :)
  let $alert    := if ($resource/opsdirResourceType = "alerts")
                   then
                     t:managed-resource($resource/opsdirClusterId,
                                        (),
                                        $resource/resourceId)
                   else
                     ()
  let $resource := if (empty($alert))
                   then $resource
                   else $alert

  let $cluster  := $resource/opsdirClusterId
  let $type     := if (empty($resource/opsdirResourceType))
                   then "alerts" (: If there are still alerts from 1.0 around... :)
                   else $resource/opsdirResourceType
  let $id       := $resource/opsdirId

  (: Applicable defaults:
     all resources on all clusters
     {$type} resources on all clusters
     all resources on this cluster
     {$type} resources on this cluster
  :)
  let $all_all   := rs:default-resource-scope((), (), false())
  let $all_type  := rs:default-resource-scope((), $type, false())

  (: The cluster may not exist yet, so don't worry if these fail. :)
  let $clus_all  := try { rs:default-resource-scope($cluster, (), false())    }
                    catch ($e) { () }
  let $clus_type := try { rs:default-resource-scope($cluster, $type, false()) }
                    catch ($e) { () }

  (: This resource may not exist yet, so don't worry if this fails. :)
  let $this_rsrc := try { rs:resource-scope((), $type, $id, false()) }
                    catch ($e) { () }

  let $groups    := rs:resource-groups-containing-resources($type, $id)

  let $map       := map:new(for $rs in ($all_all, $all_type,
                                        $clus_all, $clus_type,
                                        $this_rsrc, $groups)
                            return
                              map:entry($rs/opsdirId, $rs))

  let $inherited := rs:inherited-groups($resource)
  let $_         := $inherited ! map:put($map, ./opsdirId, .)
  return
    map:keys($map) ! map:get($map, .)
};

declare function rs:resource-groups-containing-resources(
  $resource-type as xs:string,
  $ids as xs:string+
) as object-node()*
{
  let $rg-ids := for $id in $ids
                 return
                   rs:resource-groups($resource-type)[$id = resource]/opsdirId
  return
    distinct-values($rg-ids) ! rs:resource-group(.)
};

(: ====================================================================== :)

declare function rs:resource-add-permissions(
  $scope as object-node(),
  $roles as xs:string*
) as object-node()
{
  let $roles  := rs:assert-roles($roles)
  let $update := (for $role in distinct-values($roles)
                  return
                    object-node { "role": string($role), "capability": "read" },
                  for $perm in $scope/permissions
                  where not($perm/role = $roles)
                  return
                    $perm)
  return
    rs:replace-property(rs:assert-scope($scope), "permissions", array-node { $update })
};

declare function rs:resource-remove-permissions(
  $scope as object-node(),
  $roles as xs:string*
) as object-node()
{
  let $roles  := rs:assert-roles($roles)
  let $update := (for $perm in $scope/permissions
                  where not($perm/role = $roles)
                  return
                    $perm)
  return
    rs:replace-property(rs:assert-scope($scope), "permissions", array-node { $update })
};

declare function rs:resource-set-permissions(
  $scope as object-node(),
  $roles as xs:string*
) as object-node()
{
  let $roles  := rs:assert-roles($roles)
  let $update := (for $role in distinct-values($roles)
                  return
                    object-node { "role": $role, "capability": "read" })
  return
    rs:replace-property(rs:assert-scope($scope), "permissions", array-node { $update })
};

(: ====================================================================== :)

declare function rs:resource-add-privileges(
  $scope as object-node(),
  $roles as xs:string*,
  $actions as xs:string*
) as object-node()
{
  let $roles  := rs:assert-roles($roles)
  let $update := (for $priv in $scope/privileges
                  return
                    object-node {
                       "role": $priv/role,
                       "actions": array-node {
                                     if ($priv/role = $roles)
                                     then
                                       distinct-values(($priv/actions, $actions))
                                     else
                                       $priv/actions
                                  }
                    },
                  for $role in distinct-values($roles)
                  where empty($scope/privileges[role = $role])
                  return
                    object-node {
                       "role": $role,
                       "actions": array-node { distinct-values($actions) }
                    })
  return
    rs:replace-property(rs:assert-scope($scope), "privileges", array-node { $update })
};

declare function rs:resource-remove-privileges(
  $scope as object-node(),
  $roles as xs:string*,
  $actions as xs:string*
) as object-node()
{
  let $roles  := rs:assert-roles($roles)
  let $update := (for $priv in $scope/privileges
                  return
                    object-node {
                       "role": $priv/role,
                       "actions": array-node {
                                     if ($priv/role = $roles)
                                     then
                                       for $action in $priv/actions
                                       where not($action = $actions)
                                       return
                                         $action
                                     else
                                       $priv/actions
                                  }
                    })
  let $update := for $priv in $update
                 where exists($priv/actions)
                 return
                   $priv
  return
    rs:replace-property(rs:assert-scope($scope), "privileges", array-node { $update })
};

declare function rs:resource-set-privileges(
  $scope as object-node(),
  $roles as xs:string*,
  $actions as xs:string*
) as object-node()
{
  let $roles  := rs:assert-roles($roles)
  let $update := (for $role in distinct-values($roles)
                  return
                    object-node {
                       "role": $role,
                       "actions": array-node { distinct-values($actions) }
                    })
  let $update := for $priv in $update
                 where exists($priv/actions)
                 return
                   $priv
  return
    rs:replace-property(t:assert-scope($scope), "privileges", array-node { $update })
};

(: ====================================================================== :)

declare function rs:resource-add-roles(
  $scope as object-node(),
  $roles as xs:string*,
  $add-roles as xs:string*
) as object-node()
{
  let $roles     := rs:assert-roles($roles)
  let $add-roles := rs:assert-roles($add-roles)

  let $update := (for $robj in $scope/roles
                  return
                    object-node {
                       "role": $robj/role,
                       "addRoles": array-node {
                                     if ($robj/role = $roles)
                                     then
                                       distinct-values(($robj/addRoles, $add-roles))
                                     else
                                       $robj/addRoles
                                  }
                    },
                  for $role in distinct-values($roles)
                  where empty($scope/roles[role = $role])
                  return
                    object-node {
                       "role": $role,
                       "addRoles": array-node { distinct-values($add-roles) }
                    })
  return
    rs:replace-property(rs:assert-scope($scope), "roles", array-node { $update })
};

declare function rs:resource-remove-roles(
  $scope as object-node(),
  $roles as xs:string*,
  $add-roles as xs:string*
) as object-node()
{
  let $roles     := rs:assert-roles($roles)
  let $add-roles := rs:assert-roles($add-roles)

  let $update := (for $robj in $scope/roles
                  return
                    object-node {
                       "role": $robj/role,
                       "addRoles": array-node {
                                     if ($robj/role = $roles)
                                     then
                                       for $role in $robj/addRoles
                                       where not($role = $add-roles)
                                       return
                                         $role
                                     else
                                       $robj/addRoles
                                  }
                    })
  let $update := for $robj in $update
                 where exists($robj/addRoles)
                 return
                   $robj
  return
    rs:replace-property(rs:assert-scope($scope), "roles", array-node { $update })
};

declare function rs:resource-set-roles(
  $scope as object-node(),
  $roles as xs:string*,
  $add-roles as xs:string*
) as object-node()
{
  let $roles     := rs:assert-roles($roles)
  let $add-roles := rs:assert-roles($add-roles)

  let $update := (for $role in distinct-values($roles)
                  return
                    object-node {
                       "role": $role,
                       "addRoles": array-node { distinct-values($add-roles) }
                    })
  let $update := for $robj in $update
                 where exists($robj/addRoles)
                 return
                   $robj
  return
    rs:replace-property(rs:assert-scope($scope), "roles", array-node { $update })
};

(: ====================================================================== :)

declare %private function rs:create-resource-group-role(
  $name as xs:string
) as xs:string
{
  let $secret := "d72e95f66c03b31d855e7546854f2882afd339eb"
  let $seed   := xdmp:get-current-user() || "/" || $name
  let $hash   := xdmp:hmac-sha1($secret, $seed)

  (: If the role exists, so will the document. :)
  let $hash   := if (ts:role-exists("opsdir-rg-internal-" || $hash))
                 then
                   let $_ := trace("Hash collision for " || $seed, $t:trace-status)
                   return
                     string(xdmp:random())
                 else $hash
  return
    "opsdir-rg-internal-" || $hash
};

declare %private function rs:replace-property(
  $obj as object-node(),
  $property as xs:string,
  $value as item()*
) as object-node()
{
  xdmp:to-json(
    map:new(
      (map:entry($property, $value),
       for $prop in $obj/node()
       let $name := local-name-from-QName(node-name($prop))
       where $name ne $property
       return
         map:entry($name, $prop)
    ))
  )/object-node()
};

declare function rs:assert-scope(
  $scope as object-node()
) as object-node()
{
  rs:assert-scope($scope, $t:RESOURCE-SCOPES-LIST)
};

declare function rs:assert-scope(
  $scope as object-node(),
  $scopes as xs:string+
) as object-node()
{
  if ($scope/opsdirKind = $scopes)
  then
    $scope
  else
    error((), "OD-BADSCOPE", ("Unexpected scope: " || $scope/opsdirKind))
};

declare %private function rs:assert-roles(
  $roles as xs:string*
) as xs:string*
{
  for $role in $roles
  return
    try {
       string(xdmp:role($role))
    } catch ($e) {
      if ($role castable as xs:unsignedLong)
      then
        if (empty(ts:get-role-names($role cast as xs:unsignedLong)))
        then error((), "OD-BADROLE", ("Bad role: " || $role))
        else $role
      else
        error((), "OD-BADROLE", ("Bad role: " || $role))
   }
};
