xquery version "1.0-ml";

(: Copyright 2015-2018 MarkLogic Corporation.  All Rights Reserved. :)

(: Note: This module has been written so that it can be shared between
   XQuery and JavaScript. Consequently it uses camelCase naming conventions
   and passes ID values around as strings instead of xs:unsignedLongs.
   Because Javascript.
:)

module namespace sess="http://marklogic.com/v1/opsdirector/session";

import module namespace t="http://marklogic.com/v1/opsdirector"
    at "/common/opsdir.xqy";

import module namespace config="http://marklogic.com/v1/opsdirector/config"
    at "/common/config.xqy";

declare default function namespace "http://www.w3.org/2005/xpath-functions";

declare option xdmp:mapping "false";

declare %private variable $CONFIG-BASE  := "/config/sessions/";
declare %private variable $DURATIONFUZZ := 30 + xdmp:random(10);
declare %private variable $DURATION     := xs:dayTimeDuration("PT" || $DURATIONFUZZ || "M");
declare %private variable $APP-TOKEN    := "this-is-the-random-app-token";

declare function sess:get-session-uri(
  $cluster-id as xs:string,
  $host-id as xs:string
) as xs:string
{
  let $cluster-id := $cluster-id cast as xs:unsignedLong
  let $host-id := $host-id cast as xs:unsignedLong
  return
    "/sessions/" || $cluster-id || "/" || $host-id || ".json"
};

declare function sess:get(
  $cluster-id as xs:string,
  $host-id as xs:string
) as object-node()?
{
  doc(sess:get-session-uri($cluster-id, $host-id))/node()
};

declare function sess:remaining-time(
  $session as object-node()?
) as xs:dayTimeDuration?
{
  if (exists($session) and $session/expires castable as xs:dateTime)
  then
    xs:dateTime($session/expires) - current-dateTime()
  else
    ()
};

declare function sess:expired(
  $session as object-node()
) as xs:boolean
{
  let $expires := xs:dateTime($session/expires)
  return
    $expires lt current-dateTime()
};

declare function sess:setup(
  $data as object-node()
) as object-node()
{
  let $cluster-id   := if ($data/clusterId castable as xs:unsignedLong)
                       then $data/clusterId
                       else error((), "Invalid cluster-id: " || $data/clusterId)
  let $host-id      := if ($data/hostId castable as xs:unsignedLong)
                       then $data/hostId
                       else error((), "Invalid host-id: " || $data/hostId)

  let $session      := xdmp:hmac-md5($APP-TOKEN, string(xdmp:random()), "hex")

  let $hosts        := sess:hosts($cluster-id)
  let $duration     := (config:get-atomic-property("session-duration-" || $cluster-id),
                        config:get-atomic-property("session-duration"))[1]
  let $duration     := if ($duration castable as xs:dayTimeDuration)
                       then xs:dayTimeDuration($duration)
                       else $DURATION

  let $_            := if (xdmp:trace-enabled($t:trace-sessions))
                       then
                         let $c := t:managed-resource($cluster-id, "clusters", $cluster-id)
                         let $h := t:managed-resource($cluster-id, "hosts", $host-id)
                         return
                           (trace("Starting " || $duration || " session for "
                                  || $c/opsdirName || "/" || $h/opsdirName
                                  || " to " || $hosts[1],
                                  $t:trace-sessions),
                            trace("  Session: " || $session, $t:trace-sessions))
                       else
                         ()

  (: Get rid of fractional seconds (just round down to the minute) :)
  let $expires      := current-dateTime() + $duration
  let $expires      := adjust-dateTime-to-timezone($expires, xs:dayTimeDuration("PT0S"))
  let $fracsecs     := seconds-from-dateTime($expires)
  let $expires      := $expires - xs:dayTimeDuration("PT" || $fracsecs || "S")

  let $streams      := $data/streams
  let $extra-data   := $data/node() except ($data/streams
                                            | $data/sessionId
                                            | $data/created
                                            | $data/clusterId
                                            | $data/hostId
                                            | $data/expires
                                            | $data/response)

  let $sessuri   := sess:get-session-uri($cluster-id, $host-id)
  let $cursess   := sess:get($cluster-id, $host-id)
  return
    let $resp    := sess:construct-response($cluster-id, $host-id, $hosts,
                                            $session, $expires)
    let $map     := map:new((
                       map:entry("sessionId", $session),
                       map:entry("created", string(current-dateTime())),
                       map:entry("clusterId", $cluster-id),
                       map:entry("opsdirClusterId", $cluster-id),
                       map:entry("hostId", $host-id),
                       map:entry("opsdirHostId", $host-id),
                       map:entry("expires", string($expires)),
                       map:entry("response", $resp),
                       for $node in $cursess/*
                                    except ($cursess/sessionId
                                            | $cursess/created
                                            | $cursess/clusterId
                                            | $cursess/hostId
                                            | $cursess/expires
                                            | $cursess/response)
                       return
                         map:entry(local-name-from-QName(node-name($node)), $node),
                       for $node in $extra-data
                       return
                         map:entry(local-name-from-QName(node-name($node)), $node)))
    let $sess    := xdmp:to-json($map)/node()
    let $_       := if (empty($cursess))
                    then
                      let $_ := trace("Ins: " || $sessuri, $t:trace-debug-uri)
                      return
                        (xdmp:document-insert($sessuri, $sess, $t:DBPERMS,
                                              ($t:COLLROOT||"sessions",
                                              ($t:COLLROOT||"sessions/"||$cluster-id))),
                         trace("Insert: " || $sessuri, $t:trace-debug))
                    else
                      (xdmp:node-replace($cursess, $sess),
                       trace("Update: " || $sessuri, $t:trace-debug))
    return
      $resp
};

declare %private function sess:hosts(
  $cluster-id as xs:string
) as xs:string+
{
  let $hosts := (config:get-atomic-property("session-hosts-" || $cluster-id),
                 config:get-atomic-property("session-hosts"))[1]
  let $hosts := if (empty($hosts) or (normalize-space($hosts) eq ''))
                then
                  (: By default, spread the load over hosts in this group :)
                  xdmp:group-hosts() ! xdmp:host-name(.)
                else
                  (: If the user has specified host(s), assume they know
                     what they're doing. :)
                  tokenize($hosts, "(,\s*|\s+)")
  return
    fold-left(
      function($list, $item) {
        if (empty($list))
        then
          $item
        else
          let $pick := xdmp:random(count($list))
          return
            (subsequence($list, 1, $pick), $item, subsequence($list, $pick+1))
      }, (), $hosts)
};

declare %private function sess:construct-response(
  $cluster-id as xs:string,
  $host-id as xs:string,
  $hosts as xs:string+,
  $session as xs:string,
  $expires as xs:dateTime
) as object-node()
{
  let $sessionExpires := string($expires)
  let $uploadExpires  := string($expires) (: allowed to be different :)

  let $sesuris  := $hosts ! concat("https://", .,
                                   ":", xdmp:get-request-port(),
                                   "/v1/opsdirector/session")
  let $upluris  := $hosts ! concat("https://", .,
                                   ":", xdmp:get-request-port(),
                                   "/v1/opsdirector/upload",
                                   "/{cluster-id}/{host-id}",
                                   "/{type}/{year}/{month}/{day}",
                                   "/{type}_{cluster-id}_{host-id}_{timestamp}_{md5}.{ext}")

  let $maxLog  := (config:get-atomic-property("session-max-logs-level-" || $cluster-id),
                   config:get-atomic-property("session-max-logs-level", "finest"))[1]

  let $maxLog  := if ($cluster-id eq string(xdmp:cluster())
                      and matches($maxLog, "fine.*"))
                  then
                    let $_ := trace("OpsDirector monitoring OpsDirector: "
                                    || "logs limited to 'debug'", $t:trace-status)
                    return
                     "debug"
                  else
                    $maxLog

  let $maxMeters := (config:get-atomic-property("session-max-meters-level-" || $cluster-id),
                     config:get-atomic-property("session-max-meters-level", "raw"))[1]

  return
    object-node { "apiVersion": "1.0.0",
                  "sessionId": string($session),
                  "sessionEndpoint": object-node {
                     "expires": $sessionExpires,
                     "urls": array-node {
                        $sesuris
                     }
                  },
                  "uploadEndpoint": object-node {
                     "expires": $uploadExpires,
                     "urls": array-node {
                        $upluris
                     }
                  },
                  "streams": array-node {
                     object-node {
                        "type": "errorLog",
                        "level": $maxLog
                     },
                     object-node {
                        "type": "meters",
                        "level": $maxMeters
                     },
                     object-node {
                        "type": "config",
                        "level": "frequent"
                     }
                  }
                }
};
