xquery version "1.0-ml";

(: Copyright 2015-2018 MarkLogic Corporation.  All Rights Reserved. :)

(: This module exists so that some functions can be amped for the
   opsdir:8011/connect endpoint; it should never be used outside
   that endpoint.
:)

module namespace setup="http://marklogic.com/v1/opsdirector/setup";

import module namespace sec="http://marklogic.com/xdmp/security"
    at "/MarkLogic/security.xqy";

import module namespace pki = "http://marklogic.com/xdmp/pki"
    at "/MarkLogic/pki.xqy";

import module namespace admin = "http://marklogic.com/xdmp/admin"
    at "/MarkLogic/admin.xqy";

import module namespace t="http://marklogic.com/v1/opsdirector"
    at "/common/opsdir.xqy";

import module namespace rs="http://marklogic.com/v1/opsdirector/resource-scopes"
    at "/common/rscopes.xqy";

import module namespace ts="http://marklogic.com/v1/opsdirector/security"
    at "/common/opsec.xqy";

declare namespace x509 = "http://marklogic.com/xdmp/x509";

declare default function namespace "http://www.w3.org/2005/xpath-functions";

declare option xdmp:mapping "false";

declare %private variable $iso-opts :=
  <options xmlns="xdmp:eval">
    <isolation>different-transaction</isolation>
    <transaction-mode>update-auto-commit</transaction-mode>
    <prevent-deadlocks>false</prevent-deadlocks>
  </options>;

declare %private variable $sec-opts :=
  <options xmlns="xdmp:eval">
    <database>{xdmp:security-database()}</database>
  </options>;

declare %private variable $credential-duration := xs:dayTimeDuration("P3650D");
declare %private variable $certificate-duration := xs:dayTimeDuration("P3650D");


declare function setup:add-managed-cluster(
  $hostname as xs:string,
  $port as xs:string,
  $cluster-id as xs:string,
  $credname as xs:string
)
{
  let $external-certs := string(doc("/config/certificate-authority.txt")) eq ""
  return
    if ($external-certs)
    then
      let $_ := xdmp:document-insert("/config/cluster/"||$cluster-id||".json",
                                     object-node {
                                       "created": string(current-dateTime()),
                                       "hostname": $hostname,
                                       "port": $port
                                     },
                                     $t:DBPERMS,
                                     $t:COLLROOT || "cluster-config")
      return
        "ok"
    else
      setup:add-managed-cluster-go($hostname, $port, $cluster-id, $credname)
};

declare function setup:add-managed-cluster-go(
  $hostname as xs:string,
  $port as xs:string,
  $cluster-id as xs:string,
  $credname as xs:string
)
{
  (: N.B. We explicitly don't check for an execute permission because the
     point of this function is to allow the unauthenticated user calling
     the connect endpoint to complete the process of creating the ability
     to make authenticated calls to us.
  :)
  let $cred  := xdmp:invoke-function(
                  function () {
                    setup:create-credential-to-access-managed-cluster(
                       $hostname, $port, $cluster-id)
                  }, $sec-opts)
  let $ok    := xdmp:invoke-function(
                  function () {
                    setup:can-connect-to-managed-cluster($hostname, $port, $cred)
                  }, $iso-opts)
  return
    if ($ok)
    then
      let $parts   := setup:create-managed-cluster-cert($cluster-id, $hostname)
      let $cert    := $parts[1]
      let $pubkey  := $parts[2]
      let $privkey := $parts[3]
      let $_       := setup:create-credential-on-managed-cluster(
                        $hostname, $port, $cred, $credname, $cert, $privkey)
      let $passwd  := xdmp:hmac-sha256("somesecret", xdmp:random()||xdmp:random(), "hex")
      let $user    := try {
                        xdmp:user("opsdir-" || $cluster-id)
                      } catch ($e) {
                        xdmp:invoke-function(
                          function () {
                            sec:create-user("opsdir-" || $cluster-id,
                                            "Ops Director data from cluster " || $cluster-id,
                                            $passwd, ("opsdir-admin",
                                                      "opsdir-data-internal"), (), (),
                                            xdmp:x509-certificate-subject-name($cert))
                          }, $sec-opts)
                      }

      let $_ := xdmp:document-insert("/config/cluster/"||$cluster-id||".json",
                                     object-node {
                                       "created": string(current-dateTime()),
                                       "hostname": $hostname,
                                       "port": $port
                                     },
                                     $t:DBPERMS,
                                     $t:COLLROOT || "cluster-config")
      return
        "ok"
    else
      let $credname := "OpsDirector-" || $cluster-id
      let $_ := xdmp:invoke-function(
                   function () {
                      sec:remove-credential($credname)
                   }, $sec-opts)
      return
        error((), "OPSDIR-NOCONN", "Failed to connect")
};

declare function setup:remove-managed-cluster(
  $hostname as xs:string,
  $cluster-id as xs:string
)
{
  let $_ := trace("Removing managed cluster " || $cluster-id || " for " || $hostname,
                  $t:trace-status)
  let $_ := xdmp:invoke-function(
              function () {
                setup:remove-credential-to-access-managed-cluster($hostname, $cluster-id)
              }, $sec-opts)
  let $_ := setup:remove-managed-cluster-cert($cluster-id, $hostname)
  let $_ := try {
              xdmp:invoke-function(
                function () {
                  let $_ := trace("Removing user opsdir-" || $cluster-id, $t:trace-status)
                  return
                    sec:remove-user("opsdir-" || $cluster-id)
                }, $sec-opts)
            } catch ($e) {
              if ($e/error:code = "SEC-USERDNE")
              then ()
              else xdmp:rethrow()
            }

  let $_ := setup:cleanup-data($cluster-id)
  return
    "ok"
};

declare %private function setup:cleanup-data(
  $cluster-id as xs:string
) as empty-sequence()
{
  let $alluris := cts:uris((), (),
                    cts:and-query(
                      (cts:collection-query($t:COLLROOT || "resource"),
                       cts:element-range-query(
                          xs:QName("opsdirClusterId"), "=", $cluster-id,
                                   ("collation=http://marklogic.com/collation/")))))
  let $_ := trace("Resetting collections for resources from disconnected cluster: "
                  || count($alluris) || " uris", $t:trace-status)
  let $_ := setup:reset-collections-on-uris($alluris)

  (: Now what about resource groups... :)
  for $type in $t:RESOURCE-TYPES
  let $ids := t:managed-resource-ids($cluster-id, $type)
  where $ids
  return
    let $rg  := rs:resource-groups-containing-resources($type, $ids)
    where exists($rg)
    return
      let $_ := trace("Removing " || $type || " resources from resource group(s)",
                      $t:trace-status)
      return
        ts:remove-resources-from-resource-group($type, $ids)
};

declare %private function setup:reset-collections-on-uris(
  $uris as xs:string*
) as empty-sequence()
{
  (: Loop and do this in groups of $size; use the task queue so that
     we don't try to do too many, and we don't blow the stack :)
  let $size := 1000
  where exists($uris)
  return
    xdmp:spawn-function(
      function () {
        setup:do-reset-collections-on-uris($uris, $size)
      },
      <options xmlns="xdmp:eval">
        <update>true</update>
        <commit>auto</commit>
      </options>)
};

declare %private function setup:do-reset-collections-on-uris(
  $uris as xs:string*,
  $size as xs:unsignedLong
) as empty-sequence()
{
  let $fix  := subsequence($uris, 1, $size)
  let $_    := trace("Resetting collections for " || count($fix) || " uris",
                     $t:trace-status)
  let $_    := $fix ! xdmp:document-set-collections(., ())
  let $uris := subsequence($uris, $size+1)
  where exists($uris)
  return
    xdmp:spawn-function(
      function () {
        setup:do-reset-collections-on-uris($uris, $size)
      },
      <options xmlns="xdmp:eval">
        <update>true</update>
        <commit>auto</commit>
      </options>)
};

declare %private function setup:create-credential-to-access-managed-cluster(
  $hostname as xs:string,
  $port as xs:string,
  $cluster-id as xs:string
) as xs:unsignedLong
{
  let $credname := "opsdir-" || $cluster-id
  return
    try {
      let $cred := sec:get-credential($credname)
      return
        $cred/sec:credential-id
    } catch ($e) {
      let $keys    := xdmp:rsa-generate()
      let $privkey := $keys[1]
      let $pubkey  := $keys[2]
      let $x509    :=  element x509:cert {
                         element x509:version {2},
                         element x509:serialNumber {pki:integer-to-hex(xdmp:random())},
                         element x509:validity {
                           element x509:notBefore {
                             fn:current-dateTime()
                           },
                           element x509:notAfter {
                             fn:current-dateTime() + $credential-duration
                           }
                         },
                         element x509:subject {
                           element x509:organizationName { "MarkLogic Corporation" },
                           element x509:commonName
                             { "OpsDirector" }
                         },
                         element x509:publicKey {$pubkey},
                         element x509:v3ext {
                           element x509:subjectKeyIdentifier {
                             attribute critical {"false"},
                             pki:integer-to-hex(xdmp:random())
                           }
                         }
                       }
      let $certificate := xdmp:x509-certificate-generate($x509, (),
                             <options xmlns="ssl:options">
                               <credential-id>{xdmp:credential-id($t:CRED-NAME)}</credential-id>
                             </options>)

      let $_ := xdmp:document-insert("/certs/out/"||$cluster-id,
                                     text { $certificate },
                                     $t:ADMIN-DBPERMS,
                                     $t:COLLROOT || "certificates")

      let $uri-pattern := "https://.*:" || $port || "/manage/.*"
      return
        sec:create-credential($credname,
                              "A credential for accessing cluster " || $cluster-id,
                              (), (), $certificate, $privkey, fn:false(),
                              sec:uri-credential-target($uri-pattern, "basic"),
                              (xdmp:permission("app-user", "read"),
                               xdmp:permission("opsdir-data-internal", "read"),
                               xdmp:permission("opsdir-user", "read")))
    }
};

declare %private function setup:remove-credential-to-access-managed-cluster(
  $hostname as xs:string,
  $cluster-id as xs:string
) as empty-sequence()
{
  let $credname := "opsdir-" || $cluster-id
  let $_        := try {
                     sec:remove-credential($credname)
                   } catch ($e) {
                     if ($e/error:code = "SEC-CREDENTIALDNE")
                     then ()
                     else xdmp:rethrow()
                   }
  let $_        := try {
                     xdmp:document-delete("/certs/out/"||$cluster-id)
                   } catch ($e) {
                     if ($e/error:code = "XDMP-DOCNOTFOUND")
                     then ()
                     else xdmp:rethrow()
                   }
  return
    ()
};

declare %private function setup:can-connect-to-managed-cluster(
  $hostname as xs:string,
  $port as xs:string,
  $cred as xs:unsignedLong
) as xs:boolean
{
  let $uri  := "https://" || $hostname || ":" || $port || "/manage/v2"
  let $resp := try {
                 xdmp:http-get($uri,
                    <options xmlns="xdmp:http">
                      <headers>
                        <accept>application/json</accept>
                      </headers>
                      <credential-id>{$cred}</credential-id>
                      <timeout>600</timeout>
                      <verify-cert>false</verify-cert>
                    </options>)
               } catch ($e) {
                 xdmp:log($e)
               }
  return
    if ($resp[1]/*:code = 200)
    then
      true()
    else
      (xdmp:log(("Failed to connect to managed cluster: " || $uri), $resp),
       false())
};

declare %private function setup:create-managed-cluster-cert(
  $cluster-id as xs:string,
  $hostname as xs:string
)
{
  let $keys     := xdmp:rsa-generate()
  let $privkey  := $keys[1]
  let $pubkey   := $keys[2]
  let $x509     :=
    element x509:cert {
      element x509:version {2},
      element x509:serialNumber {pki:integer-to-hex(xdmp:random())},
      element x509:validity {
        element x509:notBefore {
          fn:current-dateTime()
        },
        element x509:notAfter  {
          fn:current-dateTime() + $certificate-duration
        }
      },
      element x509:subject {
        element x509:organizationName { "MarkLogic Corporation" },
        element x509:commonName { "OpsDirector-" || $cluster-id }
      },
      element x509:publicKey { $pubkey },
      element x509:v3ext {
        element x509:subjectKeyIdentifier {
          attribute critical {"false"},
          pki:integer-to-hex(xdmp:random())
        }
      }
    }

  let $certificate :=
    xdmp:x509-certificate-generate($x509, (),
      <options xmlns="ssl:options">
        <credential-id>{xdmp:credential-id($t:CRED-NAME)}</credential-id>
      </options>)

  let $_ := xdmp:document-insert("/certs/in/"||$cluster-id,
                                 text { $certificate },
                                 $t:ADMIN-DBPERMS,
                                 $t:COLLROOT || "certificates")

  return
    ($certificate, $pubkey, $privkey)
};

declare %private function setup:remove-managed-cluster-cert(
  $cluster-id as xs:string,
  $hostname as xs:string
)
{
  let $uri := "/certs/in/"||$cluster-id
  where doc-available($uri)
  return
    let $_ := trace("Removing cert: " || $uri, $t:trace-status)
    return
      xdmp:document-delete($uri)
};

declare %private function setup:create-credential-on-managed-cluster(
  $hostname as xs:string,
  $port as xs:string,
  $cred as xs:unsignedLong,
  $credname as xs:string,
  $certificate as xs:string,
  $privkey as xs:string
)
{
  let $config := admin:get-configuration()
  let $ioport := admin:appserver-get-port($config, xdmp:server())
  let $payload :=
    <secure-credential-properties xmlns="http://marklogic.com/manage">
      <name>{$credname}</name>
      <description>Credential for accessing OpsDirector system server</description>
      <certificate>{$certificate}</certificate>
      <private-key>{$privkey}</private-key>
      <targets>
        <target>
          <uri-pattern>https://.*:{$ioport}/.*</uri-pattern>
          <authentication>basic</authentication>
        </target>
      </targets>
      <signing>true</signing>
      <permissions></permissions>
    </secure-credential-properties>
  let $uri  := "https://" || $hostname || ":" || $port || "/manage/v2/credentials/secure"
  let $resp := try {
                 xdmp:http-post($uri,
                    <options xmlns="xdmp:http">
                      <headers>
                        <content-type>application/xml</content-type>
                      </headers>
                      <credential-id>{$cred}</credential-id>
                      <timeout>15</timeout>
                      <verify-cert>false</verify-cert>
                    </options>,
                    $payload)
               } catch ($e) {
                 xdmp:log($e),
                 xdmp:rethrow()
               }
  return
    ()
};
