xquery version "1.0-ml";

(: Copyright 2014-2018 MarkLogic Corporation.  All Rights Reserved. :)

import module namespace t="http://marklogic.com/v1/opsdirector"
    at "/common/opsdir.xqy";

import module namespace p="http://marklogic.com/v1/opsdirector/polling"
    at "/common/polling.xqy";

import module namespace config="http://marklogic.com/v1/opsdirector/config"
    at "/common/config.xqy";

declare variable $infinity := xs:yearMonthDuration("P9999Y");
declare variable $max-meters-age
  := local:duration(config:get-atomic-property("max-meters-age"));
declare variable $max-logs-age
  := local:duration(config:get-atomic-property("max-logs-age"));
declare variable $max-resource-age
  := local:duration(config:get-atomic-property("max-resource-age"));

declare function local:duration(
  $duration as xs:string?
) as xs:duration
{
  if ($duration castable as xs:dayTimeDuration)
  then xs:dayTimeDuration($duration)
  else if ($duration castable as xs:yearMonthDuration)
       then xs:yearMonthDuration($duration)
       else ($infinity,
             if ($duration != "infinity")
             then xdmp:log("Unparsable maximum age: " || $duration)
             else ())
};

declare function local:expire-resources(
) as empty-sequence()
{
  let $cutoff := current-dateTime() - $max-resource-age
  let $q-end  := cts:json-property-range-query("opsdirDtend", "<", $cutoff)
  let $q-rsrc := cts:not-query(cts:collection-query($t:COLLROOT||"resource"))
  let $uris   := cts:uris((), (), cts:and-query(($q-end, $q-rsrc)))
  let $_      := trace("Expire resources removes " || count($uris) || " documents "
                       || "older than " || string($max-resource-age),
                       $t:trace-expires)
  return
    local:delete-uris($uris)
};

declare function local:expire-logs(
) as empty-sequence()
{
  let $cutoff := current-dateTime() - $max-logs-age
  let $q-end  := cts:json-property-range-query("ts", "<", $cutoff)
  let $q-logs := cts:collection-query($t:COLLROOT||"logs")
  let $uris   := cts:uris((), (), cts:and-query(($q-end, $q-logs)))
  let $_      := trace("Expire logs removes " || count($uris) || " documents "
                       || "older than " || string($max-logs-age),
                       $t:trace-expires)
  return
    local:delete-uris($uris)
};

declare function local:expire-meters(
) as empty-sequence()
{
  let $cutoff   := current-dateTime() - $max-meters-age
  let $qname    := fn:QName("http://marklogic.com/manage/meters", "end-time")
  let $q-end    := cts:element-range-query($qname, "<", $cutoff)
  let $q-meters := cts:collection-query($t:COLLROOT||"meters")
  let $uris     := cts:uris((), (), cts:and-query(($q-end, $q-meters)))
  let $_       := trace("Expire meters removes " || count($uris) || " documents "
                       || "older than " || string($max-meters-age),
                        $t:trace-expires)
  return
    local:delete-uris($uris)
};

declare function local:delete-uris(
  $uris as xs:string*
) as empty-sequence()
{
  (: Loop and do this in groups of $size; use the task queue so that
     we don't try to do too many, and we don't blow the stack :)
  let $size := 1000
  where exists($uris)
  return
    xdmp:spawn-function(
      function () {
        local:do-delete-uris($uris, $size)
      },
      <options xmlns="xdmp:eval">
        <update>true</update>
        <commit>auto</commit>
      </options>)
};

declare function local:do-delete-uris(
  $uris as xs:string*,
  $size as xs:unsignedLong
) as empty-sequence()
{
  let $_    := subsequence($uris, 1, $size) ! xdmp:document-delete(.)
  let $_    := if (xdmp:trace-enabled($t:trace-expires)
                   and xdmp:trace-enabled($t:trace-debug))
               then trace(("Deleting:", subsequence($uris, 1, $size)), $t:trace-expires)
               else ()
  let $uris := subsequence($uris, $size+1)
  where exists($uris)
  return
    xdmp:spawn-function(
      function () {
        local:do-delete-uris($uris, $size)
      },
      <options xmlns="xdmp:eval">
        <update>true</update>
        <commit>auto</commit>
      </options>)
};

if (p:host-to-run())
then
  let $_ := trace("Running expire on " || xdmp:host-name(), $t:trace-expires)
  return
    (local:expire-logs(),
     local:expire-resources(),
     local:expire-meters())
else
  let $_ := trace("Not running alerts poll on " || xdmp:host-name(), $t:trace-expires)
  return
    ()